(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

open List;;

open Trace;;
open Location;;
open Abstract;;
open Symbols;;
open Termutil;;

exception Alias_error of string * location;;
 

(**************************************************************
 * Detection of loop in alias decl
 * TBC
 **************************************************************)
  
let rec aliaslist_of_lterm sl term = 
  match term.lte_term with
      EVA_TERM_ALIAS (i, _) -> (
        try
          if (is_alias (assoc_symbol i sl))
          then [i]
          else raise (Internal_error ("aliaslist_of_lterm:^"^
                                      " symbol "^i^" is not an alias"))
        with
            Symbols_not_found -> 
              raise (Alias_error (("undeclared alias symbol "^i), term.lte_loc))
      )  
    | EVA_P t      
    | EVA_A t      
    | EVA_SA t      
    | EVA_AA t      
    | EVA_U (_, t) ->
        aliaslist_of_lterm sl t
    | EVA_CONS tl -> 
        set_flatten (map (aliaslist_of_lterm sl) tl)
    (* it is assumed (not checked) that f is not the id of an alias *)
    (* app_root = f; app_args = tl *)
    | EVA_APP (_, tl) ->
        set_flatten (map (aliaslist_of_lterm sl) tl)
    (* it is assumed (not checked) that a does not contain alias ids *)
    (* crypt_algo = a; crypt_key = k; crypt_content = t *)
    | EVA_CRYPT (_, k, t) ->
        set_flatten (map (aliaslist_of_lterm sl) [k; t])
    (* it is assumed (not checked) that a does not contain alias ids *)
    (* crypt_algo = a; crypt_key = k; crypt_content = t *)
    | EVA_SIGN (_, k ,t) -> 
        set_flatten (map (aliaslist_of_lterm sl) [k; t])
    | EVA_PCENT (t1, t2) ->
        set_flatten (map (aliaslist_of_lterm sl) [t1; t2])
    (* it is assumed (not checked) that v is not the id of an alias *)
    | EVA_TERM_VALUE v -> []
    (* it is assumed (not checked) that var is not the id of an alias *)
    | EVA_LOCATED (var, role) -> []
    | _ -> []
(* flatten without repetition *)
and set_flatten = function
    [] -> []
  | (l::ll) -> (set_append l (set_flatten ll))
and set_append l1 l2 = 
  match l1 with
      [] -> l2
    | (x::l) -> 
        if (mem x l2)
        then (set_append l l2)
        else (x::(set_append l l2));;

(* TBC!! construction du graph des alias et detection de circuit *)



(* has_alias sl term
 * returns true if the given located term term contains
 * a symbol registered as an alias in the given symbol list sl.
 * Returns false otherwise. 
 *
 * It is assumes that only the leaves EVA_TERM_ID
 * and EVA_TERM_ALIAS can contain alias symbols.  *)
let rec has_alias sl term =
  (* trace "Alias.has_alias\n"; *)
  match term.lte_term with
      EVA_TERM_ID i -> (
        try
          (* TBC: limiter les TERM_ID aux id du f.o.: tjs false dans ce cas *)
          (is_alias (assoc_symbol i sl))
        with
            Symbols_not_found -> 
              raise (Alias_error (("undeclared symbol "^i), term.lte_loc))
      )
    | EVA_TERM_ALIAS (i, _) -> (
        try
          if (is_alias (assoc_symbol i sl))
          then true
          else raise (Internal_error 
                        ("has_alias: symbol "^i^" is not an alias"))
          with
              Symbols_not_found -> 
                raise (Alias_error (("undeclared alias symbol "^i), term.lte_loc))
      )
    | EVA_P t      
    | EVA_A t      
    | EVA_SA t      
    | EVA_AA t      
    | EVA_U (_, t) ->
        has_alias sl t
    | EVA_CONS tl -> 
        fold_left (fold_alias sl) false tl
      (* we assume that the root symbol is not an alias *)
      (* app_root = i; app_args = tl *)
    | EVA_APP (_, tl) -> 
        fold_left (fold_alias sl) false tl
      (* crypt_algo = a; crypt_key = k; crypt_content = t *)
    | EVA_CRYPT (_, k, t) -> 
        (has_alias sl k) || (has_alias sl t)
      (* crypt_algo = a; crypt_key = k; crypt_content = t *)
    | EVA_SIGN (_, k, t) -> 
        (has_alias sl k) || (has_alias sl t)
    | EVA_PCENT (t1, t2) -> 
        (has_alias sl t1) || (has_alias sl t2)
    | _ -> false
and fold_alias sl b term = b || (has_alias sl term);;


(**************************************************************
 * aliases replacement
 **************************************************************)

(* unalias_lterm sl term
 * returns the term obtained by 
 * replacement of the aliases in the given located term term,
 * according to the given symbol list sl *)
let rec unalias_lterm sl term = 
  (* trace "Alias.unalias_lterm\n"; *)
  if (not (has_alias sl term))
  then term
  else (
    let unt = (
      match term.lte_term with
          (* the id i must be an alias *)
          EVA_TERM_ALIAS (a, ta) -> (
            trace "Alias.unalias_lterm: %s\n" a;
            ta
          (*
            try
              (get_term (assoc_symbol a sl))
            with
                Symbols_not_found -> 
                  raise (Alias_error (("undeclared alias symbol "^i), t.lte_loc))
              | (Symbols_error msg) -> 
                  raise (Internal_error ("unalias_lterm "^i^": "^msg))
          *)
          )
          (* replace if the id i is an alias *)
        | EVA_TERM_ID i -> (
            try
              if (is_alias (assoc_symbol i sl))
              then (get_term (assoc_symbol i sl))
              else term
            with
                Symbols_not_found -> 
                  raise (Alias_error (("undeclared alias symbol "^i), term.lte_loc))
              | (Symbols_error msg) -> 
                  raise (Internal_error ("unalias_lterm "^i^": "^msg))
          )
        | EVA_P t ->
            { term with lte_term =  EVA_P (unalias_lterm sl t) }
        | EVA_A t ->
            { term with lte_term =  EVA_A (unalias_lterm sl t) }
        | EVA_SA t ->    
            { term with lte_term =  EVA_SA (unalias_lterm sl t) }
        | EVA_AA t ->     
            { term with lte_term =  EVA_AA (unalias_lterm sl t) }
        | EVA_U (i, t) ->
            { term with lte_term =  EVA_U (i, (unalias_lterm sl t)) }
        | EVA_CONS tl -> 
            { term with lte_term = EVA_CONS (map (unalias_lterm sl) tl) }
        (* app_root = i; app_args = tl *)
        | EVA_APP (i, tl) -> 
            (* note: the syntax does not allow an alias to a functional symbol
               i.e. no TERM_ID i with i symbol of function type *)
            { term with lte_term = EVA_APP (i, (map (unalias_lterm sl) tl)) }
        (* crypt_algo = a; crypt_key = k; crypt_content = t *)
        | EVA_CRYPT (a, k, t) ->   
            { term with lte_term = 
                EVA_CRYPT (a, (unalias_lterm sl k), (unalias_lterm sl t)) }
        (* crypt_algo = a; crypt_key = k; crypt_content = t *)
        | EVA_SIGN (a, k, t) ->   
            { term with lte_term = 
                EVA_SIGN (a, (unalias_lterm sl k), (unalias_lterm sl t)) }
        | EVA_PCENT (t1, t2) -> 
            { term with lte_term = 
                EVA_PCENT ((unalias_lterm sl t1), (unalias_lterm sl t2)) }
        (* the other terms cannot contain aliases, should not be here *)
        | _ -> raise (Internal_error ("unalias_lterm: term has no alias"))
    )
    (* loop until all alias sumbols disapeared *)
    in (unalias_lterm sl unt)
  ) (* end else *)
;;


(* unalias_ldcl sl d
 * unalias the terms in the given located declaration d,
 * only if the declaration is not an alias decl. *)
let rec unalias_ldcl sl d = 
  trace "Alias.unalias_ldcl\n";
  { d with ldc_dcl = (unalias_dcl sl d.ldc_dcl) }
(* returns a eva_dcl *)
and unalias_dcl sl = function
    (* know_principal = i; know_contents = tl; *)
    EVA_DCL_KNOW (i, tl) -> 
      EVA_DCL_KNOW (i, (map (unalias_lterm sl) tl))
  | EVA_DCL_EVERY tl -> 
      EVA_DCL_EVERY (map (unalias_lterm sl) tl)
    (* ax_forall = itl; ax_left = t1; ax_right = t2 *)
  | EVA_DCL_AXIOM (t1, t2, t, itl) ->
      EVA_DCL_AXIOM ((unalias_lterm sl t1), (unalias_lterm sl t2), t, itl)
  | dd -> dd;;
  
(* unalias_spec sl sp
 * unalias all the terms in the given spec sp,
 * except in alias declarations *)
let rec unalias_spec sl sp = 
  trace "Alias.unalias_spec\n";
  let tr = (unalias_lterm sl) in
    {  sp with 
         sp_declarations = (map (unalias_ldcl sl) sp.sp_declarations);
         sp_protocol = (lterm_replace_lprotocol tr sp.sp_protocol);
         sp_statements = (map (lterm_replace_lstatement tr) sp.sp_statements);
    };;

