(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(*
 * Dump specifications in abstract syntax for
 * debugging puposes.
 *)

open Format;;

open Abstract;;
open Symbols;;
open Pp;;


(***************************************************************
 * Separators for pp_list
 ***************************************************************)

(* standard comma separation to use for evapp_list *)
let comma_sep () = 
  print_string ","; print_space ();;

(* standard semi-column separation to use for evapp_list *)
let semicol_sep () = 
  print_string ";"; print_space ();;

(* standard null separation to use for evapp_list *)
let null_sep () = ();;

(* standard new line separation to use for evapp_list *)
let newline_sep () = 
  print_newline ();;



(***************************************************************
 * Pretty printing functions
 ***************************************************************)

let rec debugpp_label l =
  begin
    open_box 0;
    (match l with
         EVA_NOLABEL -> 
           print_string "NOLABEL"
       | (EVA_LABEL s) -> 
           print_string "LABEL";
           print_space ();
           print_string s);
    close_box ();
  end
and debugpp_llabel { lla_loc = _; lla_label = l } =
  debugpp_label l;;

let rec debugpp_type t =
  begin
    open_box 0;
    (match t with
         EVA_VOID -> print_string "\"void\""
       | EVA_PRINCIPAL -> print_string "\"principal\""
       | EVA_NUMBER -> print_string "\"number\""
       | EVA_AALGO -> print_string "\"asym_algo\""
       | EVA_SALGO -> print_string "\"sym_algo\""
       | EVA_TALGO -> print_string "\"*algo*\""
       | (EVA_USERTYPE s) 
       | (EVA_BASETYPE s) -> 
           print_string "\""; print_string s; print_string "\""
       | (EVA_FUN t1) ->
           debugpp_type EVA_NUMBER;
           print_space (); print_string "->"; print_space ();
           debugpp_type t1;
       | (EVA_HFUN t1) ->
           debugpp_type EVA_NUMBER;
           print_space (); print_string "-h->"; print_space ();
           debugpp_type t1;           
       | (EVA_KEY t1) ->
           debugpp_type EVA_NUMBER;
           print_space (); print_string "-k->"; print_space ();
           debugpp_type t1;                      
    );
    close_box ();
  end
and debugpp_ltype { lty_loc = _; lty_type = t } =
  debugpp_type t;;

let rec debugpp_scope s =
  begin
    open_box 0;
    (match s with
       (*  EVA_SCOPE_VOID -> print_string "" *)
         EVA_CST -> print_string "constant"
       | EVA_PARAM -> print_string "parameter"
       | EVA_PRIVATE -> print_string "private");
    close_box ();
  end
and debugpp_lscope { lsc_loc = _; lsc_scope = s } =
  debugpp_scope s;;

let rec debugpp_id i =
  begin
    open_box 0;
    (match i with
         (* EVA_ID_VOID -> print_string "" *)
         (EVA_ID s) -> 
           print_string "ID"; print_space ();
           print_string s);
    close_box ();
  end
and debugpp_lid { lid_loc = _; lid_id = i }  =
  debugpp_id i;;

let rec debugpp_value v =
  begin
    open_box 0;
    (match v with
       (*  EVA_VALUE_VOID -> print_string "" *)
         EVA_INTRUDER -> print_string "INTRUDER"
       | (EVA_VALUE s) -> 
           print_string "VALUE";
           print_space ();
           print_string s);
    close_box ();
  end
and debugpp_lvalue { lva_loc = _; lva_value = v } =
  debugpp_value v;;

let rec debugpp_term t =
  begin
    open_box 0;
    (match t with
         (* EVA_TERM_VOID -> print_string "" *)
         EVA_TERM_ID s -> 
           print_string "TERM_ID"; print_space ();
           print_string s
       | EVA_TERM_CST s -> 
           print_string "TERM_CST"; print_space ();
           print_string s
       | EVA_CONS tl -> 
           print_string "CONS(";
           pp_list debugpp_lterm comma_sep tl;
           print_string ")";           
       | EVA_P t ->
           print_string "P(";           
           debugpp_lterm t;
           print_string ")"                      
       | EVA_A t ->
           print_string "A(";           
           debugpp_lterm t;
           print_string ")"                      
       | EVA_SA t ->
           print_string "SA(";           
           debugpp_lterm t;
           print_string ")"                      
       | EVA_AA t ->
           print_string "AA(";           
           debugpp_lterm t;
           print_string ")"                      
       | EVA_U (i, t) ->
           print_string "U(";           
           print_string i;
           comma_sep ();
           debugpp_lterm t;
           print_string ")"                      
       | EVA_APP (i, etl) -> 
           print_string "APP(";           
           debugpp_app (i, etl);
           print_string ")"           
       | EVA_CRYPT (a, k, t) -> 
           print_string "CRYPT(";           
           debugpp_cipher (a, k, t);
           print_string ")"           
       | EVA_SIGN (a, k, t) -> 
           print_string "SIGN(";           
           debugpp_cipher (a, k, t);
           print_string ")"           
       | EVA_PCENT (t1, t2) -> 
           print_string "PCENT(";           
           pp_list debugpp_lterm comma_sep [t1; t2];
           print_string ")"                      
       | (EVA_TERM_VALUE v) -> 
           print_string "TERM_VALUE"; print_space ();           
           debugpp_value v
       | (EVA_TERM_ALIAS (a, ta)) -> 
           print_string "ALIAS(";           
           print_string a;
           comma_sep ();
           debugpp_lterm ta;
           print_string ")"
       | (EVA_LOCATED (i, r)) -> 
           print_string "@(";           
           debugpp_lid i; 
           comma_sep ();
           debugpp_lid r; 
           print_string ")"
       | (EVA_QUANTIFIED (i, t)) -> 
           print_string "QUANTIFIED(";           
           print_string i;
           comma_sep ();
           debugpp_type t; 
           print_string ")"           
    );
    close_box ();
  end
and debugpp_lterm { lte_loc = _; lte_term = t } =
  debugpp_term t
(* app_root = i; app_args = etl *)
and debugpp_app (i, etl) =
  debugpp_lid i; 
  comma_sep ();
  pp_list debugpp_lterm comma_sep etl; 
(* crypt_algo = a; crypt_key = k; crypt_content = t *)
and debugpp_cipher (a, k, t) =
  debugpp_algo a;
  comma_sep ();
  debugpp_lterm k;
  comma_sep ();
  debugpp_lterm t  
and debugpp_algo = function
    EVA_VANILLA -> 
      print_string "VANILLA"
  | (EVA_ALGO t) -> 
      print_string "ALGO"; print_space ();
      debugpp_lterm t;;

let rec debugpp_dcl_ids (ibl, s, t) =
  begin
    open_box 0;
    debugpp_lscope s;
    print_space ();
    pp_list debugpp_singleid comma_sep ibl;
    print_string " :";
    print_space ();
    debugpp_ltype t;
    close_box ();
  end
and debugpp_singleid (i, f) =
  debugpp_fresh f; 
  debugpp_lid i 
and debugpp_fresh f =
  if f 
  then begin print_string "fresh"; print_space (); end
  else ();;

let rec debugpp_dcl_fun (i, s, t, tl, bh, bs) =
  begin
    open_box 0;
    debugpp_lscope s;
    print_space ();
    debugpp_lid i;
    print_string "(";
    pp_list debugpp_ltype comma_sep tl;
    print_string ") :";
    print_space ();
    debugpp_ltype t;
    debugpp_hash bh;
    debugpp_secret bs;
    close_box ();
  end
and debugpp_hash h =
  if h 
  then begin print_string "hash"; print_space (); end
  else ()
and debugpp_secret s =
  if s
  then print_string "secret" 
  else ();;

let debugpp_dcl_keypair (a, k1, k2, s, t, tl) =
  begin
    open_box 0;
    debugpp_lscope s; print_space ();
    print_string "keypair"; print_space ();
    debugpp_algo a; print_space ();
    debugpp_lid k1; comma_sep ();
    debugpp_lid k2;
    print_space (); print_string "(";
    pp_list debugpp_ltype comma_sep tl;
    print_string ") :"; print_space ();
    debugpp_ltype t;
    close_box ();
  end;;

(* know_principal = i; know_contents = tl *)
let debugpp_dcl_know (i, tl) =
  begin 
    debugpp_lid i; print_space (); 
    print_string "knows"; print_space ();
    pp_list debugpp_lterm comma_sep tl; 
  end;;

let rec debugpp_dcl d =
  begin
    open_box 0;
    (match d with
         EVA_DCL_ID (ibl, s, t) -> 
           debugpp_dcl_ids (ibl, s, t)
       | EVA_DCL_FUN (i, s, t, tl, bh, bs) ->  
           debugpp_dcl_fun (i, s, t, tl, bh, bs)
       | EVA_DCL_KEYPAIR (a, k1, k2, s, t, tl) -> 
           debugpp_dcl_keypair (a, k1, k2, s, t, tl)
       | EVA_DCL_ALIAS (i, t) -> 
           begin 
             print_string "alias"; print_space (); 
             debugpp_lid i; print_space (); 
             print_string "="; print_space ();
             debugpp_lterm t; 
           end
       | EVA_DCL_LOCALALIAS (il, i, t) -> ()
       | EVA_DCL_BASETYPE s -> 
           begin 
             print_string "basetype"; print_space (); 
             print_string s; 
           end
       | EVA_DCL_KNOW (i, tl) -> debugpp_dcl_know (i, tl)
       | EVA_DCL_EVERY tl -> 
           begin 
             print_string "everybody knows"; print_space (); 
             pp_list debugpp_lterm comma_sep tl; 
           end
       | EVA_DCL_AXIOM (t1, t2, t, ql) -> debugpp_axiom (t1, t2, t, ql)
       | EVA_DCL_VALUE (vl, t) -> debugpp_dcl_value (vl, t)
    );
    close_box ();
    print_newline ();
  end
and debugpp_ldcl { ldc_loc = _; ldc_dcl = d } =
  debugpp_dcl d
(* ax_forall = ql; ax_left = t1; ax_right = t2 *)
and debugpp_axiom (t1, t2, t, ql) =
  begin 
    print_string "axiom"; print_space (); 
    debugpp_lterm t1; print_space ();
    print_string "="; print_space (); 
    debugpp_lterm t2; print_space (); 
    debugpp_type t; print_space (); 
    print_string "["; print_space (); 
    pp_list debugpp_optional_quantification comma_sep ql;
    print_space (); print_string "]"; 
  end
and debugpp_optional_quantification (i, t) =
  begin 
    debugpp_lid i;
    (if ( t.lty_type <> EVA_VOID )
     then 
       begin 
         print_space (); print_string ":"; print_space (); 
         debugpp_ltype t; 
       end);
  end
and debugpp_dcl_value (vl, t) =
  begin 
    open_box 0;
    print_string "value"; print_space ();
    pp_list debugpp_lvalue comma_sep vl; print_space ();
    print_string ":"; print_space (); 
    debugpp_ltype t; 
    close_box ();
    print_newline ();
  end;;

let rec debugpp_instr i = 
  begin
    open_box 0;
    (match i with
         EVA_SKIP -> 
           print_string "SKIP"
       | EVA_MSG (l, is, ir, t) -> 
           debugpp_msg (l, is, ir, t)
       | EVA_ASSIGN (ir, iv, t) -> 
           debugpp_assign (ir, iv, t)
       | EVA_COMP (ir, iv, t)
         -> debugpp_compare (ir, iv, t)
       | EVA_BLOCK b -> debugpp_block b;
       | EVA_SWITCH (t, cl) -> 
           begin 
             print_string "switch"; 
             debugpp_lterm t; 
             print_newline(); print_string "{";
             open_box 1; 
             pp_list debugpp_case null_sep cl; 
             close_box (); 
             print_string "}" 
           end     
    );
    close_box ();
    print_newline ();
  end
and debugpp_linstr { lin_loc = _; lin_instr = i } =
  debugpp_instr i;
(* msg_label = l; msg_sender = is; msg_receiver = ir; msg_body = t *)
and debugpp_msg (l, is, ir, t) =
  begin 
    debugpp_llabel l; print_space (); 
    debugpp_lid is; print_space ();
    print_string "->"; print_space (); 
    debugpp_lid ir; print_space ();
    print_string ":"; print_space (); 
    debugpp_lterm t; 
  end
(* as_role = ir; as_variable = iv; as_value = t *)
and debugpp_assign (ir, iv, t) =
  begin 
    debugpp_lid ir; print_space (); 
    print_string ":"; print_space ();
    debugpp_lid iv; print_space (); 
    print_string ":="; print_space (); 
    debugpp_lterm t; 
  end
(* co_role = ir; co_left_variable = iv; co_right_variable = t *)
and debugpp_compare (ir, iv, t) =
  begin 
    debugpp_lid ir; print_space (); 
    print_string ":"; print_space ();
    debugpp_lid iv; print_space (); 
    print_string "=="; print_space (); 
    debugpp_lterm t; 
  end      
and debugpp_case { case_case = t; case_block = b; case_loc = _ } =
  begin open_box 0;
    print_string "case"; 
    debugpp_lterm t; print_space (); 
    print_string ":"; print_newline (); 
    debugpp_block b; 
    close_box (); 
  end
and debugpp_caselist = function
    [] -> ()
  | c::l -> 
      begin 
        debugpp_case c;  
        debugpp_caselist l 
      end
and debugpp_block b = 
  begin 
    open_box 0; 
    print_string "{"; print_newline ();
    open_box 1; 
    pp_list debugpp_linstr null_sep b; 
    close_box (); 
    print_string "}";
    close_box (); 
    print_newline (); 
  end;;

let rec debugpp_protocol = function
    (EVA_MP l) -> debugpp_block l
  | (EVA_PROCESS pl) -> debugpp_processlist pl
and debugpp_lprotocol { lpr_loc = _; lpr_protocol = p } =
  debugpp_protocol p
and debugpp_processlist = function
    [] -> ()
  | p::l -> 
      begin 
        debugpp_process p; 
        print_newline (); 
        debugpp_processlist l 
      end
and debugpp_process { pr_role = i; pr_privates = il; pr_messages = b } =
  begin 
    open_box 0;
    print_string "process"; print_space (); 
    debugpp_lid i; print_newline ();
    print_string "private"; print_space (); 
    pp_list debugpp_lid comma_sep il; print_newline ();
    debugpp_block b;
  end;;


let rec debugpp_constraint = function
    EVA_EQ_CONSTRAINT (i1, i2) -> 
      begin 
        debugpp_lid i1; print_space ();
        print_string "="; 
        debugpp_lid i2; 
      end
  | EVA_NEQ_CONSTRAINT (i1, i2) -> 
      begin 
        debugpp_lid i1; print_space ();
        print_string "!="; 
        debugpp_lid i2; 
      end
  | EVA_MEMBERSHIP_CONSTRAINT (i, tl) -> 
      begin 
        debugpp_lid i; print_space ();
        print_string ":"; print_space(); 
        print_string "{"; 
        pp_list debugpp_lterm comma_sep tl; print_space (); 
        print_string "}"; 
      end
  | EVA_DOMAIN_CONSTRAINT (i, s) -> 
      begin 
        debugpp_lid i; print_space ();
        print_string ":"; print_space(); 
        print_string s; 
      end
and debugpp_lconstraint { lco_loc = _; lco_constraint = c } =
  debugpp_constraint c
and debugpp_lconstraintlist = function
    [] -> ()
  | [ c ] -> debugpp_lconstraint c
  | c::l -> 
      begin 
        debugpp_lconstraint c; 
        print_string ";"; print_space(); 
        debugpp_lconstraintlist l 
      end;;


let rec debugpp_session s =
  begin
    open_box 0;
    (match s with
         EVA_BANG -> print_string "session*"
       | EVA_CONSTRAINED_SESSIONS l -> 
           begin 
             print_string "session*"; print_space(); 
             print_string "["; print_space();
             open_box 1; 
             debugpp_lconstraintlist l; 
             close_box (); 
             print_space();  print_string "]"; 
           end
       | EVA_SESSIONS al -> 
           begin  
             print_string "session*"; print_space (); 
             print_string "("; print_space ();  
             open_box 1; 
             pp_list debugpp_sassignment semicol_sep al; 
             close_box ();
             print_space (); print_string ")"; 
           end
       | EVA_SESSION (l, al) -> 
           begin 
             debugpp_llabel l; print_space (); 
             print_string "session"; print_space (); 
             print_string "("; print_space (); 
             open_box 1; 
             pp_list debugpp_sassignment semicol_sep al; 
             close_box ();
             print_space (); print_string ")"; 
           end
    );
    close_box ();
    print_newline ();
  end   
and debugpp_lsession { lse_loc = _; lse_session = s } = 
  debugpp_session s
and debugpp_sassignment (i, v) = 
  begin 
    debugpp_lid i; print_space(); 
    print_string "="; print_space (); 
    debugpp_lterm v; 
  end;;

let rec debugpp_atom = function
    EVA_TRUE -> print_string "true"
  | EVA_FALSE -> print_string ""
  | EVA_ATOM_EQ (t1, t2) -> 
      begin 
        debugpp_lterm t1; print_space ();
        print_string "=="; print_space (); 
        debugpp_lterm t2; 
      end
  | EVA_ATOM_NEQ (t1, t2) -> 
      begin debugpp_lterm t1; print_space ();
        print_string "!="; print_space (); 
        debugpp_lterm t2; 
      end
  | EVA_ATOM_HONEST t -> 
      begin 
        print_string "honest("; 
        debugpp_lterm t; 
        print_string ")"; 
      end
  | EVA_ATOM_SECRET  t -> 
      begin 
        print_string "secret("; 
        debugpp_lterm t; 
        print_string ")"; 
      end
  | EVA_ATOM_USER (i, t) -> 
      begin 
        print_string "(";
        debugpp_lid i;
        print_string ")"; 
        print_string "(";
        debugpp_lterm t; 
        print_string ")"; 
      end
and debugpp_latom { lat_loc = _; lat_atom = a } =
  debugpp_atom a;;


let rec debugpp_statement s = 
  begin
    open_box 0;
    (match s with
         EVA_ASSUME (ql, al, a) ->
           begin 
             print_string "assume"; print_space ();
             debugpp_formula al a; print_space();
             (match ql with
                  [] -> ()
                | _  -> 
                    begin 
                      print_string "["; print_space (); 
                      pp_list debugpp_optional_quantification comma_sep ql;
                      print_space (); print_string "]"; 
                    end
             );
           end
       | EVA_CLAIM (l, al, a) ->
           begin 
             print_string "claim"; print_space() ;
             debugpp_llabel l; print_space ();
             debugpp_formula al a; 
           end
    );
    close_box ();
    print_newline ();
  end
and debugpp_lstatement { lst_loc = _; lst_statement = s } =
  debugpp_statement s
(* for_tail = al; for_head = a *)
and debugpp_formula al a =
  begin 
    (match al with
         [] -> ()
       | _  -> 
           begin 
             pp_list debugpp_latom comma_sep al; print_space (); 
             print_string "=>"; print_space (); 
           end
    );
    debugpp_latom a; 
  end;;

let debugpp_spec = function 
    { sp_label = l;
      sp_declarations = dl;
      sp_protocol = p;
      sp_valuedcl = vdl;
      sp_sessions = sl;
      sp_statements = stl } ->
      begin
        open_box 0;
        debugpp_lid l;
        print_newline (); print_newline ();
        pp_list debugpp_ldcl null_sep dl;(* debugpp_ldcllist dl; *) 
        print_newline ();
        debugpp_lprotocol p;
        print_newline ();
        pp_list debugpp_ldcl null_sep vdl; (* debugpp_ldcllist vdl; *)
        print_newline ();
        pp_list debugpp_lsession null_sep sl;
        print_newline ();
        pp_list debugpp_lstatement null_sep stl;
        print_newline ();
        close_box ();
      end


(***************************************************************
 * dump symbol lists
 ***************************************************************)

let rec dump_symbol sy = 
  if (is_first_order sy)
  then 
    begin
      open_box 0;
      print_string "sy_fos{";
      open_box 1;
      print_string "name = ";
      print_string (get_name sy);
      print_string ";"; print_space ();
      print_string "scope = ";
      debugpp_scope (get_scope sy);
      print_string ";"; print_space ();
      print_string "type = ";
      debugpp_type (get_type sy);
      print_string ";"; print_space ();
      print_string "fresh = ";
      print_bool (get_fresh sy);
      close_box ();
      print_string "}";
      close_box ();
    end
  else if (is_functional sy)
  then
    begin
      open_box 0;
      print_string "sy_fun{";
      open_box 1;        
      print_string "name = ";
      print_string (get_name sy);
      print_string ";"; print_space ();
      print_string "scope = ";
      debugpp_scope (get_scope sy);
      print_string ";"; print_space ();
      print_string "type = ";
      debugpp_type (get_type sy);
      print_string ";"; print_space ();
      print_string "signature = ";
      dump_typelist (get_signature sy);
      print_string ";"; print_space ();
      print_string "hash = ";
      print_bool (get_hash sy);
      print_string ";"; print_space ();
      print_string "secret = ";
      print_bool (get_secret sy);
      close_box ();
      print_string "}";
      close_box ();
    end      
  else if (is_key sy)
  then
    begin
      open_box 0;
      print_string "sy_key{";
      open_box 1;
      print_string "name = ";
      print_string (get_name sy);
      print_string ";"; print_space ();
      print_string "scope = ";
      debugpp_scope (get_scope sy);
      print_string ";"; print_space ();
      print_string "type = ";
      debugpp_type (get_type sy);
      print_string ";"; print_space ();
      print_string "signature = ";
      dump_typelist (get_signature sy);
      print_string ";"; print_space ();
      print_string "symkey = ";
      print_string (get_symkey sy);
      print_string ";"; print_space ();
      print_string "algo = ";
      dump_algo (get_algo sy);
      close_box ();
      print_string "}";
      close_box ();
    end      
  else if (is_value sy)
  then
    begin
      open_box 0;
      print_string "sy_val{";
      open_box 1;
      print_string "name = ";
      print_string (get_name sy);
      print_string ";"; print_space ();
      print_string "type = ";
      debugpp_type (get_type sy);
      close_box ();
      print_string "}";
      close_box ();
    end       
  else if (is_intruder sy)
  then
    begin
        open_box 0;
        print_string "sy_intruder";
    end
  else if (is_quantified sy)
  then
    begin
      open_box 0;
      print_string "sy_qua{";
      open_box 1;
      print_string "name = ";
      print_string (get_name sy);
      print_string ";"; print_space ();
      print_string "type = ";
      debugpp_type (get_type sy);
      close_box ();
      print_string "}";
      close_box ();
    end      
  else if (is_alias sy)
  then
    begin
      open_box 0;
      print_string "sy_ali{";
      open_box 1;
      print_string "name = ";
      print_string (get_name sy);
      print_string ";"; print_space ();
      print_string "alias = ";
      debugpp_lterm (get_term sy);
      close_box ();
      print_string "}";
      close_box ();
    end     
  else if (is_type sy)
  then
      begin
        open_box 0;
        print_string "sy_typ{";
        open_box 1;       
        print_string "name = ";
        print_string (get_name sy);
        print_string ";"; print_space ();
        print_string "base = ";
        print_bool (is_basetype sy);
        close_box ();
        print_string "}";
        close_box ();
      end       
and dump_typelist tl = 
  pp_list debugpp_type comma_sep tl
and dump_algo = function
    EVA_VANILLA -> print_string "VANILLA";
  | EVA_ALGO t -> debugpp_lterm t;;


let rec dump_symbollist sl = 
    (iter_symbollist (fun s -> dump_symbol s; print_newline ()) sl)
  

