(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

open Lexing;;
open Format;;
open Arg;;

open Trace;;
open Error;;
open Location;;
open Abstract;;
open Symbols;;
open Evalex;;
open Evaparse;;
open Translator;;
open Alias;;
open Types;;
open Evapp;;
open Debugpp;;

type output_format =
    OUT_NONE
  | OUT_EVA
  | OUT_DEBUG;;

let sl = ref empty_symbol_list;;

let output = ref OUT_NONE;;

let usage () =
  print_string "Usage: evatrans [options] file.eva"; 
  print_newline ();
  print_string "options:";
  print_newline ();
  open_box 2;
  print_string "-eva : LAEVA output format";
  print_newline ();
  print_string "-debug : debugging output format";
  print_newline ();
  print_string "-trace : output debug info";
  print_newline ();
  print_string "-help : this message";
  print_newline ();
  close_box ();
  exit 1;;

let parse_option args i =
  let arg = args.(i) in
    if (arg = "-eva")
    then  output := OUT_EVA
    else if (arg = "-debug")
    then  output := OUT_DEBUG
    else if (arg = "-trace")
    then Trace.trace_on ()
    else if (arg = "-help")
    then usage ()
    else if ((String.get arg 0) = '-')
    then begin (error "unknown option %s\n" arg); usage () end
    else Evalex.set_filename arg;;

let _ =
  try
    let args = Sys.argv in
      begin
        (* commandline must have at least 1 param *)
        if ( (Array.length args) < 2 )
        then usage ();
        (* parse command line arguments *)
        for i = 1 to ((Array.length args)-1)
        do
          parse_option args i
        done; 
        (* no filename given *)
        if ( (Evalex.get_filename ()) = "" )
        then usage ()
        (* the filename must exist *)
        else if ( not (Sys.file_exists (Evalex.get_filename ())) )
        then 
          begin 
            (error "%s file not found\n" (Evalex.get_filename ())); 
            exit 3; 
          end;
        (* parse *)
        trace "open and parse file %s\n" (Evalex.get_filename ());       
        let ch = (open_in (Evalex.get_filename ())) in
        let spec = ref (Evaparse.spec Evalex.evatoken (Lexing.from_channel ch)) in
          (* check and pretty print *)             
          begin
            close_in ch;
            trace "extract symbol list:\n";
            sl := (symbollist_of_spec !spec);
            (* add the intruder symbol *)
            (* sl := (append_symbol intruder_symbol !sl); *)
            (* add the coercion symbols *)
            (* sl := (append_symbollist  coercion_symbols !sl); *)
            trace "BEGIN dump symbol list (before translation)\n\n";
            if (Trace.is_trace_mode ()) then dump_symbollist !sl;
            trace "END   dump symbol list\n";

            (* post parsing conversions *)
            trace "convert spec\n";
            spec := (convert_spec !sl !spec);

            (* type checking and addition of needed type coercions functions *)
            trace "type coerce\n";
            spec := coerce_spec !sl !spec;

            trace "BEGIN Dump spec (after convert, coerce, before extra check):\n\n"; 
            if (Trace.is_trace_mode ()) then debugpp_spec !spec;
            trace "END   Dump spec (after convert, coerce, before extra check)\n\n"; 

            (* check extra restrictions in spec (see EVA grammar) *)
            trace "check extra restrictions in spec\n";
            spec := (check_spec !sl !spec);

            trace "unalias\n";
            spec := unalias_spec !sl !spec; 

            trace "translation\n";
            trace "BEGIN Dump spec (before translation):\n\n"; 
            if (Trace.is_trace_mode ()) then debugpp_spec !spec;
            trace "END   Dump spec (before translation)\n\n"; 

            trace "translation\n";
            spec := translate !sl !spec;            

            (* recompute the symbol list (there can be additional private variables) *)
            (* USELESS for translator *)
            (* sl := (symbollist_of_spec !spec); *)
            trace "BEGIN dump symbol list (after translation)\n\n";
            if (Trace.is_trace_mode ()) then dump_symbollist !sl;
            trace "END   dump symbol list\n"; 

            trace "BEGIN Dump spec (after translation):\n\n"; 
            if (Trace.is_trace_mode ()) then debugpp_spec !spec;
            trace "END   Dump spec (after translation)\n\n"; 

            trace "Pretty print (after translation):\n"; 
            (match !output with
                 OUT_NONE -> print_string "Compilation successful, exit...\n"
               | OUT_EVA -> evapp_spec !spec
               | OUT_DEBUG -> debugpp_spec !spec );
          end
      end
  with
      Sys_error msg -> 
        error "\n%s\n" msg
    | Evalex.Invalid_char (c, loc) -> 
        error "\n%s:\nInvalid character: %s\n" (Location.string_of_location loc) c
    | Evalex.Eof_in_comment -> error "Unexpected end of file in comment\n"
    | Parsing.Parse_error -> 
        error "\n%s:\nSyntax error\n" (Evalex.string_of_location ())
    | Translator_error (msg, loc) ->
        error "\n%s:\nError: %s\n" (Location.string_of_location loc) msg
    | Alias_error (msg, loc) ->
        error "\n%s:\nUnalias error: %s\n" (Location.string_of_location loc) msg
    | Type_error (msg, loc) ->
        error "\n%s:\nType error: %s\n" (Location.string_of_location loc) msg
      (* TBC: error in this module reported as internal *)
    | Symbols.Symbols_error msg -> 
        error "\nINTERNAL: ERROR: (Symbols) %s (please report!)\n" msg
    | Trace.Internal_error msg -> 
        error "\nINTERNAL ERROR: %s (please report!)\n" msg;;
(*     | (Translator_error msg) -> error "Error %s\n" msg;; *)
