{

(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)


open String;;
open Format;;
open Lexing;;

open Trace;;
open Location;;
open Evaparse;;

exception Eof_in_comment;;
exception Invalid_char of string * location;;

(*** Locations ***)

(* type representing the line number in a file. *)
type file_info = 
{ 
  (* name of the file currently parsed *)
  filename : string;
  (* current lne number *)
  line_number : int;
  (* position of the first char of the current line *)
  first_char : int;
}

let current_info =
  ref { line_number = 1;
        first_char = 0;
        filename = "" };;

let get_filename () =
  !current_info.filename;;

let set_filename s =
  current_info := { !current_info with filename = s };;

let last_loc = ref None;;

let bumpcr lexbuf =
  begin
    trace "Evalex.bumpcr: (line %i, fst char %i)\n" 
      !current_info.line_number
      (lexeme_end lexbuf);
    current_info :=
    { !current_info with 
        line_number = !current_info.line_number + 1;
        first_char = (lexeme_end lexbuf);
    }
  end;;

(* returns the location of the current lexeme in the given lexbuf *)
let bump lexbuf =
  let ln = !current_info.line_number
  and sc = (lexeme_start lexbuf) - !current_info.first_char 
  and ec = (lexeme_end lexbuf) - !current_info.first_char - 1 in
    trace "Evalex.bump: %s (%i, %i-%i)\n" (lexeme lexbuf) ln sc ec;
    let loc =
      {
        file = !current_info.filename;
        start_line = ln;
        end_line = ln;
        start_char = sc;
        end_char = ec;
      }
    in 
      begin
        last_loc := Some loc;
        loc
      end;;
  

(* print the current location on the standard output *)
let print_location () =
  match !last_loc with 
    | None -> 
        begin
          print_string "begining of";
          print_space ();
          print_string (!current_info.filename);
        end
    | Some loc -> Location.print_location loc;;


(* returns a formatted string describing the current location *)
let string_of_location () =
  match !last_loc with 
    | None -> "begining of "^(!current_info.filename)
    | Some loc -> Location.string_of_location loc;;
}

let space = [' ''\t']+
let digit = ['0'-'9']
let int = digit+
(* let fmt = [' ' '\t' '\n' '\r' '\f'] *)
let letter = ['A'-'Z''a'-'z']
let alphanum = (letter | digit | '_')
let id = letter(alphanum)*

rule evatoken = parse 
  (* ignore spaces *)
    space+	{ evatoken lexbuf }
  (* single line of comment *)
  | "//"[^'\n']*'\n' { bumpcr lexbuf; evatoken lexbuf } 
  (* begin comment *)
  | "/*"        { comment lexbuf }
  | '\n'        { trace "newline\n"; bumpcr lexbuf; evatoken lexbuf }
  | ','         { KW_COMMA (bump lexbuf) }
  | ';'         { KW_SEMI (bump lexbuf) }
  | ':'         { KW_COL (bump lexbuf) }
  | '='         { KW_EQ (bump lexbuf) }
  | '"'         { KW_QUOTE (bump lexbuf) }
  | '['         { KW_LSQ (bump lexbuf) }
  | ']'         { KW_RSQ (bump lexbuf) }
  | '('         { KW_LPAR (bump lexbuf) }
  | ')'         { KW_RPAR (bump lexbuf) }
  | '{'         { KW_LBR (bump lexbuf) }
  | '}'         { KW_RBR (bump lexbuf) }
  | '<'         { KW_LANG (bump lexbuf) }
  | '>'         { KW_RANG (bump lexbuf) }
  | '^'         { KW_UP (bump lexbuf) }
  | '*'         { KW_STAR (bump lexbuf) }
  | '_'         { KW_UNDER (bump lexbuf) }
  | '%'         { KW_PCENT (bump lexbuf) }
  | '@'         { KW_AT (bump lexbuf) }
  | "->"	{ KW_TO (bump lexbuf) }
  | "=>"	{ KW_IMPLIES (bump lexbuf) }
  | "=="	{ KW_COMP (bump lexbuf) }
  | "!="	{ KW_NEQ (bump lexbuf) }	
  | ":="	{ KW_ASSIGN (bump lexbuf) }	
  | "asym_algo" { KW_ASYM_ALGO (bump lexbuf) }
  | "alias"     { KW_ALIAS (bump lexbuf) }
  | "assume"	{ KW_ASSUME (bump lexbuf) }
  | "axiom"	{ KW_AXIOM (bump lexbuf) }
  | "basetype"  { KW_BASETYPE (bump lexbuf) }
  | "case"	{ KW_CASE (bump lexbuf) }
  | "claim"	{ KW_CLAIM (bump lexbuf) }
  | "constant"  { KW_CONSTANT (bump lexbuf) }
  | "everybody" { KW_EVERYBODY (bump lexbuf) }
  | "false"     { KW_FALSE (bump lexbuf) }
  | "fresh"	{ KW_FRESH (bump lexbuf) }
  | "hash"	{ KW_HASH (bump lexbuf) }
  | "honest"	{ KW_HONEST (bump lexbuf) }
  | "keypair"   { KW_KEYPAIR (bump lexbuf) }
  | "knows"	{ KW_KNOWS (bump lexbuf) }
  | "number"    { KW_NUMBER (bump lexbuf) }
  | "parameter" { KW_PARAMETER (bump lexbuf) }
  | "principal" { KW_PRINCIPAL (bump lexbuf) }
  | "secret"    { KW_SECRET (bump lexbuf) }
  | "session"   { KW_SESSION (bump lexbuf) }
  | "sym_algo"  { KW_SYM_ALGO (bump lexbuf) }
  | "switch"	{ KW_SWITCH (bump lexbuf) }
  | "value"	{ KW_VALUE (bump lexbuf) }
  | id('.')	{ LABEL (lexeme lexbuf , bump lexbuf) }
  | int('.')	{ LABEL (lexeme lexbuf, bump lexbuf) }
  | id	        { ID (lexeme lexbuf, bump lexbuf) }
  | int 	{ INT (int_of_string(lexeme lexbuf), bump lexbuf) }
  | eof         { EOF }
  | _           { raise (Invalid_char (lexeme lexbuf, bump lexbuf) ) } 
and comment = parse 
    [^ '*''\n']*       { comment lexbuf }
  | '*'+[^'*''/''\n']* { comment lexbuf }
  | '\n'               { bumpcr lexbuf; comment lexbuf }
  | '*'+'/'            { evatoken lexbuf }
  | eof                { raise Eof_in_comment }
