%{

(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(** Parsing of a protocol specification in LAEVA concrete syntax. *)

(** the protocol messages and instruction presented
  as one unique block (list) for every principal
  (presentation a la BAN or Clark-Jacob).
  
  The others syntactic restrictions verified after parsing are marked with - *)

  open List;;

  open Trace;;
  open Evalex;;
  open Location;;
  open Abstract;;

  (** the official name of the intruder *)
  let eva_INTRUDER_NAME = "I";;
  
  (** parse exception *)
  exception Evaparse_error of string * location;;

  (** @return an eva_id from the given token ID *)
  let make_id (str, loc) =
    if (str = "" )
    then (raise (Internal_error 
                   ((Location.string_of_location loc)^
                    "empty identifier")))  (* EVA_ID_VOID *)
    else (EVA_ID str);;

  (** @return an located id (eva_lid) from the given token ID *)
  let make_lid (str, loc) =
    { 
      lid_loc = loc;
      lid_id = make_id (str, loc)
    };;

  (** @return an eva_term from the given token ID *)
  let make_termid (str, loc) =
    if (str = "" )
    then (raise (Internal_error 
                   ((Location.string_of_location loc)^
                    "empty identifier")))  (* EVA_ID_VOID *)
    else (EVA_TERM_ID str);;


  (** @return an located term from the given token ID *)
  let make_ltermid (str, loc) =
    { 
      lte_loc = loc;
      lte_term = make_termid (str, loc)
    };;
 

  (** @return an eva_value from the given token ID *)
  let make_value (str, loc) =
    match str with
        "" -> raise (Internal_error 
                       ((Location.string_of_location loc)^
                        "empty value identifier"))  (* EVA_ID_VOID *)
   (* | ""  -> EVA_VALUE_VOID *)
      | "I" -> EVA_INTRUDER
      | _   -> EVA_VALUE str;;
 

  (** @return an located value (eva_lvalue) from the given token ID *)
  let make_lvalue (str, loc) =
    {
      lva_loc = loc;
      lva_value = make_value (str, loc)
    };;

  
%}

%token <string * Location.location> ID
%token <string * Location.location> LABEL
%token <int * Location.location> INT
%token EOF
%token <Location.location> KW_COMMA
%token <Location.location> KW_SEMI 
%token <Location.location> KW_COL 
%token <Location.location> KW_QUOTE
%token <Location.location> KW_EQ 
%token <Location.location> 
%token <Location.location> KW_LSQ 
%token <Location.location> KW_RSQ 
%token <Location.location> KW_LPAR 
%token <Location.location> KW_RPAR 
%token <Location.location> KW_LBR 
%token <Location.location> KW_RBR 
%token <Location.location> KW_LANG 
%token <Location.location> KW_RANG 
%token <Location.location> KW_UP 
%token <Location.location> KW_STAR 
%token <Location.location> KW_UNDER 
%token <Location.location> KW_PCENT 
%token <Location.location> KW_AT 
%token <Location.location> KW_TO 
%token <Location.location> KW_IMPLIES 
%token <Location.location> KW_EQ 
%token <Location.location> KW_NEQ 
%token <Location.location> KW_COMP
%token <Location.location> KW_ASSIGN
%token <Location.location> KW_ALIAS 
%token <Location.location> KW_ASSUME 
%token <Location.location> KW_AXIOM 
%token <Location.location> KW_BASETYPE 
%token <Location.location> KW_CASE 
%token <Location.location> KW_CLAIM 
%token <Location.location> KW_CONSTANT 
%token <Location.location> KW_EVERYBODY 
%token <Location.location> KW_FALSE 
%token <Location.location> KW_FRESH
%token <Location.location> KW_HASH
%token <Location.location> KW_HONEST 
%token <Location.location> KW_KEYPAIR
%token <Location.location> KW_KNOWS 
%token <Location.location> KW_PARAMETER
%token <Location.location> KW_PRINCIPAL 
%token <Location.location> KW_NUMBER 
%token <Location.location> KW_ASYM_ALGO 
%token <Location.location> KW_SYM_ALGO
%token <Location.location> KW_SECRET 
%token <Location.location> KW_SESSION 
%token <Location.location> KW_SWITCH 
%token <Location.location> KW_VALUE

%start spec

%type <Abstract.eva_spec> spec 

%nonassoc KW_PCENT
%right KW_COMMA
%right KW_SEMI

%%


/** Specification */

/*  An EVA specification, contains in this order: 
 * a label (the protocol name)
 * the declaration of the specification identifiers, aliases, base types, principals knowledge, axioms,
 * a presentation a la BAN of the protocol messages,
 * the declaration of the identifiers used to construct the sessions values,
 * the declaration of sessions (principals initial state)
 * the formulas: {assume} formulas to define the initial state of the environment and {claim} formulas to be verified
 */
spec : 
 ID declarations block value_declarations sessions statements EOF
{ 
  { 
    sp_label = make_lid $1;
    sp_declarations = (List.rev $2);
    sp_protocol = { lpr_loc = (snd $3); lpr_protocol = EVA_MP (fst $3) };
    sp_valuedcl = (List.rev $4);
    sp_sessions = (List.rev $5);
    sp_statements = (List.rev $6);
  }
}
;


/** Type identifiers */

type_id :
  KW_PRINCIPAL
{
  { 
    lty_loc = $1;
    lty_type = EVA_PRINCIPAL
  }
}
| KW_NUMBER
{
  { 
    lty_loc = $1;
    lty_type = EVA_NUMBER
  }
}
| KW_ASYM_ALGO 
{   
  { 
    lty_loc = $1;
    lty_type = EVA_AALGO
  }
}
| KW_SYM_ALGO
{
  { 
    lty_loc = $1;
    lty_type = EVA_SALGO
  }
}
| ID
{
  { 
    lty_loc = (snd $1);
    (* no VOID, no BASETYPE *)
    lty_type = EVA_USERTYPE (fst $1)
  }
}
;

/** Declarations */

/* reversed list of declarations. 
 * - every symbol declared here must not be declared twice 
 *   in the sections {declarations} and {value_declarations}
 */
declarations : 
  declarations declaration 
{ 
  $2::$1
}
| 
{ 
  []
}
;

/* declaration */
declaration : 
/* declaration of first order or functional identifiers */
  typing_declaration
{
  $1
}
/* declaration of global alias 
 *  - the {ID} must be distinct from 'I' (intruder's name)
 */
| KW_ALIAS ID KW_EQ atomic_term_or_ciphertext 
{
  {
    ldc_loc = spawn $1 $4.lte_loc;
    ldc_dcl = EVA_DCL_ALIAS (make_lid $2, $4)
  }
}
/* declaration of local alias 
 *  - the {ID} must be distinct from 'I' (intruder's name)
 */
| ids KW_COL KW_ALIAS ID KW_EQ atomic_term_or_ciphertext
{
  {
    (* hd: ids is a non empty reversed list *)
    ldc_loc = spawn (hd $1).lid_loc $6.lte_loc;
    ldc_dcl = EVA_DCL_LOCALALIAS ((List.rev $1), make_lid $4, $6)
  }
}
/* declaration of base type 
 *  - the {ID} must be distinct from 'I' (intruder's name)
 */
| KW_BASETYPE ID
{
  {
    ldc_loc = spawn $1 (snd $2);
    ldc_dcl = EVA_DCL_BASETYPE (fst $2)
  }
}
/* initial knowledge common to every principal */
| KW_EVERYBODY KW_KNOWS tuple
{
   {
     ldc_loc = spawn $1 (hd (List.rev $3)).lte_loc;
     ldc_dcl = EVA_DCL_EVERY (List.rev $3)
   }
}
/* initial knowledge of principal {ID} 
 * - {ID} must be declared as a first order identifier of type {principal}
 */
| ID KW_KNOWS tuple
{
  {
    (* hd: tuple is non empty reversed list *)
    ldc_loc = spawn (snd $1) (hd $3).lte_loc; 
    (** TODO suppr make_lknowledge (just terms) **)
    ldc_dcl = EVA_DCL_KNOW (make_lid $1, (List.rev $3)) 
  }
}
/* global equational axiom 
 * - every leave in the {term}s must be an variable of the
 *   {optional_quantification}
 * - both {term}s (members of the atom) must have the same type
 */
| KW_AXIOM atomic_term_or_ciphertext KW_EQ atomic_term_or_ciphertext optional_quantification
{
  {
    ldc_loc = spawn $1 (
      match $5 with 
          [] -> $4.lte_loc
        | (i, {lty_loc = _; lty_type = EVA_VOID})::l -> i.lid_loc 
        | (_, t)::l -> t.lty_loc );
    ldc_dcl = EVA_DCL_AXIOM ($2, $4, EVA_VOID, $5)
  }
}
;

/* quantification for the formulas ({assume}) and {axiom} 
 * - the symbols are bounded, they can be declared elsewhere
 *   in a {declaration} or {value_declaration}
 * - every (variable) symbol must occur at most once in quantification
 */
optional_quantification : 
  KW_LSQ quantified_vars KW_RSQ
{ 
  (List.rev $2)
}
| 
{ 
  []
}
;

/* non-empty reversed list of universally quantified variables */
quantified_vars : 
  quantified_var
{ 
  [ $1 ]
}
| quantified_vars KW_COMMA quantified_var
{ 
  $3::$1
}
;

/* universally quantified variable with optional type 
 * - the default type is {number}
 */
quantified_var : 
  ID
{
  (* the location of the (missing) type is arbitrary set 
   * equal to the location of the ID *)
  (make_lid $1, { lty_loc = (snd $1); lty_type = EVA_NUMBER })
}
| ID KW_COL type_id
{
  (make_lid $1, $3)
}
;

/* type declaration for identifiers */
typing_declaration : 
/* declaration of first order identifiers 
 */
  scope ids_decl KW_COL type_id
{
  {
    ldc_loc = spawn $1.lsc_loc $4.lty_loc;
    ldc_dcl = EVA_DCL_ID ((List.rev $2), $1, $4)
  }
}
/* declaration of functional identifier 
 *  - the {ID} must be distinct from 'I' (intruder's name)
 */
| scope ID KW_LPAR type_list KW_RPAR KW_COL type_id optional_hash optional_secret
{
  {
    ldc_loc = spawn $1.lsc_loc 
                (if (fst $9)
                 then (snd $9)
                 else 
                   (if (fst $8)
                    then (snd $8)
                    else $7.lty_loc));
    ldc_dcl = EVA_DCL_FUN (make_lid $2, $1, $7, (List.rev $4), (fst $8), (fst $9))
  }
}
/* declaration of a pair of asymmetric keys, both {ID}s
 * can be first order or functional symbols.
 * note: no {hash} keyword, every keypair is assumed to be hash. 
 * - the two {ID}s (public and private keys) must be distinct
 * - the two {ID}s must be distinct of 'I' (intruder's name)
 */
| scope KW_KEYPAIR optional_encryption_algorithm 
     ID KW_COMMA ID optional_type_list KW_COL type_id
{
  {
    ldc_loc = spawn $1.lsc_loc $9.lty_loc;
    ldc_dcl = EVA_DCL_KEYPAIR ($3, make_lid $4, make_lid $6, $1, $9, (List.rev $7))
  }
}
;

/* non empty reversed list of identifiers */
ids : 
  ID
{ 
  [ (make_lid $1) ] 
}
| ids KW_COMMA ID
{ 
  (make_lid $3)::$1
}
;



/* non-empty reversed list of identifiers for declaration */
ids_decl : 
  id_decl
{ 
  [ $1 ]
}
| ids_decl KW_COMMA id_decl
{ 
  $3::$1
}
;

/* first order identifier declared with freshness */
id_decl :
/*  fresh first order identifier:
 *  - the {ID} must be distinct from 'I' 
 *     (intruder's name)
 *  - the {ID} must be declared with the {scope} {parameter} 
 */
  KW_FRESH ID
{
  (make_lid $2, true)
}
/*  non fresh first order identifier:
 *  - the {ID} must be distinct from 'I' (intruder's name) 
 */
| ID
{
  (make_lid $1, false)
}
;


/* reversed list of types */
type_list : 
  non_empty_type_list 
{ 
  $1
}
| 
{ 
  []
}
;

/* non-empty reversed list of types */
non_empty_type_list : 
  type_id
{ 
  [ $1 ]
}
| non_empty_type_list KW_COMMA type_id 
{ 
  $3::$1 
}
;


/* list of types */
optional_type_list : 
  KW_LPAR type_list KW_RPAR
{ 
  $2
}
| 
{ 
  [] 
}
;

/* optional qualifier of one-way functions */
optional_hash : 
  KW_HASH
{ 
  (* the location is used to evaluate the end of the location of the fun declaration *)
  (true, $1)
}
| 
{ 
  (false, default_location)
}
;

/* optional qualifier of secret functions */
optional_secret : 
  KW_SECRET
{ 
  (* the location is used to evaluate the end of the location of the fun declaration *)
  (true, $1)
}
| 
{ 
  (false, default_location)
}
;


/* quality of identifier */
scope : 
  /* global constant */
  KW_CONSTANT
{
  {
    lsc_loc = $1;
    lsc_scope = EVA_CST
  }
}
  /* session parameter */
| KW_PARAMETER
{
  {
    lsc_loc = $1;
    lsc_scope = EVA_PARAM
  }
}
;


/* invocation of an encryption algorithm 
 * - must have an algorithm type
 */
optional_encryption_algorithm : 
/* the {atomic_term} describes a symmetric or asymmetric key algorithm */
  KW_UP atomic_term 
{ 
  (EVA_ALGO $2)
}
/* default symmetric key algorithm */
| 
{ 
  EVA_VANILLA
}
;


/** Protocol messages */

/* protocol or sub-protocol */
block : 
  KW_LBR messages KW_RBR
{ 
  (* we attach the location to fill the corresponding kind of message (sub-protocol) *)
  ((List.rev $2), (spawn $1 $3))
}
;


/* reversed list of instructions */
messages : 
  messages message 
{ 
  $2::$1
}
| 
{ 
  []
}
;

/* message label */
label : 
  LABEL 
{ 
  {
    lla_loc = (snd $1);
    lla_label = EVA_LABEL (fst $1)
  }
}
;

/* protocol instruction
 * every {term} and {atomic_term_or_ciphertext} in the instruction:
 * - can contain (at leaves positions) first order symbols, 
 *   and functional symbols, and key symbols,
 *   declared either with {scope} {constant} or  {parameter}
 * - can contain declared alias symbols
 * - must not contain located variables
 * - must not contain identifiers declared as values
 * - can contain '\%' (Lowe's notation)
 */
message : 
/* sending/receiving message 
 * - the two {ID}'s (sender and receiver) must be declared as first order identifiers with type {principal}
 */
  label ID KW_TO ID KW_COL term
{
  {
    lin_loc = spawn $1.lla_loc $6.lte_loc;
    lin_instr = EVA_MSG ($1, make_lid $2, make_lid $4, $6)
  }
}
/* equality test 
 * - the first {ID} must be declared as first order identifier with type {principal}
 * - the second {ID} must not be known to the first {ID} (call it {r}),
 *   i.e. it must not have be assigned in the local state of {r} at this point.
 */
| ID KW_COL ID KW_COMP atomic_term_or_ciphertext
{
  {
    lin_loc = spawn (snd $1) $5.lte_loc;
    lin_instr = EVA_COMP (make_lid $1, make_lid $3, $5)
  }
}
/* local assignment 
 * - the first {ID} must be declared as first order identifier with type {principal}
 * - the second {ID} must not be known to the first {ID} (call it {r}),
 *   i.e. it must not have be assigned in the local state of {r} at this point
 *   (reassignements are not allowed).
 */
| ID KW_COL ID KW_ASSIGN atomic_term_or_ciphertext
{
  {
    lin_loc = spawn (snd $1) $5.lte_loc;
    lin_instr =  EVA_ASSIGN (make_lid $1, make_lid $3, $5)
  }
}
/* sub-protocol */
| block
{
   {
     lin_loc = (snd $1);
     lin_instr = EVA_BLOCK (fst $1)
   }
}
/* conditional */
| KW_SWITCH atomic_term_or_ciphertext KW_LBR cases KW_RBR
{
   {
     lin_loc = spawn $1 $5;
     lin_instr = EVA_SWITCH ($2, (List.rev $4))
   }
}
;

/* reversed list of cases */
cases : 
  cases case 
{ 
  $2::$1
}
| 
{ 
  []
}
;

/* conditional branching */
case : 
  KW_CASE atomic_term_or_ciphertext KW_COL block
{
  {
    case_case = $2;
    case_block = (fst $4);
    case_loc = spawn $1 (snd $4)
  }
}
;

/** Terms */

/* term */
term : 
  tuple 
{ 
  trace "EVA_CONS\n";
  {
    (* hd: tuple is not empty *)
    lte_loc = spawn (hd $1).lte_loc (hd (List.rev $1)).lte_loc;
    lte_term = EVA_CONS (List.rev $1)
  }
}
;

/* non-empty reversed list of terms */
tuple : 
  atomic_term_or_ciphertext
{ 
  [ $1 ]
}
| tuple KW_COMMA atomic_term_or_ciphertext 
{ 
  $3::$1
}
;

/* non-empty parenthesized reversed list of terms */
non_empty_term_list : 
  KW_LPAR tuple KW_RPAR 
{ 
  (* we attach the location to evaluate location of atomic term *)
  ($2, spawn $1 $3)
}
;

/* parenthesized reversed list of terms */
term_list : 
  KW_LPAR KW_RPAR 
{ 
  (* we attach the location to evaluate location of atomic term *)
  ([], spawn $1 $2)
}
| non_empty_term_list 
{ 
  $1 
}
;

/* atomic term */
atomic_term_or_ciphertext : 
  atomic_term 
{ 
  $1
}
/* ciphertext  */
| KW_LBR term KW_RBR KW_UNDER atomic_term optional_encryption_algorithm
{
  trace "EVA_CRYPT\n";
  {
    lte_loc = spawn $1 (match $6 with
                            (EVA_ALGO t) -> t.lte_loc
                          | EVA_VANILLA -> $5.lte_loc);
    lte_term = EVA_CRYPT ($6, $5, $2)
  }
}
/* signature */
| KW_LSQ term KW_RSQ KW_UNDER atomic_term optional_encryption_algorithm
{
  trace "EVA_SIGN\n";
  {
    lte_loc = spawn $1 (match $6 with
                            (EVA_ALGO t) -> t.lte_loc
                          | EVA_VANILLA -> $5.lte_loc);
    lte_term = EVA_SIGN ($6, $5, $2)
  }  
}
/* schizo-notation a la Lowe 
 * - '\%' is not allowed inside the terms
 */
| atomic_term_or_ciphertext KW_PCENT atomic_term_or_ciphertext
{
  trace "EVA_PCENT\n";
  {
    lte_loc = spawn $1.lte_loc $3.lte_loc;
    lte_term = EVA_PCENT ($1, $3)
  }
}
/* located term
 * - for formulas {claim} only
 * - the second {ID} must be declared as first order identifier with type {principal}
 * - the first {ID} must be declared as first order or functional or key identifier
 * - the type of the first {ID} only is considered in type evaluation
 */
| ID KW_AT ID
{
  trace "EVA_LOCATED %s@%s\n" (fst $1) (fst $3);
  {
    lte_loc = spawn (snd $1) (snd $3);
    lte_term = EVA_LOCATED (make_lid $1, make_lid $3)
  }
}
;

/* atomic term */
atomic_term : 
/* function call
 * - {ID} must be declared as functional symbol or functional member of {keypair}
 * - the length of the {term_list} must match the declared signature of {ID}
 *   (use parenthezing tuples to apply function to more arguments).
 * - the types of terms in the {term_list} must match the declared signature of {ID}
 */ 
  ID term_list
{
  trace "EVA_APP %s(...)\n" (fst $1);
  {
    lte_loc = spawn (snd $1) (snd $2);
    lte_term = EVA_APP (make_lid $1, List.rev (fst $2))
  }
}
/* identifier
 * - {ID} must be declared as first order symbol or value
 */
| ID
{ 
  trace "EVA_TERM_ID %s\n" (fst $1);
  make_ltermid $1
}
/* singleton list for optional parenthezing of tuples */
| non_empty_term_list
{
  (* add the size of () in location *)
  {
    lte_loc = snd $1;
    lte_term = EVA_CONS (List.rev (fst $1))
  }
}
;

/** Declaration of values (domain constructors) */

/* reversed list of value declarations for instantiating the sessions 
 * - the value identifiers must not be declared otherwise
 */
value_declarations :
  value_declarations value_declaration
{ 
  $2::$1
}
|  
{
  []
}
;

/* non-empty reversed list of values identifiers */
values : 
  ID
{ 
  [ (make_lvalue $1) ] 
}
| values KW_COMMA ID
{ 
  (make_lvalue $3)::$1
}
;

/* declaration of the symbols used to construct the terms of interpretation domain,
 * they can be used below to instanciate the sessions declared 
 */
value_declaration : 
/* declaration of nullary value symbols */
  KW_VALUE values KW_COL type_id
{
  {
    ldc_loc = spawn $1 ($4.lty_loc);
    ldc_dcl = EVA_DCL_VALUE ((List.rev $2), $4)
  }
}
/* declaration of constructor for values
 * - it must be a declaration of functional or key identifier (not first order id)
 * - the {scope} of declaration must be {constant}
 */
| typing_declaration
{
  $1
}
;

/** Sessions */

/* reversed list of parallel sessions
 * (initial state of the principals)
 */
sessions :
  sessions session 
{ 
  $2::$1
}
| 
{ 
  []
}
;

/* Session definition  */
session : 
/* arbitrary number of sessions over arbitrary domains */
  KW_SESSION KW_STAR 
{ 
  {
   lse_loc = spawn $1 $2;
   lse_session = EVA_BANG
  }
}
/* arbitrary number of sessions with domain satisfying a constraint */
| KW_SESSION KW_STAR KW_LSQ session_constraints KW_RSQ
{
  {
   lse_loc = spawn $1 $5;
   lse_session = EVA_CONSTRAINED_SESSIONS (List.rev $4)
  }
}
/* arbitrary number of copies of the given session */
| KW_SESSION KW_STAR KW_LPAR session_assignments KW_RPAR
{
 {
  lse_loc = spawn $1 $5;
  lse_session = EVA_SESSIONS (List.rev $4)
 }
}
/* fixed session */
| KW_SESSION KW_LPAR session_assignments KW_RPAR
{
  {
    lse_loc = spawn $1 $4;
    (* the location of the missing label is arbitrary set
     * equal to the location of the opening parenthesis *)
    lse_session = EVA_SESSION 
                    ({ lla_loc = $2; lla_label = EVA_NOLABEL }, (List.rev $3))
  } 
}
/* fixed session with label */
| label KW_SESSION KW_LPAR session_assignments KW_RPAR
{
  {
    lse_loc = spawn $1.lla_loc $5;
    lse_session = EVA_SESSION ($1, (List.rev $4))
  }
}
;

/* non empty reversed list of constraints on domains for instantiating sessions
 */
session_constraints : 
  session_constraint
{
  [ $1 ]
}
| session_constraints KW_SEMI session_constraint
{
  $3::$1
}
;

/* constraint on domains for session instances */
session_constraint : 
/* equality constaint 
 * - both {ID}s must be declared with {scope} {parameter}
 */
  ID KW_COMP ID
{
  {
    lco_loc = spawn (snd $1) (snd $3);
    lco_constraint = EVA_EQ_CONSTRAINT (make_lid $1, make_lid $3)
  }
}
/* disequality constaint 
 * - both {ID}s must be declared with {scope} {parameter}
 */
| ID KW_NEQ ID
{
  {
    lco_loc = spawn (snd $1) (snd $3);
    lco_constraint = EVA_NEQ_CONSTRAINT (make_lid $1, make_lid $3)
  }
}
/* membership constraint:
 * the {parameter} {ID} belongs to a finite set of values.
 * - {ID} must be declared with {scope} {parameter}
 * - the {tuple} is a sequence of terms can contain only at leaves positions:
 *   values symbols, first order, functional and key symbols declared with {scope} {constant}.
 */
| ID KW_COL KW_LBR tuple KW_RBR
{
  {
    lco_loc = spawn (snd $1) $5;
    lco_constraint = EVA_MEMBERSHIP_CONSTRAINT (make_lid $1, (List.rev $4))
  }
}
/* domain constraint:
 * the first {ID} is in the model of the {predicate}.
 * - {ID} must be declared with {scope} {parameter}
 */
| ID KW_COL predicate
{
  {
    lco_loc = spawn (snd $1) (snd $3);
    lco_constraint = EVA_DOMAIN_CONSTRAINT (make_lid $1, (fst $3))
  }
}
;

/* predicate symbol for membership constraints in session declarations */
predicate :
/* reserved unary predicate "secret" */
  KW_SECRET
{
  ("secret", $1)
}
/* reserved unary predicate "honest" */
| KW_HONEST
{
  ("honest", $1)
}
/* user predicate 
 * - {ID}, the name of the predicate,
 *   must not be a declared identifer.
 *   it can only occur as a predicate symbol in 
 *   an {atom} of a {statement}.
 */
| ID
{
  $1
}
; 

/* non-empty list of assignments for the parameters of a session */
session_assignments : 
  session_assignment
{
  [ $1 ]
}
| session_assignments KW_SEMI session_assignment
{
  $3::$1
}
;

/* instantiation of a parameter for the definition of a session 
 */
session_assignment : 
/*   
 * - the {ID} must have been declared as a parameter
 * - the {atomic_term_or_ciphertext} can contain (at leaves positions) either:
 *   value symbols, 
 *   first order, functional and key symbols declared with {scope} {constant},  
 *   '{I}', the name of the intruder,
 *   or aliases to terms of the above form
 * - the {ID} and {atomic_term_or_ciphertext} must have the same type
 */
  ID KW_EQ atomic_term_or_ciphertext
{
  (make_lid $1, $3)
} 
;

/** Formulas */

/* reversed list of hypotheses and properties */
statements : 
  statements assumption
{ 
  $2::$1
}
| statements claim
{ 
  $2::$1
}
| 
{
  []
}
;


/* hypothesis on the initial state of the environment */
assumption : 
  KW_ASSUME formula optional_quantification 
{
  let (label, tail, head) = $2 in 
    {
      lst_loc = spawn $1 (
        match $3 with
            (* location of the last element (head) of the formula *)
            [] -> head.lat_loc
              (* location of the last quantified variable in the reversed list *)
          | (v, {lty_loc = _; lty_type = EVA_VOID})::l -> v.lid_loc
          | (v, t)::l -> t.lty_loc
      );
      lst_statement = EVA_ASSUME ($3, tail, head)
    }
}
;

/* property to prove */
claim : 
  KW_CLAIM formula
{
  let (label, tail, head) = $2 
  in 
    {
      lst_loc = spawn $1 head.lat_loc;
      lst_statement = EVA_CLAIM (label, tail, head)
    }
}
;

/* Horn clause */
formula : 
  optional_label atom
{
   ($1, [], $2)
}
| optional_label atoms KW_IMPLIES KW_FALSE
{
 (* trick: the location of the last atom (FALSE) is the location of => 
  * (for location of assume) *)
 ($1, (List.rev $2), { lat_loc = $4; lat_atom = EVA_FALSE } )
}
| optional_label atoms KW_IMPLIES atom
{
  ($1, (List.rev $2), $4)
}
;

/* optional formula name */
optional_label : 
  label
{ 
  $1
}
| 
{
  { lla_loc = default_location; lla_label = EVA_NOLABEL }
}
;


/* reversed list of atoms */
atoms:
  non_empty_atom_list
{
  $1
}
|
{
  []
}
;


/* non-empty reversed list of  atoms */
non_empty_atom_list : 
  atom
{
  [ $1 ]
}
| non_empty_atom_list KW_COMMA atom
{
  $3::$1
}
;

/* atomic proposition.
 *
 * Every {term} in the folloing atoms must be such that:
 * - in {assume}: every leave in the {term} must be
 *   a protocol {constant},
 *   a declared {value}, 
 *   or a quantified variable.
 * - in {claim}: every leave in the {term} must be
 *   a protocol {constant}, 
 *   a declared {value}, 
 *   or a located variable
 * - in both: the {term} must not contain '\%' (Lowe's notation)
 */
atom : 
/* term equality */
 atomic_term_or_ciphertext KW_COMP atomic_term_or_ciphertext 
{
 {
  lat_loc = spawn $1.lte_loc $3.lte_loc;
  lat_atom = EVA_ATOM_EQ ($1, $3)
 }
}
/* term disequality */
| atomic_term_or_ciphertext KW_NEQ atomic_term_or_ciphertext 
{
 {
  lat_loc = spawn $1.lte_loc $3.lte_loc;
  lat_atom = EVA_ATOM_NEQ ($1, $3)
 }
}
/* the intruder ignores {term} */
| KW_SECRET KW_LPAR atomic_term_or_ciphertext KW_RPAR
{
 {
  lat_loc = spawn $1 $4;
  lat_atom = EVA_ATOM_SECRET $3
 }
}
/* the {term} is the identifier of an honest principal.
 */
| KW_HONEST KW_LPAR atomic_term_or_ciphertext KW_RPAR
{
 {
  lat_loc = spawn $1 $4;
  lat_atom = EVA_ATOM_HONEST $3
 }
}
/* user defined predicate
 * - {ID}, the name of the predicate,
 *   must not be an identifer declared elsewhere
 * - there must be exactly one argument.
 */
| ID KW_LPAR tuple KW_RPAR
{
  let term = 
    match  $3 with
        [t] -> t
      | _ -> 
          (* Evaparse_error ("non unary atom", (spawn (snd $1) $4)) *)
          raise Parsing.Parse_error
  in 
    {
      lat_loc = spawn (snd $1) $4;
      lat_atom = EVA_ATOM_USER (make_lid $1, term)
    }
}
;

%%
