(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(*
 * Pretty print specification in abstract syntax into
 * a EVA concrete syntax, as defined in evaparse
 * with additional constructions for definition
 * of separate processes
 *)

open Format;;

open Abstract;;
open Types;;
open Pp;;


(***************************************************************
 * Separators for pp_list
 ***************************************************************)

(* standard comma separation to use for pp_list *)
let comma_sep () = 
  print_string ","; print_space();;

(* standard semicolumn separation to use for pp_list *)
let semicol_sep () = 
  print_string ";"; print_space();;

(* standard null separation to use for pp_list *)
let null_sep () = ();;

(* standard new line separation to use for pp_list *)
let newline_sep () = 
  print_newline ();;



(***************************************************************
 * Pretty printing functions
 ***************************************************************)

let rec evapp_label l =
  begin
    open_box 0;
    (match l with
         EVA_NOLABEL -> print_string ""
       | (EVA_LABEL s) -> print_string s);
    close_box ();
  end
and evapp_llabel { lla_loc = _; lla_label = l } =
  evapp_label l;;

let rec evapp_type t =
  begin
    open_box 0;
    (match t with
         EVA_VOID ->
           print_string "VOID"
       | EVA_PRINCIPAL ->
           print_string "principal"
       | EVA_NUMBER ->
           print_string "number"
       | EVA_AALGO -> 
           print_string "asym_algo"
       | EVA_SALGO -> 
           print_string "sym_algo"
       | EVA_TALGO -> 
           print_string "algo"
       | (EVA_USERTYPE s) -> 
           print_string "\""; print_string s; print_string "\""
       | (EVA_BASETYPE s) -> 
           print_string "\""; print_string s; print_string "\""
       | (EVA_FUN t1) ->
           evapp_type t1; print_string "*number"
       | (EVA_HFUN t1) ->
           evapp_type t1; print_string "*hnumber"
       | (EVA_KEY t1) ->
           evapp_type t1; print_string "*kp"  
    );
    close_box ();
  end
and evapp_ltype { lty_loc = _; lty_type = t } =
  evapp_type t;;

let rec evapp_scope s =
  begin
    open_box 0;
    (match s with
       (*  EVA_SCOPE_VOID -> print_string "" *)
         EVA_CST -> print_string "constant"
       | EVA_PARAM -> print_string "parameter"
       | EVA_PRIVATE -> print_string "private");
    close_box ();
  end
and evapp_lscope { lsc_loc = _; lsc_scope = s } =
  evapp_scope s;;

let rec evapp_id i =
  begin
    open_box 0;
    (match i with
         (* EVA_ID_VOID -> print_string "" *)
         (EVA_ID s) -> print_string s);
    close_box ();
  end
and evapp_lid { lid_loc = _; lid_id = i }  =
  evapp_id i;;

let rec evapp_value v =
  begin
    open_box 0;
    (match v with
       (*  EVA_VALUE_VOID -> print_string "" *)
         EVA_INTRUDER -> print_string "I"
       | (EVA_VALUE s) -> print_string s);
    close_box ();
  end
and evapp_lvalue { lva_loc = _; lva_value = v } =
  evapp_value v;;

let rec evapp_term t =
  begin
    open_box 0;
    (match t with
         (* EVA_TERM_VOID -> print_string "" *)
         EVA_TERM_ID s 
       | EVA_TERM_CST s -> print_string s
       | (EVA_TERM_VALUE v) -> evapp_value v
       | (EVA_TERM_ALIAS (a, ta)) -> print_string a
       | (EVA_LOCATED (i, r)) -> 
           evapp_lid i; 
           print_string "@"; 
           evapp_lid r; 
       | (EVA_QUANTIFIED (i, _)) -> print_string i
         (* we do not print coercion symbols *)
       | EVA_P t
       | EVA_A t
       | EVA_SA t
       | EVA_AA t
       | EVA_U (_, t) -> evapp_lterm t
       | EVA_CONS tl -> pp_list evapp_lterm comma_sep tl
       | EVA_APP (i, etl) -> evapp_app (i, etl)
       | EVA_CRYPT (a, k, t) -> evapp_cipher (a, k, t)
       | EVA_SIGN (a, k, t) -> evapp_cipher (a, k, t) 
       | EVA_PCENT (t1, t2) -> 
           evapp_lterm t1; 
           print_space (); print_string "%"; print_space (); 
           evapp_lterm t2; 
    );
    close_box ();
  end
and evapp_lterm { lte_loc = _; lte_term = t } =
  evapp_term t
(* app_root = i; app_args = etl *)
and evapp_app (i, etl) =
  evapp_lid i; 
  print_string "("; 
  pp_list evapp_lterm comma_sep etl; 
  print_string ")";                 
(* crypt_algo = a; crypt_key = k; crypt_content = t *)
and evapp_cipher (a, k, t) =
   begin 
     print_string "{ "; evapp_lterm t; print_string " }";
     evapp_optional_algo a;
     print_string "_"; evapp_lterm k;
   end
and evapp_algo = function
    EVA_VANILLA -> ()
  | (EVA_ALGO t) -> evapp_lterm t
and evapp_optional_algo = function
    EVA_VANILLA -> ()
  | (EVA_ALGO t) -> 
      begin 
        print_string "^"; 
        evapp_lterm t 
      end;;

let rec evapp_dcl_ids (ibl, s, t) =
  begin
    open_box 0;
    evapp_lscope s;
    print_space ();
    evapp_ids ibl;
    print_string " :";
    print_space ();
    evapp_ltype t;
    close_box ();
  end
and evapp_ids = function
    [] -> ()
  | [ (i, f) ] -> 
      begin 
        evapp_fresh f; 
        evapp_lid i 
      end
  | (i, f)::l -> 
      begin 
        evapp_fresh f; 
        evapp_lid i; 
        print_string ","; 
        print_space(); 
        evapp_ids l; end
and evapp_fresh f =
  if f 
  then begin print_string "fresh"; print_space (); end
  else ();;

let rec evapp_dcl_fun (i, s, t, tl, bh, bs) =
  begin
    open_box 0;
    evapp_lscope s;
    print_space ();
    evapp_lid i;
    print_string "(";
    pp_list evapp_ltype comma_sep tl;
    print_string ") :";
    print_space ();
    evapp_ltype t;
    evapp_hash bh;
    evapp_secret bs;
    close_box ();
  end
and evapp_hash h =
  if h 
  then begin print_string "hash"; print_space (); end
  else ()
and evapp_secret s =
  if s
  then print_string "secret" 
  else ();;

let evapp_dcl_keypair (a, k1, k2, s, t, tl) =
  begin
    open_box 0;
    evapp_lscope s; print_space ();
    print_string "keypair";
    evapp_optional_algo a; print_space ();
    evapp_lid k1; print_string ","; print_space ();
    evapp_lid k2;
    print_space (); print_string "(";
    pp_list evapp_ltype comma_sep tl;
    print_string ") :"; print_space ();
    evapp_ltype t;
    close_box ();
  end;;

(* know_principal = i; know_contents = tl *)
let evapp_dcl_know (i, tl) =
  begin 
    evapp_lid i; print_space (); 
    print_string "knows"; print_space ();
    pp_list evapp_lterm comma_sep tl; 
  end;;

let rec evapp_dcl d =
  begin
    open_box 0;
    (match d with
         EVA_DCL_ID (ibl, s, t) -> evapp_dcl_ids (ibl, s, t)
       | EVA_DCL_FUN (i, s, t, tl, bh, bs) ->  
           evapp_dcl_fun (i, s, t, tl, bh, bs)
       | EVA_DCL_KEYPAIR (a, k1, k2, s, t, tl) -> 
           evapp_dcl_keypair (a, k1, k2, s, t, tl)
       | EVA_DCL_ALIAS (i, t) -> 
           begin 
             print_string "alias"; print_space (); 
             evapp_lid i; print_space (); 
             print_string "="; print_space ();
             evapp_lterm t; 
           end
       | EVA_DCL_LOCALALIAS (il, i, t) -> ()
       | EVA_DCL_BASETYPE s -> 
           begin 
             print_string "basetype"; print_space (); 
             print_string s; 
           end
       | EVA_DCL_KNOW (i, tl) -> evapp_dcl_know (i, tl)
       | EVA_DCL_EVERY tl -> 
           begin 
             print_string "everybody knows"; print_space (); 
             pp_list evapp_lterm comma_sep tl; 
           end
       | EVA_DCL_AXIOM (t1, t2, t, ql) -> evapp_axiom (t1, t2, t, ql)
       | EVA_DCL_VALUE (vl, t) -> evapp_dcl_value (vl, t)
    );
    close_box ();
    print_newline ();
  end
and evapp_ldcl { ldc_loc = _; ldc_dcl = d } =
  evapp_dcl d
(* ax_forall = ql; ax_left = t1; ax_right = t2 *)
and evapp_axiom (t1, t2, _, ql) =
  begin 
    print_string "axiom"; print_space (); 
    evapp_lterm t1; print_space ();
    print_string "="; print_space (); 
    evapp_lterm t2; 
    print_string "[";
    pp_list evapp_optional_quantification comma_sep ql;
    print_string "]"; 
  end
and evapp_optional_quantification (i, t) =
  begin 
    evapp_lid i;
    (if ( t.lty_type <> EVA_VOID )
     then 
       begin 
         print_space (); print_string ":"; print_space (); 
         evapp_ltype t; 
       end);
  end
and evapp_dcl_value (vl, t) =
  begin 
    open_box 0;
    print_string "value"; print_space ();
    pp_list evapp_lvalue comma_sep vl; print_space ();
    print_string ":"; print_space (); 
    evapp_ltype t; 
    close_box ();
    print_newline ();
  end;;

let rec evapp_instr i = 
  begin
    open_box 0;
    (match i with
         EVA_SKIP -> print_space ()
       | EVA_MSG (l, is, ir, t) -> 
           evapp_msg (l, is, ir, t)
       | EVA_ASSIGN (ir, iv, t) -> 
           evapp_assign (ir, iv, t)
       | EVA_COMP (ir, iv, t)
         -> evapp_compare (ir, iv, t)
       | EVA_BLOCK b -> evapp_block b;
       | EVA_SWITCH (t, cl) -> 
           begin 
             print_string "switch"; 
             evapp_lterm t; 
             print_newline(); print_string "{";
             open_box 1; 
             pp_list evapp_case null_sep cl; 
             close_box (); 
             print_string "}" 
           end     
    );
    close_box ();
    print_newline ();
  end
and evapp_linstr { lin_loc = _; lin_instr = i } =
  evapp_instr i;
(* msg_label = l; msg_sender = is; msg_receiver = ir; msg_body = t *)
and evapp_msg (l, is, ir, t) =
  begin 
    evapp_llabel l; print_space (); 
    evapp_lid is; print_space ();
    print_string "->"; print_space (); 
    evapp_lid ir; print_space ();
    print_string ":"; print_space (); 
    evapp_lterm t; 
  end
(* as_role = ir; as_variable = iv; as_value = t *)
and evapp_assign (ir, iv, t) =
  begin 
    evapp_lid ir; print_space (); 
    print_string ":"; print_space ();
    evapp_lid iv; print_space (); 
    print_string ":="; print_space (); 
    evapp_lterm t; 
  end
(* co_role = ir; co_left_variable = iv; co_right_variable = t *)
and evapp_compare (ir, iv, t) =
  begin 
    evapp_lid ir; print_space (); 
    print_string ":"; print_space ();
    evapp_lid iv; print_space (); 
    print_string "=="; print_space (); 
    evapp_lterm t; 
  end      
and evapp_case { case_case = t; case_block = b; case_loc = _ } =
  begin open_box 0;
    print_string "case"; 
    evapp_lterm t; print_space (); 
    print_string ":"; print_newline (); 
    evapp_block b; 
    close_box (); 
  end
and evapp_block b = 
  begin 
    open_box 0; 
    print_string "{"; print_newline ();
    open_box 2; 
    pp_list evapp_linstr null_sep b; 
    close_box (); 
    print_string "}";
    close_box (); 
    print_newline (); 
  end;;

let rec evapp_protocol = function
    (EVA_MP l) -> 
      evapp_block l
  | (EVA_PROCESS pl) -> 
      pp_list evapp_process newline_sep pl
and evapp_lprotocol { lpr_loc = _; lpr_protocol = p } =
  evapp_protocol p
and evapp_process { pr_role = i; pr_privates = il; pr_messages = b } =
  begin 
    open_box 0;
    print_string "process"; print_space (); 
    evapp_lid i; print_newline ();
    if ((List.length il) > 0)
    then 
      begin
        print_string "private "; 
        pp_list evapp_lid comma_sep il; print_newline ();
      end;
    evapp_block b;
  end;;


let rec evapp_constraint = function
    EVA_EQ_CONSTRAINT (i1, i2) -> 
      begin 
        evapp_lid i1; print_space ();
        print_string "="; print_space ();
        evapp_lid i2; 
      end
  | EVA_NEQ_CONSTRAINT (i1, i2) -> 
      begin 
        evapp_lid i1; print_space ();
        print_string "!="; print_space ();
        evapp_lid i2; 
      end
  | EVA_MEMBERSHIP_CONSTRAINT (i, tl) -> 
      begin 
        evapp_lid i; print_space ();
        print_string ":"; print_space(); 
        print_string "{"; 
        pp_list evapp_lterm comma_sep tl; 
        print_string "}"; 
      end
  | EVA_DOMAIN_CONSTRAINT (i, s) -> 
      begin 
        evapp_lid i; print_space ();
        print_string ":"; print_space(); 
        print_string s;
      end
and evapp_lconstraint { lco_loc = _; lco_constraint = c } =
  evapp_constraint c;;

let rec evapp_session s =
  begin
    open_box 0;
    (match s with
         EVA_BANG -> print_string "session*"
       | EVA_CONSTRAINED_SESSIONS l -> 
           begin 
             print_string "session*"; print_space(); 
             print_string "["; print_space();
             open_box 1; 
             pp_list evapp_lconstraint semicol_sep l; 
             close_box (); 
             print_space();  print_string "]"; 
           end
       | EVA_SESSIONS al -> 
           begin  
             print_string "session*"; print_space (); 
             print_string "("; print_space ();  
             open_box 1; 
             pp_list evapp_sassignment semicol_sep al; 
             close_box ();
             print_space (); print_string ")"; 
           end
       | EVA_SESSION (l, al) -> 
           begin 
             evapp_llabel l; print_space (); 
             print_string "session"; print_space (); 
             print_string "("; print_space (); 
             open_box 1; 
             pp_list evapp_sassignment semicol_sep al; 
             close_box ();
             print_space (); print_string ")"; 
           end
    );
    close_box ();
    print_newline ();
  end   
and evapp_lsession { lse_loc = _; lse_session = s } = 
  evapp_session s
and evapp_sassignment (i, v) = 
  begin 
    evapp_lid i; print_space(); 
    print_string "="; print_space (); 
    evapp_lterm v; 
  end;;

let rec evapp_atom = function
    EVA_TRUE -> print_string "true"
  | EVA_FALSE -> print_string ""
  | EVA_ATOM_EQ (t1, t2) -> 
      begin 
        evapp_lterm t1; print_space ();
        print_string "=="; print_space (); 
        evapp_lterm t2; 
      end
  | EVA_ATOM_NEQ (t1, t2) -> 
      begin evapp_lterm t1; print_space ();
        print_string "!="; print_space (); 
        evapp_lterm t2; 
      end
  | EVA_ATOM_HONEST t -> 
      begin 
        print_string "honest("; 
        evapp_lterm t; 
        print_string ")"; 
      end
  | EVA_ATOM_SECRET  t -> 
      begin 
        print_string "secret("; 
        evapp_lterm t; 
        print_string ")"; 
      end
  | EVA_ATOM_USER (i, t) -> 
      begin 
        evapp_lid i;
        print_string "("; 
        evapp_lterm t; 
        print_string ")"; 
      end
and evapp_latom { lat_loc = _; lat_atom = a } =
  evapp_atom a;;


let rec evapp_statement s = 
  begin
    open_box 0;
    (match s with
         EVA_ASSUME (ql, al, a) ->
           begin 
             print_string "assume"; print_space ();
             evapp_formula al a; print_space();
             (match ql with
                  [] -> ()
                | _  -> 
                    begin 
                      print_string "["; 
                      pp_list evapp_optional_quantification comma_sep ql;
                      print_string "]"; 
                    end
             );
           end
       | EVA_CLAIM (l, al, a) ->
           begin 
             print_string "claim"; print_space() ;
             evapp_llabel l; print_space ();
             evapp_formula al a; 
           end
    );
    close_box ();
    print_newline ();
  end
and evapp_lstatement { lst_loc = _; lst_statement = s } =
  evapp_statement s
(* for_tail = al; for_head = a *)
and evapp_formula al a =
  begin 
    (match al with
         [] -> ()
       | _  -> 
           begin 
             pp_list evapp_latom comma_sep al; print_space (); 
             print_string "=>"; print_space (); 
           end
    );
    evapp_latom a; 
  end;;

let evapp_spec = function 
    { sp_label = l;
      sp_declarations = dl;
      sp_protocol = p;
      sp_valuedcl = vdl;
      sp_sessions = sl;
      sp_statements = stl } ->
      begin
        open_box 0;
        evapp_lid l;
        print_newline (); print_newline ();
        pp_list evapp_ldcl null_sep dl;(* evapp_ldcllist dl; *) 
        print_newline ();
        evapp_lprotocol p;
        print_newline ();
        pp_list evapp_ldcl null_sep vdl; (* evapp_ldcllist vdl; *)
        print_newline ();
        pp_list evapp_lsession null_sep sl;
        print_newline ();
        pp_list evapp_lstatement null_sep stl;
        print_newline ();
        close_box ();
      end
