(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

open List;;

open Trace;;
open Abstract;;


(***************************************************************
 * constants, exceptions
 ***************************************************************)

(* let LIST_SIZE = 20;; *)

(* exception for accessors *)
exception Symbols_error of string;;

(* raised by assoc_symbol *)
exception Symbols_not_found;;

(* raised when adding a symbol already present to a symbol list 
 * the argumeent is the name of the symbol *)
exception Symbols_already of string;;


(***************************************************************
 * symbols types
 ***************************************************************)

(* first order symbol *)
type sy_fo =
    {
      sfo_name : string;
      sfo_scope : eva_scope;
      sfo_type: eva_type;
      sfo_fresh : bool;  
    };;

(* function symbol *)
type sy_fun =
    {
      sfu_name : string;
      sfu_scope : eva_scope;
      sfu_type: eva_type;
      sfu_signature : eva_type list;
      sfu_hash : bool;
      sfu_secret : bool;  
    };;

(* key symbol *)
type sy_key =
  {
    ske_name : string;
    ske_scope : eva_scope;
    ske_type: eva_type;
    ske_signature : eva_type list;
    ske_symkey : string;
    ske_algo : eva_algo;
  };;

(* value symbol *)
type sy_value =
{
  sva_name : string;
  sva_type : eva_type;
};;


(* alias symbol *)
type sy_alias =
{
  sal_name : string;
  sal_term : eva_lterm;
};;

(* user type *)
type sy_type =
{
  sty_name : string;
  sty_base : bool;
};;

(* quantified variable *)
type sy_quantified =
{
  squ_name : string;
  squ_type : eva_type;
};;

(* user predicate *)
type sy_predicate =
{
  spr_name : string;
};;


(* symbol associed to an id (string) *)
type eva_symbol = 
  (* declared first order identifier *)
    SY_FO of sy_fo
  (* declared function symbol *)
  | SY_FUN of sy_fun
  (* declared key *)
  | SY_KEY of sy_key
  (* value for instanciation in session *)
  | SY_VALUE of sy_value
  (* value for instanciation in session *)
  | SY_INTRUDER
  (* universally quantified var bounded to an axiom declaration 
   * or assume formula *)
  | SY_QUANTIFIED of sy_quantified
  (* name of an alias *)  
  | SY_ALIAS of sy_alias
  (* user type *)
  | SY_TYPE of sy_type
  | SY_PREDICATE of sy_predicate;;

(*
let empty_symbol =
{
  sym_name = "";
  sym_scope = EVA_SCOPE_VOID;
  sym_type = EVA_VOID;
  sym_signature = [];
  sym_isfresh = false;  
  sym_ishash = false;
  sym_issecret = false;  
  sym_symkey = "";
};;
*)

(***************************************************************
 * constructors
 ***************************************************************)

(* returns a symbol of kind first order
 * with the given attributes *)
let new_first_order_symbol name scope typ fresh =
  (SY_FO { sfo_name = name;
           sfo_scope = scope;
           sfo_type = typ;
           sfo_fresh = fresh; });;

(* returns a symbol of kind functional
 * with the given attributes *)
let new_functional_symbol name scope typ signature hash secret =
  (SY_FUN { sfu_name = name;
            sfu_scope = scope;
            sfu_type = typ;
            sfu_signature = signature;
            sfu_hash = hash;
            sfu_secret = secret; });;


(* returns a symbol of kind functional
 * with the given attributes *)
let new_key_symbol name scope typ signature symkey algo =
  (SY_KEY { ske_name = name;
            ske_scope = scope;
            ske_type = typ;
            ske_signature = signature;
            ske_symkey = symkey;
            ske_algo = algo; });;


(* returns a symbol of kind value
 * with the given attributes *)
let new_value_symbol name typ =
  (SY_VALUE { sva_name = name;
              sva_type = typ; });;


(* returns a symbol corresponding to the value "intruder" *)
let intruder_symbol = SY_INTRUDER;;


(* returns a symbol of kind alias
 * with the given attributes *)
let new_alias_symbol name term =
  (SY_ALIAS { sal_name = name;
              sal_term = term });;


(* returns a symbol of kind quantified variable
 * with the given attributes *)
let new_quantified_symbol name typ =
  (SY_QUANTIFIED { squ_name = name;
                   squ_type = typ });;

(* returns a symbol of kind type
 * with the given attributes *)
let new_type_symbol name bbase =
  (SY_TYPE { sty_name = name;
             sty_base = bbase; });;

(* returns a symbol of kind "predicate"
 * with the given attributes:
 * - name *)
let new_predicate_symbol name =
  (SY_PREDICATE { spr_name = name });;

(***************************************************************
 * accessors
 ***************************************************************)

(* returns true if the given symbol is of kind "first order",
 * false otherwise *)
let is_first_order = function
    SY_FO _ -> true
  | _ -> false;;


(* returns true if the given symbol is of kind "functional",
 * false otherwise *)
let is_functional = function
    SY_FUN _ -> true
  | _ -> false;;


(* returns true if the given symbol is of kind "key",
 * false otherwise *)
let is_key = function
    SY_KEY _ -> true
  | _ -> false;;


(* returns true if the given symbol is of kind "value",
 * false otherwise *)
let is_value = function
    SY_VALUE _ -> true
  | SY_INTRUDER -> true
  | _ -> false;;


(* returns true if the given symbol is the "intruder" value,
 * false otherwise *)
let is_intruder = function
    SY_INTRUDER -> true
  | _ -> false;;


(* returns true if the given symbol is of kind "quantified",
 * false otherwise *)
let is_quantified = function
    SY_QUANTIFIED _ -> true
  | _ -> false;;


(* returns true if the given symbol is of kind "alias",
 * false otherwise *)
let is_alias = function
    SY_ALIAS _ -> true
  | _ -> false;;


(* returns true if the given symbol is a user type,
 * false otherwise *)
let is_type = function
    SY_TYPE _ -> true
  | _ -> false;;


(* returns true if the given symbol is a user type
 * declared as basetype,
 * false otherwise *)
let is_basetype = function
    SY_TYPE t -> t.sty_base
  | _ -> false;;


(* returns true if the given symbol is a user predicate,
 * false otherwise *)
let is_predicate = function
    SY_PREDICATE _ -> true
  | _ -> false;;


(* returns the name of the given symbol.
   Symbols of every kinds have a name. *)
let get_name = function
    (SY_FO i) -> i.sfo_name
  | (SY_FUN f) -> f.sfu_name
  | (SY_KEY k) -> k.ske_name
  | (SY_VALUE v) -> v.sva_name
  (* the official name of the intruder *)
  | SY_INTRUDER -> "I"
  | (SY_QUANTIFIED q) -> q.squ_name
  (* name of an alias *)  
  | (SY_ALIAS a) -> a.sal_name
  | (SY_TYPE t) -> t.sty_name
  | (SY_PREDICATE t) -> t.spr_name;;


(* returns the type of the given symbol  
 *
 * raise Symbols_error if the given symbol is not typed *)
let get_type = function
    (SY_FO i) -> i.sfo_type
  | (SY_FUN f) -> f.sfu_type
  | (SY_KEY k) -> k.ske_type
  | (SY_VALUE v) -> v.sva_type
  | SY_INTRUDER -> EVA_PRINCIPAL
  | (SY_QUANTIFIED q) -> q.squ_type   
  | _  -> raise (Symbols_error "symbol has no type");;


(* returns the scope of the given symbol  
 *
 * raise Symbols_error if the given symbol has no such characteristic *)
let get_scope = function
    (SY_FO i) -> i.sfo_scope
  | (SY_FUN f) -> f.sfu_scope
  | (SY_KEY k) -> k.ske_scope
  | _  -> raise (Symbols_error "symbol has no scope");;


(* returns the "fresh" characteristic of the given symbol  
 *
 * raise Symbols_error if the given symbol has no such characteristic *)
let get_fresh = function
    (SY_FO i) -> i.sfo_fresh
  | _  -> raise (Symbols_error "symbol has no field fresh");;


(* returns the "hash" characteristic of the given symbol  
 *
 * raise Symbols_error if the given symbol has no such characteristic *)
let get_hash = function
    (SY_FUN f) -> f.sfu_hash
  | _  -> raise (Symbols_error "symbol has no field hash");;


(* returns the "secret" characteristic of the given symbol  
 *
 * raise Symbols_error if the given symbol has no such characteristic *)
let get_secret = function
    (SY_FUN f) -> f.sfu_secret
  | _  -> raise (Symbols_error "symbol has no field secret");;


(* returns the signature of the given symbol  
 *
 * raise Symbols_error if the given symbol has no such characteristic *)
let get_signature = function
    (SY_FUN f) -> f.sfu_signature
  | (SY_KEY k) -> k.ske_signature
  | _  -> raise (Symbols_error "symbol has no signature");;


(* returns the name of the symmetric key associed to the given symbol 
 *
 * raise Symbols_error if the given symbol has no such characteristic *)
let get_symkey = function
    (SY_KEY k) -> k.ske_symkey
  | _  -> raise (Symbols_error "symbol has no symmetric key");;


(* returns the algorithm associed with the given symbol  
 *
 * raise Symbols_error if the given symbol has no such characteristic *)
let get_algo = function
    (SY_KEY k) -> k.ske_algo
  | _  -> raise (Symbols_error "symbol has no algo");;


(* returns the term associed with the given alias symbol  
 *
 * raise Symbols_error if the given symbol is not of kind alias *)
let get_term = function
    (SY_ALIAS a) -> a.sal_term
  | _ -> raise (Symbols_error "symbol is not an alias");;


(***************************************************************
 * symbol list
 ***************************************************************)

type eva_symbol_list =
    SYMBOLLIST of eva_symbol list;;

let empty_symbol_list = (SYMBOLLIST []);;

(* returns true if the given symbol list sl 
 * contains a symbol declared with the given name,
 * returns false otherwise. *)
let rec memassoc_symbol name (SYMBOLLIST sl) = 
  memassoc_symbold name sl
and  memassoc_symbold name = function
    [] -> false
  | s::l -> 
      if (name = (get_name s))
      then true
      else memassoc_symbold name l;;


(* returns the symbol declared with the given name
 * in the given symbol list sl.
 *
 * raise Symbols_not_found if the given name is not declared in sl *)
let rec assoc_symbol name (SYMBOLLIST sl) = 
  (* trace "assoc_symbol: %s\n" name; *)
  assoc_symbold name sl
and assoc_symbold name = function
    [] -> raise Symbols_not_found
  | s::l -> 
      (* trace "assoc_symbold name: name = %s, symbol.name = %s\n" name (get_name s); *)
      if (name = (get_name s))
      then s
      else assoc_symbold name l;;


(* returns a symbol list obtained 
 * by adding the given symbol s at the begining
 * of the given symbol list 
 *
 * raise Symbols_already with the name of s
 * if there is already a symbol with the same name as s 
 * in the given symbol list *)
let rec cons_symbol s (SYMBOLLIST sl) = 
  let n = (get_name s) 
  in begin
      (* trace "Symbols.cons_symbol %s\n" n; *)
      if (memassoc_symbol n (SYMBOLLIST sl))
      then raise (Symbols_already n)
      else (SYMBOLLIST (s::sl))
    end;;


(* returns a symbollist obtained 
 * by adding the given symbol s at the end of
 * the given symbol list.
 *
 * raise Symbols_already with the name of s
 * if there is already a symbol with the same name as s 
 * in the given symbol list *)
let rec append_symbol s (SYMBOLLIST sl) = 
  (* trace "Symbols.apend_symbol %s\n" (get_name s); *)
  (SYMBOLLIST (append_symbold s sl))
and append_symbold s = function
    [] -> [s]
  | (a::l) -> 
      let n = (get_name a) in
        if (n = (get_name s))
        then raise (Symbols_already n)
        else (a::(append_symbold s l));;
        


(* add_symbol s sl 
 * returns a symbollist obtained 
 * by adding the given symbol s at the end of
 * the given symbol list sl only if sl does
 * not already contain a symbol with the same name as s. *)
let rec add_symbol s (SYMBOLLIST sl) = 
  (* trace "Symbols.add_symbol %s\n" (get_name s); *)
  (SYMBOLLIST (add_symbold s sl))
and add_symbold s = function
    [] -> [s]
  | (a::l) -> 
      if ((get_name a) = (get_name s))
      then (a::l)
      else (a::(add_symbold s l));;

        
let rec overwrite_symbol s (SYMBOLLIST sl) = 
  (* trace "Symbols.overwrite_symbol %s\n" (get_name s); *)
  (SYMBOLLIST (overwrite_symbold s sl))
and overwrite_symbold s = function
    [] -> [s]
  | a::l -> 
      if ((get_name a) = (get_name s))
      then (s::l)
      else (a::(overwrite_symbold s l));;

(* return the symbols list obtained by
 * adding the symbols in the given list l2 at the end of l1.
 *
 * raise Symbols_already if l1 contains already one symbol with
 * the same name as a symbol of l2 *)
let rec append_symbollist (SYMBOLLIST l1) (SYMBOLLIST l2) = 
  match l1 with
      [] -> (SYMBOLLIST l2)
    | s::l -> 
        (cons_symbol s (append_symbollist (SYMBOLLIST l) (SYMBOLLIST l2)));;

(* return the symbols list obtained by
 * adding the symbols in the given list l2 
 * such that there is no symbol with the same name in l1
 * at the end of l1. *)
let rec add_symbollist (SYMBOLLIST l1) (SYMBOLLIST l2) = 
  add_symbollistd l1 (rev l2)
and add_symbollistd l1 l2 = 
  match l2 with
      [] -> (SYMBOLLIST l1)
    | s::l -> 
        (add_symbol s (add_symbollist (SYMBOLLIST l1) (SYMBOLLIST l)));;


(* return the symbols list obtained by
 * overwriting the symbols in the given list l2
 * with the symbols in the given list l1
 * (see overwrite_symbol). *)
let rec overwrite_symbollist (SYMBOLLIST l1) (SYMBOLLIST l2) =
  match l1 with
      [] -> (SYMBOLLIST l2)
    | s::l -> 
        (overwrite_symbol s 
           (overwrite_symbollist (SYMBOLLIST l) (SYMBOLLIST l2)));;

let symbol_list_length (SYMBOLLIST l) =
  List.length l;;


let iter_symbollist f (SYMBOLLIST l) =
  (List.iter f l);;

let map_symbollist f (SYMBOLLIST l) =
  (List.map f l);;

let fold_left_symbollist f a (SYMBOLLIST l) =
  List.fold_left f a l;;

let fold_right_symbollist f (SYMBOLLIST l) a =
  List.fold_right f l a;;



