(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(** Manage the abstract data type of symbols of the protocol specifications *)

(** There are 8 kinds of symbols:
  - first order identifiers 
    (either constant or parameter or private process variable)
  - functional identifiers
    (either constant or parameter or private process variable)
  - constructors for asymetric keys, 
    declared as component of a key pair
  - values for instanciation of sessions
  - intruder, a special reserved value
  - quantified variable in an axiom declaration 
    or an "assume" formula 
  - alias
  - user type *)

open Abstract;;

(***************************************************************)
(** {2 Exceptions}                                             *)
(***************************************************************)

(** raised by accessors *)
exception Symbols_error of string

(** raised by {!Symbols.assoc_symbol} *)
exception Symbols_not_found

(** raised when adding a symbol already present to a symbol list 
   the argumeent is the name of the symbol *)
exception Symbols_already of string


(***************************************************************)
(** {2 Symbols type}                                           *)
(***************************************************************)

(** type of symbols *)
type eva_symbol

(***************************************************************)
(** {2 constructors}                                           *)
(***************************************************************)

(** @return a symbol of kind "first order"
   with the given attributes:
  - name
  - quality 
  - type
  - [true] if the identifier is fresh, [false] otherwise. *)
val new_first_order_symbol :
  string -> eva_scope -> eva_type -> bool -> eva_symbol

(** @return a symbol of kind "functional"
   with the given attributes:
  - name
  - quality 
  - type
  - signature
  - [true] if the function is one-way, [false] otherwise. 
  - [true] if the function is secret, [false] otherwise. *)
val new_functional_symbol :
  string ->
  eva_scope ->
  eva_type -> eva_type list -> bool -> bool -> eva_symbol

(** @return a symbol of kind "key" constructor
   with the given attributes:
  - name
  - quality 
  - type
  - signature
  - name of the other key of the key pair
  - associed encryption algorithm *)
val new_key_symbol :
  string ->
  eva_scope ->
  eva_type ->
  eva_type list -> string -> eva_algo -> eva_symbol

(** @return a symbol of kind "value"
   with the given attributes:
  - name
  - type  *)
val new_value_symbol :
  string -> eva_type -> eva_symbol

(** @return a symbol corresponding to the value "intruder" *)
val intruder_symbol :
  eva_symbol

(** @return a symbol of kind "alias"
   with the given attributes:
  - name
  - replacement located term  *)
val new_alias_symbol :
  string -> eva_lterm -> eva_symbol

(** @return a symbol of kind "quantified variable"
   with the given attributes:
  - identifier
  - type in quantification  *)
val new_quantified_symbol :
  string -> eva_type -> eva_symbol

(** @return a symbol of kind "type"
   with the given attributes:
   - name
   - [true] if it is a basetype, [false] otherwise *)
val new_type_symbol :
  string -> bool -> eva_symbol

(** @return a symbol of kind "predicate"
   with the given attributes:
   - name *)
val new_predicate_symbol :
  string -> eva_symbol

(***************************************************************
 * accessors
 ***************************************************************)

(** @return [true] if the given symbol is of kind "first order",
   [false] otherwise *)
val is_first_order :
  eva_symbol -> bool

(** @return [true] if the given symbol is of kind "functional",
   [false] otherwise *)
val is_functional : 
  eva_symbol -> bool

(** @return [true] if the given symbol is of kind "key",
   [false] otherwise *)
val is_key : 
  eva_symbol -> bool

(** @return [true] if the given symbol is of kind "value",
   [false] otherwise *)
val is_value : 
  eva_symbol -> bool

(** @return [true] if the given symbol is the "intruder" value,
   [false] otherwise *)
val is_intruder : 
  eva_symbol -> bool

(** @return [true] if the given symbol is of kind "quantified variable",
   [false] otherwise *)
val is_quantified : 
  eva_symbol -> bool

(** @return [true] if the given symbol is of kind "alias",
   [false] otherwise *)
val is_alias : 
  eva_symbol -> bool

(** @return [true] if the given symbol is of kind "type",
   [false] otherwise *)
val is_type : 
  eva_symbol -> bool

(** @return [true] if the given symbol is a user type
   declared as basetype, subcase of [is_type]
   [false] otherwise *)
val is_basetype : 
  eva_symbol -> bool

(** @return [true] if the given symbol is of kind "predicate",
   [false] otherwise *)
val is_predicate : 
  eva_symbol -> bool

(** @return the name of the given symbol.
   Symbols of every kinds have a name. *)
val get_name : 
  eva_symbol -> string

(** @return the type of the given symbol.
  @raise Symbols_error if the given symbol is not typed. *)
val get_type : 
  eva_symbol -> eva_type

(** @return the scope of the given symbol.  
  @raise Symbols_error if the given symbol has no such characteristic. *)
val get_scope :
  eva_symbol -> eva_scope

(** @return the "fresh" characteristic of the given symbol.  
  @raise Symbols_error if the given symbol has no such characteristic. *)
val get_fresh :
  eva_symbol -> bool

(** @return the "hash" characteristic of the given symbol.  
  @raise Symbols_error if the given symbol has no such characteristic. *)
val get_hash :
  eva_symbol -> bool

(** @return the "secret" characteristic of the given symbol.  
  @raise Symbols_error if the given symbol has no such characteristic. *)
val get_secret :
  eva_symbol -> bool

(** @return the signature of the given symbol.  
  @raise Symbols_error if the given symbol has no such characteristic. *)
val get_signature :
  eva_symbol -> eva_type list

(** @return the name of the symmetric key associed to the given symbol. 
  @raise Symbols_error if the given symbol has no such characteristic. *)
val get_symkey :
  eva_symbol -> string

(** @return the algorithm associed with the given symbol.  
  @raise Symbols_error if the given symbol has no such characteristic. *)
val get_algo :
  eva_symbol -> eva_algo

(** @return the term associed with the given alias symbol.  
  @raise Symbols_error if the given symbol is not of kind "alias". *)
val get_term :
  eva_symbol -> eva_lterm


(****************************************************************
 * symbol list
 ***************************************************************)

(** Abstract type for lists of symbols *)
type eva_symbol_list

(** generic empty symbol list. *)
val empty_symbol_list :
  eva_symbol_list

(** @return [true] if the given symbol list [sl] 
   contains a symbol declared with the given name,
   returns [false] otherwise. *)
val memassoc_symbol  :
  string -> eva_symbol_list -> bool

(** @return the symbol declared with the given name
   in the given symbol list [sl].
  
  @raise Symbols_not_found if the given name is not declared in [sl]. *)
val assoc_symbol :
  string -> eva_symbol_list -> eva_symbol

(** @return a symbols list obtained 
   by adding the given symbol [s] at the end of
   the given symbol list [sl].
 
   @raise Symbols_already with the name of [s]
   if there is already a symbol with the same name as [s]
   in the given symbol list [sl].*)
val append_symbol :
  eva_symbol -> eva_symbol_list -> eva_symbol_list

(** [add_symbol s sl] 
   @return a symbols list obtained 
   by adding the given symbol [s] at the end of
   the given symbol list [sl] only if [sl] does
   not already contain a symbol with the same name as [s]. *)
val add_symbol :
  eva_symbol -> eva_symbol_list -> eva_symbol_list

(** @return a symbols list obtained from the
  the given symbol [s] and the given symbol list [sl] as follows:
  - if [sl] contains no symbol with the name of [s], 
  append [s] to [sl]
  - if [sl] contains a symbol [s1] with the name of [s],
  returns [sl] where [s] has been replaced by [s1]. *)
val overwrite_symbol :
  eva_symbol -> eva_symbol_list -> eva_symbol_list


(** @return the symbols list obtained by
  adding the symbols in the second given list [l2] at the end 
  of the first symbols list [l1].
 
  @raise Symbols_already if [l1] contains already one symbol with
  the same name as a symbol of [l2]. *)
val append_symbollist :
  eva_symbol_list -> eva_symbol_list -> eva_symbol_list


(** @return the symbols list obtained by
  overwriting the symbols in second the given list [l2]
  with the symbols in the first given list [l1]
  (see {!Symbols.overwrite_symbol}). *)
val overwrite_symbollist :
  eva_symbol_list -> eva_symbol_list -> eva_symbol_list


(** @return the length of the given symbol list [sl]. *)
val symbol_list_length :
   eva_symbol_list -> int

(** Iterator for symbol lists. 
  Apply in a row the given function [f] to each of the symbols
  contained in the given symbol list [sl]. *)
val iter_symbollist : 
  (eva_symbol -> unit) -> eva_symbol_list -> unit

(** [map_symbollist f sl]
  @return the list [ (f s1); ...; (f sn) ]
  where [s1...sn] are the symbols of the given symbol list [sl]. *)
val map_symbollist : 
  (eva_symbol -> 'a) -> eva_symbol_list -> 'a list


(** [fold_left_symbollist f a sl]
  @return [f (... (f (f a s1) s2) ...) sn] 
  where [s1...sn] are the symbols of the given symbol list [sl]. *)
val fold_left_symbollist : 
  ('a -> eva_symbol -> 'a) -> 'a -> eva_symbol_list -> 'a

(** [fold_right_symbollist f sl b] 
  @return [f s1 (f s2 (... (f sn b) ...))]
  where [s1...sn] are the symbols of the given symbol list [sl]. *)
val fold_right_symbollist : 
  (eva_symbol -> 'b -> 'b) -> eva_symbol_list -> 'b -> 'b

