(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(* utilities for term replacement in abstract specifications *)

open List;;

open Trace;;
open Location;;
open Abstract;;


exception Termutil_error of string * location;;


(*************************************************
 *  Term replacement 
 *************************************************)

(* 
 * tr : a function lterm -> lterm
 * d : located declaration *)
let rec lterm_replace_ldcl tr d = 
  trace "Termutil.lterm_replace_ldcl\n";
  { d with ldc_dcl = (lterm_replace_dcl tr d.ldc_dcl) }
(* returns a eva_dcl *)
and lterm_replace_dcl tr = function
    EVA_DCL_ALIAS (i, t) -> 
      EVA_DCL_ALIAS (i, (tr t))
  (* know_principal = i; know_contents = tl; *)
  | EVA_DCL_KNOW (i, kl) -> 
      EVA_DCL_KNOW (i, (map tr kl))
  | EVA_DCL_EVERY tl -> 
      EVA_DCL_EVERY (map tr tl)
  (* ax_forall = itl; ax_left = t1; ax_right = t2 *)
  | EVA_DCL_AXIOM (t1, t2, t, itl) ->
      EVA_DCL_AXIOM ((tr t1), (tr t2), t, itl)
  (* the other declarations contain no terms *)
  | dd -> dd
;;

let rec lterm_replace_linstr tr i =
  trace "Termutil.lterm_replace_linstr\n";
  { i with 
      lin_instr = (lterm_replace_instr tr i.lin_instr) }
(* returns a eva_instr *)
and lterm_replace_instr tr = function
    EVA_SKIP -> EVA_SKIP
    (* msg_label = l; msg_sender = i1; msg_receiver = i2; msg_body = t; *)
  | EVA_MSG (l, i1, i2, t) ->
      EVA_MSG (l, i1, i2, (tr t))
    (* as_role = i1; as_variable = i2; as_value = t; *)
  | EVA_ASSIGN (i1, i2, t) -> 
      EVA_ASSIGN (i1, i2, (tr t))
    (* co_role = i1; co_left_variable = i2; co_right_variable = t; *)
  | EVA_COMP (i1, i2, t) ->
      EVA_COMP (i1, i2, (tr t))
  | EVA_BLOCK il -> 
      EVA_BLOCK (map (lterm_replace_linstr tr) il)
  | EVA_SWITCH (t, cl) -> 
      EVA_SWITCH ((tr t), (map (lterm_replace_case tr) cl))
and lterm_replace_case tr c =
  { c with
      case_case = (tr c.case_case);
      case_block = (map (lterm_replace_linstr tr) c.case_block);
  };;

let rec lterm_replace_lprotocol tr p =
  trace "Termutil.lterm_replace_lprotocol\n";
  { p with lpr_protocol = (lterm_replace_protocol tr p.lpr_protocol) }
and lterm_replace_protocol tr = function
    EVA_MP il -> EVA_MP (map (lterm_replace_linstr tr) il)
  | EVA_PROCESS pl -> EVA_PROCESS (map (lterm_replace_process tr) pl)
(* TBC: no replacement of the role and privates vars *)
and lterm_replace_process tr p =
  { p with      
      pr_messages = map (lterm_replace_linstr tr) p.pr_messages
  };;

let rec lterm_replace_lsession tr s = 
  { s with lse_session = (lterm_replace_session tr s.lse_session) }
and lterm_replace_session tr = function
    EVA_BANG ->
      EVA_BANG
  | EVA_CONSTRAINED_SESSIONS cl ->
      EVA_CONSTRAINED_SESSIONS (map (lterm_replace_lconstraint tr) cl)
  | EVA_SESSIONS itl ->
      EVA_SESSIONS (map (fun (i, t) -> (i, (tr t))) itl)
  | EVA_SESSION (l, itl) ->
      EVA_SESSION (l, (map (fun (i, t) -> (i, (tr t))) itl))
and lterm_replace_lconstraint tr c = 
  { c with lco_constraint = (lterm_replace_constraint tr c.lco_constraint) }
and lterm_replace_constraint tr = function
    EVA_EQ_CONSTRAINT (i1, i2) -> 
      EVA_EQ_CONSTRAINT (i1, i2) 
  | EVA_NEQ_CONSTRAINT (i1, i2) -> 
      EVA_NEQ_CONSTRAINT (i1, i2) 
  | EVA_MEMBERSHIP_CONSTRAINT (i, tl) -> 
      EVA_MEMBERSHIP_CONSTRAINT (i, (map tr tl))
  | EVA_DOMAIN_CONSTRAINT (i, s) -> 
      EVA_DOMAIN_CONSTRAINT (i, s);;


let rec lterm_replace_latom tr a =
  trace "Termutil.lterm_replace_atom\n";
  { a with lat_atom = (lterm_replace_atom tr a.lat_atom) }
and lterm_replace_atom tr = function
    EVA_TRUE -> EVA_TRUE
  | EVA_FALSE -> EVA_FALSE
  | EVA_ATOM_EQ (t1, t2) -> 
      EVA_ATOM_EQ ((tr t1), (tr t2))
  | EVA_ATOM_NEQ (t1, t2) -> 
      EVA_ATOM_NEQ ((tr t1), (tr t2))
  | EVA_ATOM_HONEST t -> EVA_ATOM_HONEST (tr t)
  | EVA_ATOM_SECRET t -> EVA_ATOM_SECRET (tr t)
  | EVA_ATOM_USER (i, t) -> EVA_ATOM_USER (i, (tr t));;

let rec lterm_replace_lstatement tr s =
  { s with lst_statement = (lterm_replace_statement tr s.lst_statement) }
and lterm_replace_statement tr = function
    EVA_ASSUME (ql, al ,a) -> 
      EVA_ASSUME (ql, map (lterm_replace_latom tr) al, lterm_replace_latom tr a)
  | EVA_CLAIM (l, al, a)  -> 
      EVA_CLAIM (l, map (lterm_replace_latom tr) al, lterm_replace_latom tr a);;

let rec lterm_replace_spec tr sp =
  {  sp with 
       sp_declarations = (map (lterm_replace_ldcl tr) sp.sp_declarations);
       sp_protocol = lterm_replace_lprotocol tr sp.sp_protocol;
       sp_sessions = (map (lterm_replace_lsession tr) sp.sp_sessions);
       sp_statements = (map (lterm_replace_lstatement tr) sp.sp_statements);
  };;


(*************************************************
 *  Term (syntactic) equality
 *************************************************)

let string_of_lid i = 
  match i.lid_id with
      EVA_ID s -> s;;
(*     | _ -> raise (Internal_error "string_of_lid");; *)

let lid_equal i1 i2 =
  (string_of_lid i1) = (string_of_lid i2);;

let rec lterm_equal t1 t2 =
  match (t1.lte_term, t2.lte_term) with
      (EVA_TERM_ID s1, EVA_TERM_ID s2) -> 
        s1 = s2
    | (EVA_TERM_CST s1, EVA_TERM_CST s2) -> 
        s1 = s2
    | (EVA_TERM_VALUE v1, EVA_TERM_VALUE v2) ->
        (* structural eq works for eva_value *)
        (v1 = v2)
    | (EVA_TERM_ALIAS (a1, t1), EVA_TERM_ALIAS (a2, t2)) ->
        (a1 = a2)&&(lterm_equal t1 t2)
    | (EVA_LOCATED (v1, r1), EVA_LOCATED (v2, r2)) ->
        (lid_equal v1 v2)&&(lid_equal r1 r2)
    | (EVA_QUANTIFIED (v1, t1), EVA_QUANTIFIED (v2, t2)) ->
        (* structural eq works for eva_type *)
        (v1 = v2)&&(t1 = t2)
    | (EVA_P t1, EVA_P t2) ->
        lterm_equal t1 t2
    | (EVA_A t1, EVA_A t2) ->
        lterm_equal t1 t2
    | (EVA_SA t1, EVA_SA t2) ->
        lterm_equal t1 t2
    | (EVA_AA t1, EVA_AA t2) ->
        lterm_equal t1 t2
    | (EVA_U (i1, t1), EVA_U (i2, t2)) ->
        (i1 = i2) && (lterm_equal t1 t2)
    | (EVA_CONS tl1, EVA_CONS tl2) ->
        ltermlist_equal tl1 tl2 
    | (EVA_APP (i1, tl1), EVA_APP (i2, tl2)) ->
        (lid_equal i1 i2)&&(ltermlist_equal tl1 tl2)
    | (EVA_CRYPT (a1, k1, t1), EVA_CRYPT (a2, k2, t2)) ->
        (algo_equal a1 a2)&&(lterm_equal k1 k2)&&(lterm_equal t1 t2)
    | (EVA_SIGN (a1, k1, t1), EVA_SIGN (a2, k2, t2)) ->
        (algo_equal a1 a2)&&(lterm_equal k1 k2)&&(lterm_equal t1 t2)
    | (EVA_PCENT (u1, v1), EVA_PCENT (u2, v2)) ->
        (lterm_equal u1 u2)&&(lterm_equal v1 v2)
    | (_, _) -> false
and ltermlist_equal tl1 tl2 =
  try 
    List.for_all2 lterm_equal tl1 tl2
  with 
      (Invalid_argument _) -> false
and algo_equal a1 a2 =
  match (a1, a2) with
      (EVA_VANILLA, EVA_VANILLA) -> 
        true
    | (EVA_ALGO t1, EVA_ALGO t2) ->
        (lterm_equal t1 t2)
    | (_, _) -> false
and value_equal v1 v2 =
  match (v1, v2) with
      (EVA_INTRUDER, EVA_INTRUDER) ->
        true
    | (EVA_VALUE s1, EVA_VALUE s2) ->
        s1 = s2
    | (_,_) -> false
      
(*************************************************
 *  Test
 *************************************************)

(* is_atomic_term t 
 * returns true if t is an id or a value,
 * and returns false otherwise. *)
let rec is_atomic_term = function
    EVA_TERM_ID _ 
  | EVA_TERM_CST _ 
  | EVA_TERM_VALUE _ 
  | EVA_LOCATED (_,_)
  | EVA_QUANTIFIED (_,_)  -> true
  | EVA_TERM_ALIAS (a, t) -> (is_atomic_lterm t)
  | _ -> false
and is_atomic_lterm t =
  is_atomic_term t.lte_term;;


(**************************************************
 * Substitutions
 *
 * A substitution is a list of pairs (string, lterm),
 * where string is the identifier of a symbol.
 * The located term in the image of a substitution
 * must 
 **************************************************)

let subst_dom s = 
  (map fst s);;


(* subst_term s t
 * returns the substitution of 
 * the given located term t
 * by the given substitution s.
 * 
 * The symbol list is used to determine which symbols are constant,
 * constants are not subsituted.
 *
 * the term t must contain no alias *)
let rec subst_term s term = 
  match term with
      EVA_TERM_ID i -> 
        subst_symbol s i
    | EVA_TERM_CST _ 
    | EVA_TERM_VALUE _ -> 
        term
    | EVA_TERM_ALIAS (_, _) ->
        raise (Internal_error "subst_lterm: alias")
    | EVA_LOCATED (_, _) ->
        (* do not substitute the located variables *)
        term
    | EVA_QUANTIFIED (_, _) -> 
        (* these var are not in subst *)
        term
    | EVA_P t ->
        EVA_P (subst_lterm s t)
    | EVA_A t ->
        EVA_A (subst_lterm s t)
    | EVA_SA t ->
        EVA_SA (subst_lterm s t)
    | EVA_AA t ->
        EVA_AA (subst_lterm s t)
    | EVA_U (i, t) ->
        EVA_U (i, (subst_lterm s t))
    | EVA_CONS tl ->
        EVA_CONS (map (subst_lterm s) tl)
    | EVA_APP (i, tl) ->      
        (* id i is a constant (composed function symbols not allowed)
         * and hence is not in the domain of the subst s *)
        EVA_APP (i, (map (subst_lterm s) tl))
    | EVA_CRYPT (a, k, t) ->
        EVA_CRYPT (subst_algo s a, subst_lterm s k, subst_lterm s t)
    | EVA_SIGN  (a, k, t) ->
        EVA_CRYPT (subst_algo s a, subst_lterm s k, subst_lterm s t)
    | EVA_PCENT (t1, t2) ->
        EVA_PCENT (subst_lterm s t1, subst_lterm s t2)
and subst_lterm s t = 
  { t with lte_term = (subst_term s t.lte_term) }
(* subst_symbol s i
 * returns the located term associed to the 
 * string i by the substitution s *)
and subst_symbol s i = 
  try 
    (List.assoc i s)
  with 
      Not_found -> 
        raise (Internal_error ("subst_symbol: "^i^" not found in subst"))
and subst_algo s = function
    EVA_VANILLA -> EVA_VANILLA
  | EVA_ALGO t -> EVA_ALGO (subst_lterm s t);;


(* term_vars t 
 * returns the list (with repetitions) 
 * of identifiers of parameters and private variables
 * in the given term t *)
let rec term_vars = function
    EVA_TERM_ID i -> 
      [i]
  | EVA_TERM_CST _ 
  | EVA_TERM_VALUE _ -> 
      []
  | EVA_TERM_ALIAS (a, t) ->
      lterm_vars t
  | EVA_LOCATED (_, _) 
  | EVA_QUANTIFIED (_, _) -> 
      []
  | EVA_P t 
  | EVA_A t 
  | EVA_SA t 
  | EVA_AA t
  | EVA_U (_, t) -> 
      lterm_vars t
  | EVA_CONS tl ->
      List.concat (List.map lterm_vars tl)
  | EVA_APP (i, tl) ->      
        (* id i is a constant (composed function symbols not allowed)
         * and hence is not returned *)
      List.concat (List.map lterm_vars tl)
  | EVA_CRYPT (a, k, t) 
  | EVA_SIGN  (a, k, t) ->
      List.concat [(algo_vars a); (lterm_vars k); (lterm_vars t)]
  | EVA_PCENT (t1, t2) ->
      List.concat [(lterm_vars t1); (lterm_vars t2)]
and lterm_vars t = 
  term_vars t.lte_term
and algo_vars = function
    EVA_VANILLA ->
      []
  | EVA_ALGO t ->
      lterm_vars t;;


(* is_grounding s t
 * returns true iff the domain of the given substitution s 
 * contains all the parameters and private variables of
 * the term t *)
let rec is_grounding s t =  
  (subset (term_vars t) (subst_dom s))
and subset l1 l2 = 
  match l1 with
      [] -> true
    (* structural eq for strings *)
    | a::l -> (List.exists (fun x -> (x = a)) l2)&&(subset l l2);;

