(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(** Utilities for term replacement in abstract protocol specifications *)

open Location;;
open Abstract;;

(** raised by functions the [termutil] module to signal errors 
  in the input file.
  - message
  - location of the error *)
exception Termutil_error of string * location;;

(*************************************************
 *  Term replacement 
 *************************************************)

(** [lterm_replace_ldcl tr d]
  @return the given located declaration [d] in which 
  every term has been replaced by 
  its image under the given term replacement function [tr]. *)
val lterm_replace_ldcl :
  (eva_lterm -> eva_lterm) -> eva_ldcl -> eva_ldcl

val lterm_replace_dcl :
  (eva_lterm -> eva_lterm) -> eva_dcl -> eva_dcl

val lterm_replace_linstr :
  (eva_lterm -> eva_lterm) -> eva_linstr -> eva_linstr

val lterm_replace_instr :
  (eva_lterm -> eva_lterm) -> eva_instr -> eva_instr

val lterm_replace_lprotocol :
  (eva_lterm -> eva_lterm) -> eva_lprotocol -> eva_lprotocol

val lterm_replace_protocol :
  (eva_lterm -> eva_lterm) -> eva_protocol -> eva_protocol

val lterm_replace_process :
  (eva_lterm -> eva_lterm) -> eva_process -> eva_process

val lterm_replace_lsession :
  (eva_lterm -> eva_lterm) -> eva_lsession -> eva_lsession

val lterm_replace_session :
  (eva_lterm -> eva_lterm) -> eva_session -> eva_session

val lterm_replace_latom :
  (eva_lterm -> eva_lterm) -> eva_latom -> eva_latom

val lterm_replace_atom :
  (eva_lterm -> eva_lterm) -> eva_atom -> eva_atom

val lterm_replace_lstatement :
  (eva_lterm -> eva_lterm) -> eva_lstatement -> eva_lstatement

val lterm_replace_statement :
  (eva_lterm -> eva_lterm) -> eva_statement -> eva_statement

val lterm_replace_spec :
  (eva_lterm -> eva_lterm) -> eva_spec -> eva_spec



(*************************************************
 *  Term (syntactic) equality
 *************************************************)

val string_of_lid :
  eva_lid -> string;;

val lid_equal :
  eva_lid -> eva_lid -> bool

val lterm_equal :
  eva_lterm -> eva_lterm -> bool

      
(*************************************************
 *  Test
 *************************************************)

(** is_atomic_term t 
  @return [true] if the given term [t] 
  is a first order id or a value,
  or an alias to an atomic located term,
  and returns false otherwise. *)
val is_atomic_term :
  eva_term -> bool

(** [is_atomic_lterm t]
  @return [true] if the given located term [t] 
  is a first order id or a value,
  or an alias to an atomic located term,
  and returns false otherwise. *)
val is_atomic_lterm :
  eva_lterm -> bool
  
(**************************************************
 * Substitutions
 *
 * A substitution is a list of pairs (string, lterm),
 * where string is the identifier of a symbol.
 * The located term in the image of a substitution
 * must 
 **************************************************)

val subst_dom : 
  ('a * 'b) list -> 'a list

(** [subst_term s t]
  returns the substitution of 
  the given term [t]
  by the given substitution [s].
  
  Constants are not subsituted.
  The term [t] must contain no alias. *)
val subst_term :
  (string * eva_term) list -> eva_term -> eva_term

val subst_lterm :
  (string * eva_term) list -> eva_lterm -> eva_lterm

(** [is_grounding s t]
   returns [true] iff the domain of the given substitution [s] 
   contains all the parameters and private variables of
   the term [t]. *)
val is_grounding : 
  (string * 'a) list -> eva_term -> bool

(** [term_vars t]
   returns the list (with repetitions) 
   of identifiers of parameters and private variables
   in the given term [t]. *)
val term_vars :
  eva_term -> string list
val lterm_vars :
  eva_lterm -> string list

