(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

open List;;

open Trace;;
open Location;;
open Abstract;;
open Symbols;;
open Types;;
open Termutil;;


exception Translator_error of string * location;;


(* internal exception: must be raised caught by the functions of this file 
 * used to catch location (line number) of error in parsed structures *)
(* exception Tcaught_error of string;; *)



(***************************************************************
 * Extraction of a symbol list from a spec
 ***************************************************************)

let string_of_dcllid i = 
  match i.lid_id with
      EVA_ID "I" -> raise (Translator_error
                             ("the identifier I is reserved for the intruder",
                              i.lid_loc))
    | EVA_ID s -> s;;
(*    | _ -> raise (Internal_error "string_of_dcllid");; *)
  

(* returns the symbol list obtained by
 * concatenation of the given symbol list and
 * the list of symbols for each symbol declared in the given eva_dcl 
 * (one eva_dcl can declare several symbols) *)
let rec append_symbollist_of_ldcl sl d =
  match d.ldc_dcl with
      EVA_DCL_ID (ibl, s, t) -> 
        append_symbollist_of_dcl_ids sl (ibl, s, t)
    | EVA_DCL_FUN  (i, s, t, tl, bh, bs) -> 
        append_symbollist_of_dcl_fun sl (i, s, t, tl, bh, bs)
    | EVA_DCL_KEYPAIR (a, k1, k2, s, t, tl) -> 
        append_symbollist_of_dcl_keypair sl (a, k1, k2, s, t, tl)
    | EVA_DCL_ALIAS (i, t) ->  
        append_symbollist_of_dcl_alias sl (i, t)
    | EVA_DCL_LOCALALIAS (l, i, t) -> sl
    | EVA_DCL_BASETYPE s -> 
        append_symbollist_of_ldcl_basetype sl d
    | EVA_DCL_KNOW (_, _) -> sl
    | EVA_DCL_EVERY l -> sl
    | EVA_DCL_AXIOM (_, _, _, _) -> sl
    | EVA_DCL_VALUE (vl, t) -> append_symbollist_of_dcl_values sl (vl, t)
(* add_symbollist_of_dcl_usertype sl t
 * returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the symbol corr. to the given located user type t,
 * if there is no symbols already in sl with the same name. *)
and add_symbollist_of_dcl_usertype sl t =
  match t.lty_type with
      EVA_USERTYPE s -> 
        add_symbol (new_type_symbol s false) sl
    | _ -> sl
(* add_symbollist_of_dcl_usertype sl t
 * returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the symbol corr. to the given user types inh the given list 
 * of located type tl,
 * if there is no symbols already in sl with the same name. *)
and add_symbollist_of_dcl_usertypes sl = function
    [] -> sl
  | t::l -> 
      add_symbollist_of_dcl_usertypes (add_symbollist_of_dcl_usertype sl t) l
(* append_symbollist_of_dcl_ids sl (ibl, s, t)
 * returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the symbol list corr. to the given declaration of symbols 
 * DCL_ID (ibl, s, t) *)
and append_symbollist_of_dcl_ids sl (ibl, s, t) =
  (match s.lsc_scope with
       EVA_CST -> ()
     | EVA_PARAM -> ()
     | _  -> raise (Translator_error 
                      ("wrong qualifier in first order symbol declaration", 
                       s.lsc_loc)));
  let tsl = add_symbollist_of_dcl_usertype sl t
  in (append_symbollist_of_dcl_ids1 tsl s t ibl)
and append_symbollist_of_dcl_ids1 sl s t = function
    [] -> sl
  | (i,f)::l -> 
      try
        (append_symbol 
           (new_first_order_symbol (string_of_dcllid i) 
              s.lsc_scope t.lty_type f)
           (append_symbollist_of_dcl_ids1 sl s t l))
      with
          (Symbols_already n) -> 
            raise (Translator_error 
                     (("redeclaration of the first order symbol "^n), 
                      i.lid_loc))
(* returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the singleton list corr. to
 * the given delaration of function symbol DCL_FUN *)
and append_symbollist_of_dcl_fun sl (i, s, t, tl, bh, bs) = 
  begin
    (match s.lsc_scope with
         EVA_CST -> ()
       | EVA_PARAM -> ()
       | _  -> raise (Translator_error 
                        ("wrong qualifier in function symbol declaration", 
                         s.lsc_loc)));
    (match t.lty_type with
         EVA_VOID -> raise (Translator_error 
                              ("empty type in function symbol declaration", 
                               t.lty_loc))
       | _ -> ());
    (try
       let tsl = add_symbollist_of_dcl_usertypes sl (t::tl)
       in (append_symbol 
             (new_functional_symbol (string_of_dcllid i) s.lsc_scope t.lty_type 
                (map (fun t -> t.lty_type) tl) bh bs)
             tsl)
     with
         (Symbols_already n) -> 
           raise (Translator_error 
                    (("redeclaration of the functional symbol "^n), 
                     i.lid_loc)) 
    );
  end
(* returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the singleton list of the symbol 
 * corr. to the given alias declaration EVA_DCL_ALIAS *)
and append_symbollist_of_dcl_alias sl (i, t) =
  try
    (append_symbol (new_alias_symbol (string_of_dcllid i)  t) sl)
  with
      (Symbols_already n) -> 
        raise (Translator_error 
                 (("redeclaration of the alias symbol "^n), 
                  i.lid_loc))
(* returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the singleton list of the symbol 
 * corr. to the given located basetype declaration *)
and append_symbollist_of_ldcl_basetype sl d =
  try 
    append_symbol (symbol_of_ldcl_basetype d) sl
  with 
      (Symbols_already n) -> 
        raise (Translator_error 
                 (("redeclaration of the basetype symbol "^n), 
                  d.ldc_loc))
(* returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the 2 elements list of symbols corr. to
 * the key symbols declared in the given EVA_DCL_KEYPAIR *)
and append_symbollist_of_dcl_keypair sl (a, k1, k2, s, t, tl) =
  begin
    (match s.lsc_scope with
         EVA_CST -> ()
       | EVA_PARAM -> ()
       | _ -> raise (Translator_error 
                       ("wrong qualifier in keypair declaration", s.lsc_loc)));
    (match t.lty_type with
         EVA_VOID -> raise (Translator_error 
                              ("empty type in keypair declaration", 
                               t.lty_loc))
       | _ -> ());    
    let puname = (string_of_dcllid k1)
    and prname = (string_of_dcllid k2) in 
    let tsl = add_symbollist_of_dcl_usertypes sl (t::tl) in
    let sl1 = 
      (try
         (append_symbol
            (new_key_symbol prname s.lsc_scope t.lty_type 
               (map (fun t -> t.lty_type) tl) puname a)
            tsl)
       with
           (Symbols_already n) -> 
             raise (Translator_error 
                      (("redeclaration of the key symbol "^n), 
                       (k2.lid_loc))))
    in try
        (append_symbol
           (new_key_symbol puname s.lsc_scope t.lty_type 
              (map (fun t -> t.lty_type) tl) prname a)
           sl1)
      with
          (Symbols_already n) -> 
            raise (Translator_error 
                     (("redeclaration of the key symbol "^n), 
                      (k1.lid_loc)))
  end
(* returns the symbol associed to the given located value v and located type t *)
and symbol_of_dcl_value v t =
  match v.lva_value with
      (* special reserved name for the intruder *)
      EVA_INTRUDER -> raise (Translator_error
                               (("redeclaration of the identifier I"^
                                 " (reserved for the intruder)"),
                                v.lva_loc)) (* intruder_symbol *)
    | EVA_VALUE s -> (new_value_symbol s t.lty_type)
and symbol_of_ldcl_basetype b = 
  match b.ldc_dcl with
      EVA_DCL_BASETYPE "I" -> 
        raise (Translator_error
                 ("the identifier I is reserved for the intruder", b.ldc_loc))
    | EVA_DCL_BASETYPE s -> 
        new_type_symbol s true
    | _ ->
        raise (Internal_error "symbol_of_ldcl_basetype")
(* returns the symbol list obtained by
 * concatenation of the given symbol list sl and
 * the list of symbols corresponding to the given
 * value declarations EVA_DCL_VALUE *)
and append_symbollist_of_dcl_values sl (vl, t) =
  let tsl = add_symbollist_of_dcl_usertype sl t 
  in append_symbollist_of_dcl_valuesd tsl (vl, t) 
and append_symbollist_of_dcl_valuesd sl (vl, t) =
  match vl with
      [] -> sl
    | (v::l) -> 
        try
          (append_symbol 
             (symbol_of_dcl_value v t) 
             (append_symbollist_of_dcl_valuesd sl (l, t)))
        with
            (Symbols_already n) -> 
              raise (Translator_error 
                       (("double declaration of the value symbol "^n), 
                        v.lva_loc));;


(* returns the symbol list 
 * corr. to the identifiers declared in the given 
 * list of located declarations 
 *
 * the symbols are added in the reversed ordering w.r.t. declarations *)
let rec symbollist_of_ldcllist = function
    [] -> empty_symbol_list
  | (d::dl) -> 
      try
        (trace "Translator.symbollist_of_ldcllist [%s]\n" (string_of_location d.ldc_loc));        (*   (Evapp.evapp_ldcl d); *)
        (append_symbollist_of_ldcl (symbollist_of_ldcllist dl) d)
      with
          (* double declaration should have been already catched *)
          (Symbols_already n) -> raise 
            (Internal_error 
               ("symbollist_of_ldcllist: double decl not catched, symbol "^n));;


(* returns the symbol list 
 * with identifiers declared in the given protocol
 * (i.e. list of private variables of processes)
 *
 * There is no check of double declaration.
 * It is assumed that private var are pairwise distinct
 * and distinct from other ids of the protocol spec. *)
let rec symbollist_of_protocol = function
    EVA_MP _ -> empty_symbol_list (* no private variables *)
  | EVA_PROCESS pl -> (symbollist_of_processlist pl)
and symbollist_of_lprotocol { lpr_protocol = p } =
  symbollist_of_protocol p
and symbollist_of_processlist = function
    [] -> empty_symbol_list
  | p::l -> (
      try 
        (append_symbollist (symbollist_of_lidlist p.pr_privates) 
           (symbollist_of_processlist l))
      with 
          (Symbols_already n) -> 
            raise (Internal_error 
                     ("symbollist_of_processlist: Symbols_already "^n)))
and symbollist_of_lidlist = function
    [] -> empty_symbol_list
  | i::l -> (append_symbol (symbol_of_lid i) (symbollist_of_lidlist l))
and symbol_of_lid { lid_id = EVA_ID s } = 
  (new_first_order_symbol s EVA_PRIVATE EVA_NUMBER false);;      



(* returns the symbol list 
 * corr. to the identifiers in the given list of quantified variables
 * (list of pairs (located identifier, located type))
 *
 * the symbols are added in the reversed ordering w.r.t. the list *)
let rec symbollist_of_quantification = function
    [] -> empty_symbol_list
  | ((i,t)::l) -> 
      try
        (append_symbol 
           (new_quantified_symbol (string_of_dcllid i) t.lty_type) 
           (symbollist_of_quantification l))
      with
          (* double declaration should have been already catched *)
          (Symbols_already n) -> raise 
            (Translator_error (("variable "^n^
                                " occurs twice in quantification"), 
                               i.lid_loc));;





(* [symbolist_of_lstatementlist sl stl]
 * returns the symbol list with the user predicate identifiers
 * in the given list of statements [stl].
 *
 * The symbols are added in the reversed ordering w.r.t. the list. 
 * 
 * raises Translator_error is the given symbol list [sl] contains a symbol
 * with the name of one of the predicate symbols in the given statement list [stl].*)
let rec symbolist_of_lstatementlist sl = function
    [] -> empty_symbol_list
  | (s::l) -> 
      let sl2 = (symbolist_of_lstatement sl s)
      and sll = (symbolist_of_lstatementlist sl l) 
      in (fold_right_symbollist (fun x y -> add_symbol x y) sl2 sll)        
(* [symbolist_of_lstatement sl s]
 * returns the symbol list with the user predicate identifiers 
 * in the given statement [s].
 *
 * the symbols are added in the reversed ordering w.r.t. their
 * order in the statement. 
 * 
 * raises Translator_error is the given symbol list [sl] contains a symbol
 * with the name of one of the predicate symbols in the given statement [s].*)
and symbolist_of_lstatement sl s =  
  match s.lst_statement with 
      EVA_ASSUME (q, al, a) ->
        let sl2 = (symbollist_of_latomlist sl (a::al))
                   (* checks that quantified vars ids are not used as predicates *)
        in (check_predicates_quantifiers sl2 q)
    | EVA_CLAIM (_, al, a) ->
        (symbollist_of_latomlist sl (a::al))
(* [symbollist_of_latomlist sl al]
 * returns the symbol list with the user predicate identifiers 
 * in the given list of atoms [al], and checks for each predicate id
 * that there is no symbol in the given symbol list [sl] with the same name. 
 *
 * the symbols are added in the reversed ordering w.r.t. the list.
 * 
 * raises Translator_error is the given symbol list [sl] contains a symbol
 * with the name of one of the predicate symbols in the given atom list [al].*)
and symbollist_of_latomlist sl = function
    [] -> empty_symbol_list
  | (a::l) -> (
      match a.lat_atom with
          EVA_ATOM_USER (i, _) -> 
            let n = (string_of_lid i)
            in 
              if (memassoc_symbol n sl)
              then 
                raise (Translator_error
                         (("the predicate name "^n^" is already declared"),
                          i.lid_loc))
              else  
                (add_symbol 
                   (new_predicate_symbol n)
                   (symbollist_of_latomlist sl l))
        | _ -> (symbollist_of_latomlist sl l)
    )
and check_predicates_quantifiers sl = function
    [] -> sl
  | ((i,_)::l) -> 
      let n = string_of_lid i in
        if (memassoc_symbol n sl)
        then raise 
          (Translator_error (("the predicate name "^n^
                              " is also used as a quantified variable"),
                             i.lid_loc))
        else check_predicates_quantifiers sl l;;
                      


(* returns the symbol list extracted from:
 * - the declarations (field Abstract.eva_spec.sp_declarations)
 *   of the given eva_spec
 * - the value declarations (field Abstract.eva_spec.sp_valuedcl)
 *   of the given eva_spec.
 *
 * Note that the other identifiers (quantified identifiers in axioms or
 * formulas and private process variables) are not included in the symbol list.
 *)
let symbollist_of_spec = function 
    {
      sp_label = _;
      sp_declarations = dl;
      sp_protocol = p;
      sp_valuedcl = vdl;
      sp_sessions = _;
      sp_statements = stl;
    } -> 
      let sl = (symbollist_of_ldcllist ((rev vdl)@(rev dl))) in
      (* add the private variables *)
      let sl = (
        try 
          (append_symbollist  (symbollist_of_lprotocol p) sl)
        with 
            (Symbols_already n) -> 
              raise 
              (Internal_error 
                 ("symbollist_of_spec: private variable "^n^" is already declared"))
      ) in
      let psl = (symbolist_of_lstatementlist sl stl) in
      let sl = (
        try
          (append_symbollist psl sl)
        with 
            (Symbols_already n) -> 
              raise 
              (Internal_error 
                 ("symbollist_of_spec: redeclared predicate symbol "^n^
                  " not catched by symbollist_of_lprotocol")))
      (* add the reserved intruder symbol *)
      in 
        (try
           (append_symbol intruder_symbol sl)
         with
             (Symbols_already n) -> 
              raise 
               (Internal_error 
                  ("symbollist_of_spec: redeclaration of intruder symbol not catched "^n)));;
             

 
(**************************************************************
 * Conversion of located terms
 * 
 * This functions check and convert the symbols in terms 
 * obtained from parsing.
 **************************************************************)

let rec convert_lterm sl lterm =
  match lterm.lte_term with
      EVA_TERM_ID i -> (
        try
          let sy = (assoc_symbol i sl) in
            (* declared first order variable or function symbol
               or member of declared keypair *)
            if ((is_first_order sy) || (is_functional sy) || (is_key sy))
            then (
              match (get_scope sy) with
                  EVA_CST -> 
                    { lterm with lte_term = EVA_TERM_CST (get_name sy) }
                | EVA_PARAM 
                | EVA_PRIVATE ->
                    { lterm with lte_term = EVA_TERM_ID (get_name sy) })
            (* declared value symbol *)
            else if (is_value sy) 
            then { lterm with lte_term = 
                     EVA_TERM_VALUE (EVA_VALUE (get_name sy)) }
            (* declared value symbol *)
            else if (is_intruder sy) 
            then { lterm with lte_term = EVA_TERM_VALUE EVA_INTRUDER }   
            (* alias symbol *)
            else if (is_alias sy)
            then { lterm with lte_term = 
                     EVA_TERM_ALIAS ((get_name sy), (get_term sy)) }   
            (* quantified variable *)
            else if (is_quantified sy)
            then { lterm with lte_term =
                     EVA_QUANTIFIED ((get_name sy), (get_type sy)) }
            (* user type *)
            else if (is_type sy)
            then raise (Translator_error ("type symbol in term", lterm.lte_loc))

            else raise (Translator_error ("term badly formed", lterm.lte_loc))
        with
            Symbols_not_found -> 
              raise (Translator_error (("undeclared symbol "^i), lterm.lte_loc))  
      )        
      (* there should not be coerce symbol here *)
    | EVA_P t ->
        { lterm with lte_term = EVA_P (convert_lterm sl t) }         
    | EVA_A t ->
        { lterm with lte_term = EVA_A (convert_lterm sl t) }         
    | EVA_SA t ->
        { lterm with lte_term = EVA_SA (convert_lterm sl t) }         
    | EVA_AA t ->
        { lterm with lte_term = EVA_AA (convert_lterm sl t) }         
    | EVA_U (i, t) ->
        { lterm with lte_term = EVA_U (i, (convert_lterm sl t)) }         
    | EVA_CONS tl -> 
        { lterm with lte_term = EVA_CONS (map (convert_lterm sl) tl) }   
    (* app_root = f; app_args = tl *)
    | EVA_APP (f, tl) ->
        { lterm with lte_term = EVA_APP (f, (map (convert_lterm sl) tl)) }   
    (* crypt_algo = a; crypt_key = k; crypt_content = t *)
    | EVA_CRYPT (a, k, t) ->
        { lterm with lte_term = 
            EVA_CRYPT (a, (convert_lterm sl k), (convert_lterm sl t)) }
    (* crypt_algo = a; crypt_key = k; crypt_content = t *)
    | EVA_SIGN (a, k ,t) -> 
        { lterm with lte_term = 
            EVA_SIGN (a, (convert_lterm sl k), (convert_lterm sl t)) }
    | EVA_PCENT (t1, t2) ->
        { lterm with lte_term = 
            EVA_PCENT ((convert_lterm sl t1), (convert_lterm sl t2)) }
    | EVA_TERM_ALIAS (a, t) -> 
        { lterm with lte_term = 
            EVA_TERM_ALIAS (a, (convert_lterm sl t)) }
    | _ -> lterm;;


let rec convert_ldcl sl d =
  match d.ldc_dcl with
      EVA_DCL_KNOW (i, kl) -> 
        { d with ldc_dcl = EVA_DCL_KNOW (i, (map (convert_lterm sl) kl)) }
    | EVA_DCL_AXIOM (t1, t2, _, ql) ->
        let slq = (overwrite_symbollist (symbollist_of_quantification ql) sl) in
        let tc1 = (convert_lterm slq t1)
        and tc2 = (convert_lterm slq t2) in          
        (* cercion is not applied at this point, hence tc1 and tc2 are not well typed,
         * use the partial typing function *)
        let type1 = (type_lterm_laeva slq tc1)
        and type2 = (type_lterm_laeva slq tc2) in
          (* structural equality *)
          if (Types.equals type1 type2)
          then { d with ldc_dcl = EVA_DCL_AXIOM (tc1, tc2, type1, ql) }
          else raise (Translator_error ("members of axiom have different types",
                                        d.ldc_loc))
    | _ -> (lterm_replace_ldcl (convert_lterm sl) d);;


let convert_lstatement sl st = 
  match st.lst_statement with
      EVA_ASSUME (ql, _, _) -> 
        lterm_replace_lstatement 
        (convert_lterm 
           (overwrite_symbollist (symbollist_of_quantification ql) sl))
        st
    | EVA_CLAIM (_, _, _) -> 
        lterm_replace_lstatement (convert_lterm sl) st
    
(* convert_spec sl sp
 * applies the convertion functions convert_lterm, convert_ldcl, 
 * convert_lstatement
 * to the elements of the given sbatract specification sp, 
 * with the given symbol list sl, and return the specification obtained *)
let rec convert_spec sl sp = 
  (*   lterm_replace_spec (convert_lterm sl) sp;; *)
  let tr = (convert_lterm sl) 
  in {  sp with 
          sp_declarations = (map (convert_ldcl sl) sp.sp_declarations);
          sp_protocol = lterm_replace_lprotocol tr sp.sp_protocol;
          sp_sessions = (map (lterm_replace_lsession tr) sp.sp_sessions);
          sp_statements = (map (convert_lstatement sl) sp.sp_statements);
     };; 


(*********************************************************************
 * Checking symbols at leaves positions in terms w.r.t. given rights
 *********************************************************************)

(* flag for term rights meaning that value symbols are allowed 
 * in terms (see check_lterm) *)
let term_VALUE   = 0x00000001;;

(* flag for term rights meaning that constant symbols are allowed 
 * in terms (see check_lterm) *)
let term_CST     = 0x00000002;;

(* flag for term rights meaning that parameters symbols are allowed 
 * in terms (see check_lterm) *)
let term_PARAM   = 0x00000100;;

(* flag for term rights meaning that private variables are allowed 
 * in terms (see check_lterm) *)
let term_PRIV    = 0x00000200;;

(* flag for term rights meaning that percent (Lowe's notation) 
 * are allowed in terms (see check_lterm) *)
let term_PCENT   = 0x00000004;;

(* flag for term rights meaning that located variables
 * are allowed in terms (see check_lterm) *)
let term_AT      = 0x00000008;;

(* flag for term rights meaning that first order symbols 
 * are allowed in terms (see check_lterm) *)
let term_FO      = 0x00000010;;

(* flag for term rights meaning that functional symbols
 * are allowed in terms (see check_lterm) *)
let term_FUN     = 0x00000020;;

(* flag for term rights meaning that keypairs symbols
 * are allowed in terms (see check_lterm) *)
let term_KEY     = 0x00000040;;

(* flag for term rights meaning that alias symbols
 * are allowed in terms (see check_lterm) *)
let term_ALIAS   = 0x00000080;;

(* flag for term rights meaning that quantified variables
 * are allowed in terms (see check_lterm) *)
let term_QUANTIF = 0x00000100;;

(* sum of all flags (see check_lterm) *)
let term_ALL = term_VALUE land  
               term_CST land
               term_PARAM land
               term_PRIV  land
               term_PCENT land
               term_AT land
               term_FO land
               term_FUN land
               term_KEY land
               term_ALIAS land
               term_QUANTIF;;

let check_flag flag right =
  (flag land right) != 0;;



(* returns true is the given located identifier has been declared
 * as first order symbol with type principal. Returns false otherwise.
 *
 * raise Translator_error if the given located identifier
 * has not been declared or has not been declared as a first order symbol.
 *)
let is_principal sl i =   
  let name = string_of_lid i in
    try
      let sy = (assoc_symbol name sl) in
        if (is_first_order sy)
        then ((get_type sy) = EVA_PRINCIPAL)
        else raise (Translator_error ((name^" is not a first order symbol"), 
                                      i.lid_loc))
    with
        Symbols_not_found ->
          raise (Translator_error (("undeclared symbol "^name), i.lid_loc))
        (* symbol has no type *)
      | Symbols_error _ ->
          raise (Internal_error "is_principal");;


(* check_ltype sl t 
 * returns the given located ttype t if it is well formed w.r.t.
 * the given symbol list sl.
 *
 * raises an exception Translator_error otherwise *)
let rec check_ltype sl t = 
  match t.lty_type with
      EVA_USERTYPE s ->
        ( try
            let ts = (assoc_symbol s sl)  in
              if ((is_type ts) && (not (is_basetype ts)))
              then t
              else raise (Translator_error 
                            ("type badly formed, user type symbol expected", 
                             t.lty_loc))
          with
              Symbols_not_found -> 
                raise (Translator_error (("unknown type symbol "^s), 
                                         t.lty_loc)))
    | EVA_BASETYPE s ->
        ( try
            let ts = (assoc_symbol s sl)  in
              if (is_basetype ts)
              then t
              else raise (Translator_error 
                            ("type badly formed, basetype expected", 
                             t.lty_loc))
          with
              Symbols_not_found -> 
                raise (Translator_error (("unknown basetype symbol "^s), 
                                         t.lty_loc)))
    | _ -> t;;


(* check_lterm rl sl lterm 
 * returns the given located term lterm
 * if its leaves are correct w.r.t.
 * the given rights list rl and the given symbol list sl.
 * The right list is a list of flags described below
 * defining which symbols are allowed in terms. 
 *
 * raises an exception Translator_error otherwise *)
let rec check_lterm rl sl lterm =
  (* sum the flags in the given rightlist *)
  let right = (fold_left (fun x y -> x lor y) 0 rl) 
  in match lterm.lte_term with
      EVA_TERM_ID i -> (
        try
          let sy = (assoc_symbol i sl) in
            (* must be first order id in sl *)
            if (is_first_order sy)
            then 
              if (check_flag term_FO right)
              then  
                check_id right sl lterm i (get_scope sy)
              else 
                raise (Translator_error 
                         ("first order symbols not allowed in this term",
                          lterm.lte_loc))
            else if (is_functional sy)
            then 
              if (check_flag term_FUN right)
              then
                check_id right sl lterm i (get_scope sy)
              else
                raise (Translator_error 
                         ("function symbols not allowed in this term",
                          lterm.lte_loc))
            else if (is_key sy)
            then 
              if (check_flag term_FUN right)
              then 
                check_id right sl lterm i (get_scope sy)                
              else 
                raise (Translator_error 
                         ("key symbols not allowed in this term",
                          lterm.lte_loc))
            else raise (Internal_error
                          ("check_lterm: "^i^
                           " first order or fun or key id expected"))
        with
            Symbols_not_found -> 
              raise (Translator_error (("undeclared symbol "^i),
                                       lterm.lte_loc)))
    | EVA_TERM_CST i -> (
        try
          let sy = (assoc_symbol i sl) in
            (* must be first order id in sl *)
            if (is_first_order sy) 
            then 
              if (check_flag term_FO right)
              then  
                check_cst right sl lterm i (get_scope sy)
              else 
                raise (Translator_error 
                         ("first order symbols not allowed in this term",
                          lterm.lte_loc))               
            else if (is_functional sy)
            then 
              if (check_flag term_FUN right)
              then
                check_cst right sl lterm i (get_scope sy)
              else 
                raise (Translator_error 
                         ("function symbols not allowed in this term",
                          lterm.lte_loc))               
            else if (is_key sy)
            then 
              if (check_flag term_KEY right)
              then 
                check_cst right sl lterm i (get_scope sy)
              else 
                raise (Translator_error 
                         ("key symbols not allowed in this term",
                          lterm.lte_loc))
            else
              raise (Internal_error
                       ("check_lterm: "^i^" first order of fun or key id expected"))
        with
            Symbols_not_found -> 
              raise (Translator_error (("undeclared symbol "^i),
                                       lterm.lte_loc)))
    | EVA_TERM_VALUE v -> 
        if (not (check_flag term_VALUE right))
        then  raise (Translator_error 
                       ("value symbols not allowed in this term", 
                        lterm.lte_loc))
        else (
          match v with
              EVA_INTRUDER -> lterm
            | (EVA_VALUE i) -> (
                try
                  let sy = (assoc_symbol i sl) in 
                    if (is_value sy)
                    then lterm
                    else raise (Internal_error
                                  ("check_lterm: "^i^" non value in TERM_VALUE")) 
                with
                    Symbols_not_found -> 
                      raise (Translator_error (("undeclared symbol "^i),
                                               lterm.lte_loc))
              )
        )
     (* checks recursively the aliased term *)
    | EVA_TERM_ALIAS (a, t) -> 
        if (not (check_flag term_ALIAS right))
        then raise (Translator_error 
                      ("aliases not allowed in this term", lterm.lte_loc))
        else (
          try 
            let sy = (assoc_symbol a sl) in
              if (is_alias sy) 
              then 
                { lterm with lte_term = EVA_TERM_ALIAS (a, (check_lterm rl sl t)) }
              else 
                raise (Internal_error
                         ("check_lterm: "^a^" non alias in TERM_ALIAS")) 
          with
              Symbols_not_found -> 
                raise (Translator_error (("undeclared symbol "^a), 
                                         lterm.lte_loc))
        )
    | EVA_LOCATED (v, r) -> 
        (* check rights for located var *)
        if (not (check_flag term_AT right))
        then raise (Translator_error 
                      ("located variables not allowed in this term", 
                       lterm.lte_loc))         
        else 
          (* r declared as first order of type principal *)
          if (not (is_principal sl r))
          then
            raise (Translator_error 
                     (("the role was not declared with type principal"^ 
                       " in this located variable"),
                      lterm.lte_loc))         
          else ( 
            let n = (string_of_lid v) in
              try
                (* check that v is declared *)
                let syv = (assoc_symbol n sl) in 
                  if ((is_first_order syv) || (is_functional syv)  || (is_key syv))
                  then lterm
                  else raise (Translator_error 
                                ("located variable "^
                                 n^" is not f.o. or fun or key symbol", 
                                 lterm.lte_loc))
              with                        
                  Symbols_not_found -> 
                    raise (Translator_error (("undeclared symbol "^n), 
                                             v.lid_loc)) )
    | EVA_QUANTIFIED (i, _) -> 
        if (not (check_flag term_QUANTIF right))
        then 
          raise (Translator_error 
                   ("quantified variables not allowed in this term",
                    lterm.lte_loc))
        else (
          try 
            let sy = (assoc_symbol i sl) in
              if (is_quantified sy)
              then lterm 
              else 
                raise (Internal_error
                         ("check_lterm: "^i^" non quantified in QUANTIFIED"))
          with
              Symbols_not_found -> 
                raise (Translator_error (("undeclared symbol "^i), lterm.lte_loc))
        )
    | EVA_P t ->
        { lterm with lte_term = EVA_P (check_lterm rl sl t) }
    | EVA_A t ->
        { lterm with lte_term = EVA_A (check_lterm rl sl t) }
    | EVA_AA t ->
        { lterm with lte_term = EVA_AA (check_lterm rl sl t) }
    | EVA_SA t ->
        { lterm with lte_term = EVA_SA (check_lterm rl sl t) }
    | EVA_U (i, t) -> (
        try
        let sy = (assoc_symbol i sl) in
          if (is_type sy)
          then { lterm with lte_term = EVA_U (i, (check_lterm rl sl t)) }
          else raise (Internal_error "check_lterm (coercion U)")
        with
            Symbols_not_found -> 
              raise (Internal_error 
                       "check_lterm (coercion U): type symbol undeclared"))
    | EVA_CONS tl -> 
        { lterm with lte_term = EVA_CONS (map (check_lterm rl sl) tl) }
    | EVA_APP (i, tl) -> (
        let s = (string_of_lid i) in
          try
            let sy = (assoc_symbol s sl) in
              if ((is_functional sy)  || (is_key sy))
              then ( 
                if (((get_scope sy) = EVA_CST) &&
                    (not (check_flag term_CST right)))
                then raise (Translator_error 
                              ("constant symbols not allowed in this term", 
                               lterm.lte_loc))
                else if (((get_scope sy) = EVA_PARAM) &&
                         (not (check_flag term_PARAM right)))
                then raise (Translator_error 
                              ("parameters not allowed in this term", 
                               lterm.lte_loc))
                else if (((get_scope sy) = EVA_PRIVATE) &&
                         (not (check_flag term_PRIV right)))
                then raise (Translator_error 
                              ("private variables not allowed in this term", 
                               lterm.lte_loc))
                else if (not (check_flag term_FUN right))
                then raise (Translator_error 
                              ("function symbols not allowed in this term",
                               lterm.lte_loc)) 
                else { lterm with 
                         lte_term = EVA_APP (i, (map (check_lterm rl sl) tl)) })
                (* note: the syntax does not allow an alias 
                   to a functional symbol
                   i.e. no TERM_ID, TERM_CST i with i symbol of function type *)
              else raise (Internal_error
                            ("check_lterm: APP "^s^" not functional"))
          with
              Symbols_not_found -> 
                raise (Translator_error (("undeclared symbol "^s), lterm.lte_loc)))
    | EVA_CRYPT (a, k, t) ->
        let ca = 
          match a with
              EVA_VANILLA -> EVA_VANILLA 
            | EVA_ALGO at -> EVA_ALGO (check_lterm rl sl at)
        in { lterm with lte_term = 
               EVA_CRYPT (ca, (check_lterm rl sl k), (check_lterm rl sl t)) }
    | EVA_SIGN (a, k ,t) -> 
        let ca = 
          match a with
              EVA_VANILLA -> EVA_VANILLA 
            | EVA_ALGO at -> EVA_ALGO (check_lterm rl sl at)
        in { lterm with lte_term = 
               EVA_SIGN (ca, (check_lterm rl sl k), (check_lterm rl sl t)) }
    | EVA_PCENT (t1, t2) ->
        if (check_flag term_PCENT right)
        then { lterm with lte_term = 
                 EVA_PCENT ((check_lterm rl sl t1), (check_lterm rl sl t2)) }
        else raise (Translator_error 
                      ("% notation not allowed in this term", lterm.lte_loc))
and check_id right sl lterm i = function
    EVA_PARAM -> 
      if (check_flag term_PARAM right)
      then lterm 
      else raise (Translator_error 
                    ("parameters not allowed in this term",
                     lterm.lte_loc))
  | EVA_PRIVATE -> 
      if (check_flag term_PRIV right)
      then lterm 
      else raise (Translator_error 
                    ("private variables not allowed in this term",
                     lterm.lte_loc))
  | EVA_CST ->
      raise (Internal_error
               ("check_lterm: "^i^" constant in TERM_ID"))
and check_cst right sl lterm i = function
    EVA_CST -> 
      if (check_flag term_CST right)
      then lterm 
      else raise (Translator_error 
                    ("constant symbols not allowed in this term",
                         lterm.lte_loc))
  | _ ->
      raise (Internal_error
               ("check_lterm: "^i^" param or priv in TERM_CST"))
;;
        

        


(***********************************************************************
 * Checking extra restrictions in the declarations
 ***********************************************************************)

(** @return the given located declaration if 
  it is not an axiom declaration 
  ot if it is an axiom declaration and the
  terms of the axioms contain only
  quantified variables or constants.
 
  @raise Translator_error otherwise *)
let rec check_ldcl_axiom sl a =
  trace "Translator.check_ldcl_axiom\n";
  let arights = [term_FO; term_FUN; term_ALIAS; term_CST; term_QUANTIF] in
    match a.ldc_dcl with
        EVA_DCL_AXIOM (_, _, _, ql) ->
          (* symbol list overwritten with the list of quantified variables *)
          let qsl = (overwrite_symbollist (symbollist_of_quantification ql) sl)
                      (* in trace "overwriten symbol list:\n"; dump_symbollist qsl; *)
          in lterm_replace_ldcl (check_lterm arights qsl) a
      | _ -> a;;


(* @return the given located knowledge declaration if 
 * the role is declared as first order principal.
 *
 * @raise Translator_error otherwise *)
let rec check_ldcl_knows sl d =
  trace "Translator.check_ldcl_knows\n";
  match d.ldc_dcl with
      EVA_DCL_KNOW (i, _) ->
        if (not (is_principal sl i))
        then 
          raise (Translator_error (((string_of_lid i)^
                                    " is not declared with type principal"),
                                   d.ldc_loc))
        else d
    | _ -> d;;

(* @return the given located declaration if 
 * it is not a first order ids declaration or
 * if it is a first order ids declaration 
 * with scope parameter, 
 * if it is a first order ids declaration 
 * with scope constant and no id declared as fresh. 
 *
 * @raise Translator_error otherwise *)
let rec check_ldcl_fresh d =
  trace "Translator.check_ldcl_fresh\n";
  match d.ldc_dcl with
      EVA_DCL_ID (il ,sc ,_ ) ->
        if (sc.lsc_scope = EVA_CST)
        then 
          if (List.for_all (fun (x,y) -> (not y)) il)
          then d
          else  
            raise (Translator_error (("constant symbols "^
                                      "cannot be declared fresh"), d.ldc_loc))
        else d
    | _ -> d;;


(***********************************************************************
 * Checking extra restrictions in the protocol messages / instructions
 * (check type of ids)
 ***********************************************************************)

(* Returns the given located instruction if 
 * - the identifiers sender and receiver of messages or
 *   roles of comp and assignment are typed as principal
 * - the bodys of messages are typed as numbers
 * - the terms of assign and comp are typed as numbers
 * - all the leaves of the located terms in the given located instruction
 * are either first order or functional or key symbols declared as 
 * constant or parameter, or aliases.
 *
 * raises an exception Translator_error otherwise *)
let check_linstr sl linstr = 
  let irights = [term_CST; term_PARAM; term_FO; term_FUN; term_KEY; 
                 term_ALIAS; term_PCENT] in
  match linstr.lin_instr with
      EVA_SKIP -> 
        linstr
    |  EVA_MSG (_, s, r,t) ->
         if (not (is_principal sl s))
         then 
           raise (Translator_error (("the sender "^(string_of_lid s)^
                                     " is not declared with type principal"),
                                    linstr.lin_loc))
         else if (not (is_principal sl r))
         then raise (Translator_error (("the receiver "^(string_of_lid s)^
                                        " is not declared with type principal"), 
                                       linstr.lin_loc))
         else if ((type_lterm sl t) <> EVA_NUMBER)
         then raise (Internal_error "check_linstr MSG (apply coercion first)")
         else lterm_replace_linstr (check_lterm irights sl) linstr
    | EVA_ASSIGN (r, x, t)
    | EVA_COMP (r, x, t) ->
        if (not (is_principal sl r))
        then 
          raise (Translator_error (("the role "^(string_of_lid r)^
                                    " is not declared with type principal"), 
                                   linstr.lin_loc))
        else if ((type_lterm sl t) <> EVA_NUMBER)
        then raise (Internal_error "check_linstr (apply coercion first)")
        else lterm_replace_linstr (check_lterm irights sl) linstr
    | _ -> lterm_replace_linstr (check_lterm irights sl) linstr;;
        

(* check all the instructions in the given protocol specification,
 * applying check_linstr with the given symbol list *)
let check_instrs_spec sl sp =
  trace "Translator.check_instrs_spec\n";
  match sp.sp_protocol.lpr_protocol with
      EVA_PROCESS _ -> raise (Internal_error "check_instrs_spec: not BAN spec")
    | EVA_MP il -> 
        let protocol = { sp.sp_protocol with 
                           lpr_protocol = (EVA_MP (map (check_linstr sl) il)) } 
        in { sp with sp_protocol = protocol };;
      


(**************************************************************
 * Checking extra restrictions in values declarations 
 **************************************************************)

(* check_value_ldcl d
 * returns the given located declaration d
 * if it is a valid value declaration, i.e.it is either:
 * - a declaration of value symbol
 * - a declaration of functional or key symbol with
 *   scope constant.
 *) 
let rec check_value_ldcl d =
  match d.ldc_dcl with
      EVA_DCL_VALUE _ -> d
    | EVA_DCL_FUN (_,s , _, _, _, _)
    | EVA_DCL_KEYPAIR (_, _, _,s , _, _) -> (
        if (s.lsc_scope = EVA_CST)
        then d
        else 
          raise (Translator_error 
                   ("constructors of values must be declared as constants",
                    s.lsc_loc)))
    | _ ->
        raise (Translator_error 
                 ("this kind of declaration is not allowed here", 
                  d.ldc_loc))        


(**************************************************************
 * Checking extra restrictions in sessions declarations
 **************************************************************)

let rec check_lsession sl s = 
  match s.lse_session with
      EVA_BANG -> s
    | EVA_CONSTRAINED_SESSIONS cl ->
        { s with 
            lse_session = EVA_CONSTRAINED_SESSIONS 
                            (List.map (check_lconstraint sl) cl) }
    | EVA_SESSIONS al ->
        { s with 
            lse_session = EVA_SESSIONS
                            (List.map (check_assignment sl) al) }
    | EVA_SESSION (l, al) ->
        { s with 
            lse_session = EVA_SESSION
                            (l, (List.map (check_assignment sl) al)) }
and check_lconstraint sl c =
  match c.lco_constraint with
      (* both ids must be parameters *)
      EVA_EQ_CONSTRAINT (i1, i2) ->
        {c with 
           lco_constraint = EVA_EQ_CONSTRAINT (check_lid_parameter sl i1, 
                                               check_lid_parameter sl i2) }
      (* both ids must be parameters *)
    | EVA_NEQ_CONSTRAINT (i1, i2) ->        
        {c with 
           lco_constraint = EVA_NEQ_CONSTRAINT (check_lid_parameter sl i1, 
                                                check_lid_parameter sl i2) }
    | EVA_MEMBERSHIP_CONSTRAINT (i, tl) ->
        let drights = [term_FO; term_FUN; term_KEY; term_VALUE; term_CST; 
                       term_ALIAS] in
        { c with 
            lco_constraint = 
            EVA_MEMBERSHIP_CONSTRAINT (check_lid_parameter sl i,
                                   List.map (check_lterm drights sl) tl) }
    | EVA_DOMAIN_CONSTRAINT (i, s) ->
        let c2 = 
          { c with 
              lco_constraint = 
              EVA_DOMAIN_CONSTRAINT (check_lid_parameter sl i, s) } in
        (* if i is a symbol of sl, it must be a predicate symbol *)
          (try  
             let sy = (assoc_symbol s sl) in 
               if (is_predicate sy)
               then c2
               else raise 
                 (Translator_error
                    (("the symbol "^s^" connot be used in a domain constraint"),
                     c.lco_loc))
           with
               (* we allow the predicate symbol not to occur in a formula *)
               Symbols_not_found -> c2)
(* check_lid_parameter sl i
 * returns the given located id i
 * if it is declared as a parameter in the given symbol list sl *)
and check_lid_parameter sl i = 
  let n = (string_of_dcllid i) in 
  let sy = (assoc_symbol n sl) in
    try
      if ((get_scope sy) = EVA_PARAM)
      then i
      else 
        raise (Translator_error 
                 (("the id "^n^" should be a parameter", i.lid_loc)))
  with
      Symbols_error msg -> 
        raise (Internal_error ("check_lid_parameter: "^n^" "^msg))
    | Symbols_not_found -> 
        raise (Translator_error (("undeclared symbol "^n), i.lid_loc))
and check_assignment sl (i, t) = 
  (* TBD: check that i and t have same type *)
  let arights = [term_FO; term_FUN; term_KEY; term_VALUE; term_CST; 
                 term_ALIAS] in 
  ((check_lid_parameter sl i), 
   (check_lterm arights sl t));;
        

(**************************************************************
 * Checking extra restrictions in statements (formulas)
 * (check symbols in terms)
 **************************************************************)

let rec check_lstatement sl st =
  trace "Translator.check_lstatement\n";
  let arights = [term_FO; term_FUN; term_KEY; term_ALIAS; term_VALUE; 
                 term_CST; term_QUANTIF]
  and crights = [term_FO; term_FUN; term_KEY; term_ALIAS; term_VALUE; 
                 term_CST; term_AT] in
    match st.lst_statement with
        EVA_ASSUME (ql, _, _) -> 
          (* symbol list overwritten with the list of quantified variables *)
          let qsl = (overwrite_symbollist (symbollist_of_quantification ql) sl)
          (* in trace "overwriten symbol list:\n"; dump_symbollist qsl; *)
          in lterm_replace_lstatement (check_lterm arights qsl) st
      | EVA_CLAIM (_, _, _) -> 
          lterm_replace_lstatement (check_lterm crights sl) st;;


(**************************************************************************
 * Checking protocol spec against 
 * the extra (non syntactical) restrictions described in the EVA grammar
 **************************************************************************)

(* check_spec sl sp
 * checks the extra (non syntactical) restriction described in the EVA grammar
 * in the given (parsed) protocol specification sp.
 * More precisally, it checks:
 * - the symbols contained in every term wrt to different rights specific to each 
 *   spec field (see check_lterm)
 * - the principals components of instruction have type {principal}
 *
 * NB: the spec sp must be obtained from the parser and must have passed 
 * the functions {convert_spec} and {Types.coerce_spec}
 *)
let rec check_spec sl sp =
  trace "Translator.check_spec\n";
  { sp with
      sp_label = sp.sp_label;

      sp_declarations = (
        let drights = [term_FO; term_FUN; term_KEY; term_ALIAS; term_CST; 
                       term_PARAM] in
        (* checks the knows declarations now *)
        let dl = List.map (check_ldcl_knows sl) sp.sp_declarations in 
        let dl = List.map check_ldcl_fresh dl in
        let decls = (get_non_axiom_ldcl dl) 
        and axioms = (get_axiom_ldcl sp.sp_declarations) 
        in (map (lterm_replace_ldcl (check_lterm drights sl)) decls)@
           (map (check_ldcl_axiom sl) axioms) );

      sp_protocol = (
        (* check conditions on instructions components *)
        match sp.sp_protocol.lpr_protocol with
            EVA_PROCESS _ -> 
              raise (Internal_error "check_instrs_spec: not BAN spec")
          | EVA_MP il -> 
              { sp.sp_protocol with 
                  lpr_protocol = (EVA_MP (map (check_linstr sl) il)) } );
      
      sp_valuedcl =  (List.map check_value_ldcl sp.sp_valuedcl);

      sp_sessions = 
              (List.map (check_lsession sl) sp.sp_sessions);                  
      (* different rights for assume and claim *)

      sp_statements = (map (check_lstatement sl) sp.sp_statements);
  }
and get_axiom_ldcl = function
    [] -> []
  | d::l -> (
      match d with 
          { ldc_dcl = EVA_DCL_AXIOM (_, _, _, _) } -> 
            d::(get_axiom_ldcl l)
        | _ -> (get_axiom_ldcl l))
and get_non_axiom_ldcl = function
    [] -> []
  | d::l -> (
      match d with 
          { ldc_dcl = EVA_DCL_AXIOM (_, _, _, _) } -> 
            (get_non_axiom_ldcl l)
        | _ -> d::(get_non_axiom_ldcl l));;
     
    



(**************************************************************
 * Translation of 
 * eva spec obtained by parsing a file in EVA concrete syntax 
 * (with simultaneous processes description)
 * into eva spec with separate processes description
 **************************************************************)

(* returns the inverse key of the given located term,
 * according to the given symbol list.
 * The inverse key of a symmetric key is the key itself.
 * An asymmetric key is the application of a function symbol
 * declared in a keypair to a list of terms.
 * Every other term is considered as a symmetric key.
 * 
 * The given located term must not contain aliases, %, values,
 * located variables, quantified variables or metaterm.
 * raise Internal_error otherwise. *)
let rec inverse_key sl lterm =
  match lterm.lte_term with
      (* f.o. identifier is a symmetric key *)
      EVA_TERM_ID _ 
    | EVA_TERM_CST _ -> 
        lterm
      (* a principal's name should not be an asymmetric key... *)
    | EVA_P t ->
        { lterm with 
            lte_term = EVA_P (inverse_key sl t) }
      (* neither an algo *)
    | EVA_A t ->
        { lterm with 
            lte_term = EVA_A (inverse_key sl t) }
      (* type EVA_TALGO, cannot be used as a key *) 
    | EVA_SA t 
    | EVA_AA t ->
        raise (Internal_error "inverse_key: coercion SA or AA")
    | EVA_U (i, t) ->
        { lterm with 
            lte_term = EVA_U (i, (inverse_key sl t)) }    
      (* a singleton tuple can contain symmetric or assymetric key,
       * descend recursively *)
    | EVA_CONS [k] -> 
        { lterm with 
            lte_term = EVA_CONS [ (inverse_key sl k) ] }
      (* a tuple with more than 1 element is a (composed) symmetric key *)
    | EVA_CONS _ -> 
        lterm
      (* composed symmetric or asymmetric key *)
    | EVA_APP (i, tl) -> (
        let name = string_of_lid i in
        try 
          let ks = (assoc_symbol name sl) in 
            (* check that i is the identifier of a asymmetric key *)
            if (is_key ks)
            then 
              { lterm with 
                  lte_term = 
                  EVA_APP ({i with lid_id = (EVA_ID (get_symkey ks))}, tl) }
            else 
              lterm
        with
            Symbols_not_found -> raise 
              (Translator_error (("undeclared symbol "^name), i.lid_loc))
      )
      (* composed symmetric key *)
    | EVA_CRYPT (_,_,_) ->
        lterm
    | EVA_SIGN (_,_,_) ->
        lterm
    | EVA_TERM_ALIAS (_,t) -> 
        inverse_key sl t
    | EVA_PCENT (_,_) -> 
        raise (Translator_error ("% not allowed inside terms", lterm.lte_loc))
    (* DEBUG: trace error cases *)
    | EVA_TERM_VALUE _ -> raise (Internal_error "Translator.inverse_key (value)")
    | EVA_LOCATED (_,_) -> raise (Internal_error "Translator.inverse_key (located)")
    | EVA_QUANTIFIED (_,_) -> raise (Internal_error "Translator.inverse_key (qtf)");;
 
        

(* The knowledge of a principal
 * is stored as a list of pairs (term, int)
 * - the leaves of the term are parameters or constant.
 * - the int, if = 0, has no significance.
 * - the int, if > 0, is the index of the private variable
 *    used to store the cipher term learned (by pattern matching)
 *    from a message received by the principal, 
 *    (but whose contents are unknown to the principal). 
 * Note that in the last case, the term is always a ciphertext 
 * or (one way) function application.
 *
 * invariant: the list contains no repetition (of term component).
 *)

(* returns the list of principal knowledge formatted as
 * described above (for the function process)
 * from the list of principal knowledge of the specs declarations *)
let rec bound_of_know tl = 
  List.map (fun x -> (x, 0)) tl;;


(* returns the identifier of the private variable
 * whose index is i (must be > 0) and attached to 
 * principal of the given located id *)
let private_variable p i =
  if i > 0
  then "_X"^(string_of_lid p)^(string_of_int i)
  else raise (Internal_error "bound_variable");;


(* returns the max index +1 of the free variables *)
let rec first_free bound =
  first_freem 0 bound
and first_freem m =  function
    [] -> m+1
  | (_, n)::l -> 
      if (n > m)
      then first_freem n l
      else first_freem m l;;


(* known_lid i bound 
 * returns true
 * if the given located id l 
 * is in the given list of principal's knowledge bound.
 * Otherwise returns false.
 * 
 * Note that the given id can not have an associed index > 0
 * in the knowledge list.
 *)
let rec known_lid i = function
    [] -> false
  | (t, n)::l -> (
      match t.lte_term with
  | EVA_TERM_ID s
  | EVA_TERM_CST s ->
      if (s = string_of_lid i)
      then true
      else (known_lid i l)
    (* ignore les coercions (supposee connues de tous et inversibles) *)
  | EVA_P t1
  | EVA_A t1
  | EVA_SA t1
  | EVA_AA t1
  | EVA_U (_, t1) ->
      known_lid i ((t1, n)::l)
  | _ -> 
      (known_lid i l));;


(* returns the located term
 * corresponding to the given located term
 * in the given list knowledge of the principal
 * who located id is given.
 *
 * It can be:
 * - the located term itself 
 * - a TERM_ID corresp. to the private variable 
 * bound to the given located term in the knowledge list.
 * 
 * raises Not_found if the given located term is not found in the
 * knowledge list.  *)
let rec bound_lterm p lterm = function
    [] -> raise Not_found
  | (tk, n)::l -> 
      (* structural equality *)
      if (lterm_equal lterm tk)
      then (
        if (n = 0)
        then lterm
        (* the private variable, with location = location of the given term *)
        else { lte_loc = lterm.lte_loc; 
               lte_term = EVA_TERM_ID (private_variable p n) }
      )
      else (bound_lterm p lterm l);;

(* returns true if the given located term can
 * be build from the given list of knowledge.
 * Returns false otherwise. *)
let rec known bound lterm =
  trace "Translator.known\n";  (* Debugpp.debugpp_lterm lterm;  print_newline (); *)
  if is_bound_lterm lterm bound
  then true
  else match lterm.lte_term with
      EVA_TERM_ID _ 
    | EVA_TERM_CST _ -> 
        false
     (* coercion symbols are assumed known and invertible *)
    | EVA_P t 
    | EVA_A t 
    | EVA_AA t 
    | EVA_SA t 
    | EVA_U (_, t) ->
        known bound t
    | EVA_CONS tl ->     
        (fold_left (fun x y -> x && y) true (map (known bound) tl))
    | EVA_APP (i, tl) ->
        (known_lid i bound) &&
        (fold_left (fun x y -> x && y) true (map (known bound) tl))
    | EVA_CRYPT (a, k, t) ->
        known_cipher bound (a, k, t)
    | EVA_SIGN (a, k, t) ->
        known_cipher bound (a, k, t)
    | EVA_PCENT (_, _) -> true
    | _ -> raise (Internal_error "Translator.known")
and known_cipher bound = function
    (EVA_VANILLA, k, t) -> 
      (known bound k) && (known bound t)
  | (EVA_ALGO a, k, t) ->
      (known bound a) && (known bound k) && (known bound t)
and is_bound_lterm lterm bound =
  (* trace "Translator.is_bound_lterm "; Debugpp.debugpp_lterm lterm;  print_newline (); *)
  match bound with 
      [] -> false
  | (tk, n)::l -> 
      (* structural equality *)
      if (lterm_equal lterm tk)
      then true
      else (is_bound_lterm lterm l);;


(* the CASRUL's compose function 
 * The argument p (located identifier of principal) 
 * is used only in error messages and for the names of private variables *)
let rec compose p bound lterm = 
  trace "Translator.compose %s\n" (string_of_lid p);
  (* if the term is bound in the knowledge list,
   * returns its bound value 
   * traverses twice the knowledge list *)
  if (is_bound_lterm lterm bound)
  then (bound_lterm p lterm bound)
  else (
    match lterm.lte_term with
        (* misconception in protocol *)
        EVA_TERM_ID i 
      | EVA_TERM_CST i -> 
          raise (Translator_error
                   (("principal "^(string_of_lid p)^
                     " cannot built the identifier "^i), 
                    lterm.lte_loc))
       (* coercion symbols are assumed known and invertible *)
      | EVA_P t ->
          { lterm with lte_term = EVA_P (compose p bound t) }
      | EVA_A t ->
          { lterm with lte_term = EVA_A (compose p bound t) }
      | EVA_AA t ->
          { lterm with lte_term = EVA_AA (compose p bound t) }
      | EVA_SA t ->
          { lterm with lte_term = EVA_SA (compose p bound t) }
      | EVA_U (i, t) ->
          { lterm with lte_term = EVA_U (i, (compose p bound t)) }
        (* descend recursively *)
      | EVA_CONS tl -> 
          let ctl = (map (compose p bound) tl)
          in { lterm with lte_term = EVA_CONS ctl }
      | EVA_APP (i, tl) -> 
          (* check that the function symbol i is known *)
          if (known_lid i bound)
          then 
            (* descend recursively *)
            let ctl = (map (compose p bound) tl)
            in { lterm with lte_term = (EVA_APP (i, ctl)) }
          else raise (Translator_error
                        (("principal "^(string_of_lid p)^
                          " does not know the function "^(string_of_lid i)), 
                         lterm.lte_loc))
      | EVA_CRYPT (EVA_VANILLA, k, t) -> 
          { lterm with lte_term = EVA_CRYPT (EVA_VANILLA, 
                                             (compose p bound k), 
                                             (compose p bound t)) }
      | EVA_CRYPT (EVA_ALGO a, k, t) -> 
          { lterm with lte_term =  EVA_CRYPT (EVA_ALGO (compose p bound a), 
                                              (compose p bound k), 
                                              (compose p bound t)) }
      | EVA_SIGN (EVA_VANILLA, k ,t) -> 
          { lterm with lte_term = EVA_SIGN (EVA_VANILLA, 
                                            (compose p bound k), 
                                            (compose p bound t)) }
      | EVA_SIGN (EVA_ALGO a, k, t) -> 
          { lterm with lte_term = EVA_SIGN (EVA_ALGO (compose p bound a),
                                            (compose p bound k), 
                                            (compose p bound t)) }
      | EVA_PCENT (t1, t2) -> t1
      | _ -> raise (Internal_error "Translator.compose"));;


(* the CASRUL's expect function 
 *
 * returns a pair made of:
 * - the expected pattern
 * - the list of new knowledge and new private variables in the patterns 
 * 
 *  The argument p (located identifier of principal) 
 * is used only for the names of private variables *)
let rec expect p sl bound lterm = 
  trace "Translator.expect\n";  (* Debugpp.debugpp_lterm lterm;  print_newline ();  *)
  (* if the given lterm belongs to the given list of knowledge *)
  if (is_bound_lterm lterm bound)
  then 
    begin
      trace " - expect BOUND \n";
      ((bound_lterm p lterm bound), [])
    end
  else (
    match lterm.lte_term with
        EVA_TERM_ID s 
      | EVA_TERM_CST s -> 
          trace " - expect ID %s \n" s;
          (* add the identifier to the knowledge list *)
          (lterm, [(lterm, 0)])
        (* descend recursively *)
      | EVA_P t ->
          trace " - expect P \n";
          let (tt, btl) = (expect p sl bound t)
          (* add the term to the list of new knowledge *)
          in ({ lterm with lte_term = (EVA_P tt) }, ((lterm, 0)::btl))
      | EVA_A t ->
          trace " - expect A \n";
          let (tt, btl) = (expect p sl bound t)
          in ({ lterm with lte_term = (EVA_A tt) }, ((lterm, 0)::btl))
      | EVA_SA t ->
          trace " - expect SA \n";
          let (tt, btl) = (expect p sl bound t)
          in ({ lterm with lte_term = (EVA_SA tt) }, ((lterm, 0)::btl))
      | EVA_AA t ->
          trace " - expect AA \n";
          let (tt, btl) = (expect p sl bound t)
          in ({ lterm with lte_term = (EVA_AA tt) }, ((lterm, 0)::btl))
      | EVA_U (i, t) ->
          trace " - expect U%s\n" i;
          let (tt, btl) = (expect p sl bound t)
          in ({ lterm with lte_term = (EVA_U (i, tt)) }, ((lterm, 0)::btl))
      | EVA_CONS tl -> 
          trace " - expect CONS \n";
          let (ttl, btl) = (iter_expect p sl bound tl)
          (* add the term to the list of new knowledge *)
          in ({ lterm with lte_term = (EVA_CONS ttl) }, ((lterm, 0)::btl))
      | EVA_APP (i, tl) -> 
          trace " - expect APP \n";
          (* check that the function symbol i is known *)
          if (known_lid i bound)
          then (
            let name = (string_of_lid i) in 
              try 
                let sy = assoc_symbol name sl in
                  (* non invertible function *)
                  (* note: symbols of keypair are always hash *)
                  if (is_key sy) || ((is_functional sy) && (get_hash sy))
                  then 
                    (* if the value to hash is known *)
                    if (known bound lterm)
                    then (lterm, [(lterm, 0)])
                    else (new_private p bound lterm)
                  else 
                    (* descend recursively *)
                    let (ttl, btl) = (iter_expect p sl bound tl)
                                       (* add the term to the list of knowledge *)
                    in ({ lterm with lte_term = (EVA_APP (i, ttl)) }, 
                        (lterm, 0)::btl)
              with
                  Symbols_not_found -> 
                    raise (Translator_error 
                             (("undeclared symbol "^name), i.lid_loc)) )
          (* the principal does not know the function symbol *)
          else (new_private p bound lterm)
            (* raise (Translator_error
                        (("principal does not know the function "^
                          (string_of_lid i)), 
                         lterm.lte_loc)) *)
        (* the algo and inverse key must be known to descend recursively *)
      | EVA_CRYPT (a, k, t) -> 
          trace " - expect CRYPT\n";
          let invk = (inverse_key sl k) in
            (* trace " - expect : inv key "; Debugpp.debugpp_lterm invk;  print_newline ();  *)
            if ((known bound invk) && (known_algo bound a))
            then 
              (* descend recursively *)
              let (te, be) = (expect p sl bound t)
              (* add the term to the list of knowledge *)
              in ( { lterm with 
                       lte_term = (EVA_CRYPT (a, k, te)) }, 
                   (lterm, 0)::be)
            else 
              (new_private p bound lterm)
      | EVA_SIGN (a, k ,t) -> 
          trace " - expect SIGN \n";
          let invk = (inverse_key sl k)
          in 
            if ((known bound invk) && (known_algo bound a))
            then 
              let (te, be) = (expect p sl bound t)
                               (* add the term to the list of knowledge *)
              in ( { lterm with 
                       lte_term = (EVA_SIGN (a, k, te)) }, 
                   (lterm, 0)::be)
            (* new private variable *)
            else 
              (new_private p bound lterm)
      | EVA_PCENT (t1, t2) -> 
          trace " - expect %% \n";
          (t2, [(t2, 0)])
      | _ -> raise (Internal_error "Translator.expect") )
(* return a pair made of
 * - a private variable fresh wrt the given knowledge list bound
 * - a singleton knowledge list containing the binding
 *   of this variable index to the given located term *)
and new_private p bound lterm = 
  let newprivate = (first_free bound)
  in ({ lterm with lte_term = 
          (EVA_TERM_ID (private_variable p newprivate)) },
      [(lterm, newprivate)])
(* iterates the expect function on the given list of located terms
 * and returns a pair made of 
 * - the list of expected pattern
 * - the list of new knowledge (new private variables in the patterns) *)
and iter_expect p sl bound = function
    [] -> ([], [])
  | t::l -> 
      let (te, be) = (expect p sl bound t) in
      let (tl, bl) = (iter_expect p sl (be@bound) l)
      in ((te::tl), be@bl)
and known_algo bound = function
    EVA_VANILLA -> true
  | EVA_ALGO a -> (known bound a);;

(* given 
 * - a located id which is a principal
 * - a symbol list
 * - a list of knowledge 
 * - a list of located instructions,
 * returns the process (eva_process) 
 * for the given principal corresponding to the 
 * given located instructions list.
 * 
 * {b Note:} switch case is NOT supported *)
let rec process p sl bound il =
  trace "Translator.process %s\n" (string_of_lid p);
  match il with
      [] -> { pr_role = p; pr_privates = []; pr_messages = [] }
    | i::l -> (
        match i.lin_instr with    
            EVA_SKIP ->
              let iu = { i with lin_instr = EVA_SKIP } in
              let proc = (process p sl bound l) 
              in { proc with pr_messages = iu::proc.pr_messages }
          | EVA_MSG (lab, s, r, t) ->
              (* p is sender *)
              if (p.lid_id = s.lid_id)
              then 
                let m = (compose p bound t) in
                let iu = { i with lin_instr = (EVA_MSG (lab, s, r, m)) } in
                let proc = (process p sl bound l) 
                in { proc with pr_messages = iu::proc.pr_messages }
              (* p is receiver *)
              else if (p.lid_id = r.lid_id) 
              then 
                let (m, ub) = (expect p sl bound t) in 
                  (* DEBUG 
                  trace "\nTranslator.process: bound list at return of expect:\n";
                  (Pp.pp_list 
                     (fun (t, n) -> print_string "("; Debugpp.debugpp_lterm t; print_string ",  "; print_int n; print_string ")") 
                     (fun () -> print_newline ())
                     ub);
                  print_newline ();  
                   END DEBUG *)
                let iu = { i with lin_instr = (EVA_MSG (lab, s, r, m)) } in
                let proc = (process p sl (ub@bound) l)
                (* the terms of ub, obtained by matching 
                 * are added to the list of knowledge *)
                in { proc with
                       pr_privates = (privates_of_bound p ub)@(proc.pr_privates);
                       pr_messages = iu::proc.pr_messages }
              (* ignore, principal not concerned *)
              else 
                let iu = { i with lin_instr = EVA_SKIP } 
                in (process p sl bound (iu::l))
          | EVA_ASSIGN (r, x, t) ->
              if (p.lid_id = r.lid_id) 
              then (
                if (known_lid x bound)
                then 
                  raise (Translator_error 
                           (("reassignement of parameter or constant "^(string_of_lid x)), 
                            x.lid_loc))
                    (* compose will fail *)
                else if (not (known bound t))
                then 
                  raise (Translator_error 
                           ("assignation of unknown value", t.lte_loc))
                else 
                  let tb = (compose p bound t) in
                  let xb = ({ lte_loc = x.lid_loc; 
                              lte_term = (EVA_TERM_ID (string_of_lid x)) }, 0) in
                  let iu = { i with lin_instr = (EVA_ASSIGN (r, x, tb)) } in 
                    (* the assigned variable x is added to the list of knowledge *)
                  let proc = (process p sl (xb::bound) l)                           
                               (* but it [the assigned variable x] is not a new private variable *)
                  in { proc with pr_messages = iu::proc.pr_messages }
              )              
              (* ignore, principal not concerned *)
              else (process p sl bound l)    
          | EVA_COMP (r, x, t) ->
              if (p.lid_id = r.lid_id) 
              then (
                if (not (known_lid x bound))
                then 
                  raise (Translator_error 
                           (("left member "^(string_of_lid x)^
                             " in comparison is unknown to "^(string_of_lid r)),
                            x.lid_loc))
                (* compose will fail *)    
                else if (not (known bound t))
                then 
                  raise (Translator_error 
                           ("right member in comparison is unknown", t.lte_loc))
                else 
                  let tb = (compose p bound t) in
                  let iu = { i with lin_instr = (EVA_COMP (r, x, tb))} in
                  let proc = (process p sl bound l)
                  in { proc with pr_messages = iu::proc.pr_messages }
              )
              (* ignore, principal not concerned *)
              else (process p sl bound l)     
          | EVA_BLOCK il ->
              (* grouping instruction in a block is useless.
               * remove the surrounding block, 
               * and just treat the  instr of the block like others *)
              (process p sl bound (il@l))
          | EVA_SWITCH (t, cl) -> 
              (* switch not allowed in this function *)
              raise (Internal_error "process: switch")
    )
and privates_of_bound p = function
    [] -> []
  | (t, n)::l ->
      if (n > 0)
      then { lid_loc = t.lte_loc; lid_id = (EVA_ID (private_variable p n)) }::(privates_of_bound p l)
      else (privates_of_bound p l);;


(* returns the list of active principals (located ids)
 * from the given list of instr *)
let rec principals = function
    [] -> []
  | {lin_instr = EVA_SKIP}::l -> 
      (principals l)
  | {lin_instr = EVA_MSG (_,s,r,_)}::l ->
      (add_principal s (add_principal r (principals l)))
  | {lin_instr = EVA_ASSIGN (r,_,_)}::l 
  | {lin_instr = EVA_COMP (r,_,_)}::l ->
      (add_principal r (principals l))
  | {lin_instr = EVA_BLOCK bl}::l ->
      (principals bl)@(principals l)
  | {lin_instr = EVA_SWITCH (_,cl)}::l ->
      (principals_caselist cl)@(principals l)
and principals_caselist = function
    [] -> []
  | {case_block = il}::l ->
      (principals il)@(principals_caselist l)
and add_principal i = function
    [] -> [i]
  | i2::l -> 
      (* structural eq *)
      if (i2.lid_id = i.lid_id)
      then i2::l
      else i2::(add_principal i l);;
 
(* @return the list of knowledge declared for the given
 * principal (located id) in the given list of located
 * declarations.
 * @return the empty list if there is no knowledge
 * declaration for the principal. *)
let rec knowledge p = function
    [] -> []
  | { ldc_dcl = EVA_DCL_KNOW (i, kl)}::l ->
      (* structural eq *)
      if (i.lid_id = p.lid_id)
      then kl
      else knowledge p l
  | _::l -> knowledge p l;;


(* returns a protocol specification
 * where the messages are presented in separates processes
 * given a symbol list and a protocol specification
 * where the messages are presented a la BAN.
 * The other fields of the spcification are left untouched *)
let translate sl sp = 
  trace "Translator.translate\n";
  let instrs = ( 
    match sp.sp_protocol.lpr_protocol with
        EVA_MP x -> x
      | EVA_PROCESS _ -> raise (Internal_error "translate: not BAN spec")) in
  let prs = (principals instrs) in
  (* returns the initial knowledge of the given principal *)
  let bound = fun
      p ->  bound_of_know (knowledge p sp.sp_declarations) in
  let protocol = 
    { sp.sp_protocol with 
        lpr_protocol = 
        EVA_PROCESS (map (fun p -> (process p sl (bound p) instrs)) prs) }
  in { sp with sp_protocol = protocol };;

