(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(**
   Module for 
   - checking a specification in abstract syntax
   - and translating a specification a la BAN (several processes in a single message list)
     into a multi processes specification.
 *)

open Abstract;;
open Symbols;;
open Location;;

(** raised by functions the module [translator] to signal an error
  in the input file.
  - message
  - location of the error *)
exception Translator_error of string * location;;


(** [symbollist_of_spec sp]
   @return the symbol list extracted from:
   - the declarations (field {!Abstract.eva_spec}[.sp_declarations])
     of the given specification [sp]
   - the value declarations (field {!Abstract.eva_spec}[.sp_valuedcl])
     of the given eva_spec.
 
   {b Note:} the other identifiers (quantified identifiers in axioms or
   formulas and private process variables) are not included in the symbol list.  *)
val symbollist_of_spec : 
  eva_spec -> eva_symbol_list

(**************************************************************
 * Conversion of located terms
 * 
 * This functions check and convert the symbols in terms 
 * obtained from parsing.
 **************************************************************)

(** [convert_lterm sl t]
  @return a located term obtained from the given located term [t],
  which must be a term returned by the parser,
  by conversion of the symbols at the leaves, 
  using the given symbol list [sl].
  
  The parser cannot distinguish between identifiers in terms
  which are first order ids, values, alias or quantified variables.
  Hence, after parsing, every identifier (string) [i] occurring in a term
  is stored in a [EVA_TERM_ID i] (see {!Abstract.eva_term}), 
  whatever its declaration.
  This function restores the given located term [t] obtained from parser
  by converting every leaf [EVA_TERM_ID] into
  [EVA_TERM_ID], [EVA_TERM_VALUE] or [EVA_TERM_ALIAS] or [EVA_QUANTIFIED]
  according to the given symbol list [sl], and returns the terms obtained. 
  
  @raise Translator_error if the given term [t] contains
  an unexpected symbol or undeclared symbol [i] in a subterm
  [EVA_TERM_ID i].  *)
val convert_lterm :
  eva_symbol_list -> eva_lterm -> eva_lterm


(** [convert_ldcl sl d]
  @return a located declaration 
  obtained from the given located declaration [d] (returned by the parser)
  and which conforms to the abstract syntax 
  according to the given symbol list [sl].
  
  - conversion of axiom declarations:
  The members of the axiom are converted ({!Translator.convert_lterm})
  with to the given symbol list overwritten by the quantified variables.
  This function evaluates then and adds the type of the axiom members
  (the type of the members of an axiom is not required in the 
  LAEVA concrete syntax but is required in the abstract syntax).
  
  - conversion of others declarations:
  the function {!Translator.convert_lterm} is applied to every embedded term.
  
  @raise Translator_error if the types of left and right members of an axiom differ
  or if [d] contains an undeclared symbol. *)
val convert_ldcl :
  eva_symbol_list -> eva_ldcl -> eva_ldcl


(** [convert_lstatement sl st]
   @return a located statement
   obtained from the given located statement [st] (returned by the parser).
   and which conforms to the abstract syntax 
   according to the given symbol list [sl].
   
   The function {!Translator.convert_lterm} is applied to the terms of the statement,
   with the symbol list [sl] overwritten by the quantified variables. *)
val convert_lstatement :
  eva_symbol_list -> eva_lstatement -> eva_lstatement

(** [convert_spec sl sp]
   applies the convertion functions {!Translator.convert_lterm}, 
   {!Translator.convert_ldcl}, 
   {!Translator.convert_lstatement}
   to the elements of the given abstract specification [sp], 
   with the given symbol list [sl], and return the specification obtained. *)
val convert_spec :
  eva_symbol_list -> eva_spec -> eva_spec


(** flag for term right meaning that value symbols are allowed in terms. *)
val term_VALUE : int

(** flag for term rights meaning that constant symbols are allowed 
  in terms (see {!Translator.check_lterm}). *)
val term_CST   : int

(** flag for term rights meaning that parameters symbols are allowed 
  in terms (see {!Translator.check_lterm}). *)
val term_PARAM : int

(** flag for term rights meaning that private variables are allowed 
  in terms (see {!Translator.check_lterm}). *)
val term_PRIV  : int

(** flag for term right meaning that percent (Lowe's notation) 
  are allowed in terms. *)
val term_PCENT : int

(** flag for term right meaning that located variables
  are allowed in terms. *)
val term_AT : int

(** flag for term right meaning that first order symbols 
  are allowed in terms. *)
val term_FO : int

(** flag for term right meaning that functional symbols
  are allowed in terms. *)
val term_FUN : int

(** flag for term rights meaning that keypairs symbols
  are allowed in terms (see check_lterm) *)
val term_KEY : int
  
(** flag for term right meaning that alias symbols
  are allowed in terms. *)
val term_ALIAS : int

(** flag for term right meaning that quantified variables
  are allowed in terms. *)
val term_QUANTIF : int
  
(** sum of all flags. *)
val term_ALL : int
  

(** [check_ltype sl t]
   @return the given located type [t] if it is well formed w.r.t.
   the given symbol list [sl].
 
   @raise Translator_error otherwise. *)
val check_ltype :
  eva_symbol_list -> eva_ltype -> eva_ltype


(** [check_lterm rl sl lterm]
   @return the given located term [lterm]
   if its leaves are correct w.r.t.
   the given rights list [rl] and the given symbol list [sl].
   The right list is a list of flags
   defining which symbols are allowed in terms. 
 
   @raise Translator_error otherwise. *)
val check_lterm :
  int list -> eva_symbol_list -> eva_lterm -> eva_lterm


(** [check_spec sl sp]
  @return  the spec obtained by 
  applying the function {!Translator.check_lterm} to every term of the
  given spec [sp], with different rights. *)
val check_spec :
  eva_symbol_list -> eva_spec -> eva_spec

(** [check_linstr sl i]
  @return the given located instruction [i] if 
  the identifiers sender and receiver of messages or
  roles of comp and assignment are declared as principal.
  
  @raise Translator_error otherwise *)
val check_linstr :
  Symbols.eva_symbol_list -> Abstract.eva_linstr -> Abstract.eva_linstr

(** [check_instrs_spec sl sp]
  @return the given located specification [sp] if 
   checking of all the instructions it contains,
   applying {! Translator.check_linstr} with the given symbol list [sl],
   is successful. *)
val check_instrs_spec :
  Symbols.eva_symbol_list -> Abstract.eva_spec -> Abstract.eva_spec

(** [translate sl sp]
   @return a protocol specification
   where the messages are presented in separates processes
   given a symbol list [sl] and a protocol specification [sp],
   where the messages are presented a la BAN.
   The other fields of the specification are left untouched *)
val translate :
  Symbols.eva_symbol_list -> Abstract.eva_spec -> Abstract.eva_spec
