(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(***************************************************************
 * Type check and type convertion
 ***************************************************************)

open List;;

open Trace;;
open Location;;
open Abstract;;
open Symbols;;
open Termutil;;

exception Type_error of string * location;;

(***************************************************************
 * Types equality
 ***************************************************************)

(* we make no differences between user and
 * base types in the translator *)
let equals t1 t2 = 
  match (t1, t2) with
    ((EVA_USERTYPE s1), (EVA_BASETYPE s2)) 
  | ((EVA_BASETYPE s1), (EVA_USERTYPE s2)) ->
      (s1 = s2)
  | (x, y) ->
      x = y;;


(***************************************************************
 * Types check
 ***************************************************************)

(* The respective type disciplines of the LAEVA concrete syntax and 
 * of the abstract syntax differ.
 * Indeed, in the lists of terms that occur under the constructors 
 * EVA_CONS and EVA_APP, 
 * components of various types are allowed in LAEVA 
 * (provided they respect the signature of the function symbol in EVA_APP)
 * and not in the abstract syntax,
 * where all the components of a tuple must have type EVA_NUMBER.
 * 
 * We provide here two function for type checking
 * which correspond to each of the discipline:
 * - type_lterm_laeva ehich conforms to the LAEVA discipline,
 *   and checks in APP terms the type of the arguments againts the declared 
 *   signature. 
 * - type_lterm which conforms to the disipline of the abstract syntax.
 *   It must be call after term conversion and type coercion. *)

(* type_lterm sl t
 * checks that the given located term t is well typed 
 * wrt to the given symbol list sl
 * according to the abstract syntax type discipline (see remark above):
 *  - in APP(c, args), the terms of args must all have type EVA_NUMBER
 *  - in CONS(t, l), t must have type EVA_NUMBER
 *  - in CRYPT(a, k, t) a must have type EVA_SALGO and
 *    k and t must have type EVA_NUMBER.
 * It returns the type (not located) of the given located term t.
 *
 * {b Note:} the term must NOT contain alias symbols
 *
 * {b Note:} this function should not be applied directly 
 * to terms obtained after parsing EVA protocol spec.
 * Translator.convert_leterm (Translator.check_lterm)
 * and coerce_lterm should be applied first.
 *
 * raise Type_error if the given term 
 * contains symbols not declared in sl
 * or is not well formed (arity) according sl
 * or is not well typed according to sl.
 *
 * raise Type_error if the given term contains alias symbols 
 * or some leaf symbol is not conform to symbol list *)
let rec type_lterm sl term = 
  match term.lte_term with
     (* returns the declared type of the identifier *) 
      EVA_TERM_ID i 
    | EVA_TERM_CST i -> (
        try 
          let sy = (assoc_symbol i sl) in
          let ty = (get_type sy) in
            if (is_first_order sy)
            then ty
            else if (is_key sy)
            then EVA_KEY ty
            else if (is_functional sy)
            then 
              if (get_hash sy)
              then EVA_HFUN ty
              else EVA_FUN ty
            else raise
              (Type_error (("unexpected identifier "^i), term.lte_loc))
        with 
            Symbols_not_found -> raise 
              (Type_error (("symbol not declared "^i), term.lte_loc))
          | (Symbols_error msg) -> raise 
              (Type_error (("unexpected symbol "^i), term.lte_loc))
      )
    (* returns the declared type of the value *)
    | EVA_TERM_VALUE EVA_INTRUDER -> 
        EVA_PRINCIPAL 
    | EVA_TERM_VALUE (EVA_VALUE name) -> (
        try 
          (get_type (assoc_symbol name sl))
        with 
            Symbols_not_found -> raise 
              (Type_error (("value symbol not declared "^name), term.lte_loc))
          | (Symbols_error msg) -> raise 
              (Type_error (("symbol "^name^
                            " used as value symbol but declared differently"),
                           term.lte_loc))
      )
    | EVA_TERM_ALIAS (a, ta) -> (type_lterm sl ta)
    (* check that the role has type principal, and returns the type of the var *)
    | EVA_LOCATED (var, role) -> 
        (* get the type of the role *)
        let cond_role = (
          let i = (string_of_lid role) in 
            try 
              let type_role = (get_type (assoc_symbol i sl)) in           
                (* check that the type of the role is principal *)
                if (type_role <> EVA_PRINCIPAL)
                then raise (Type_error (("symbol "^i^
                                         " used as role in located variable"^
                                         " has not the type principal"), 
                                        role.lid_loc))
                else true
            with 
                Symbols_not_found -> raise 
                  (Type_error (("symbol "^i^
                                " used as role in located variable"^
                                " is not declared"),
                               role.lid_loc))
              | (Symbols_error msg) -> raise 
                  (Type_error (("symbol "^i^
                              " used as role in located variable"^
                              " but not declared as first order"),
                             role.lid_loc)))
        in let i = (string_of_lid var) in
          (            
            try 
              (* returns the type of the var *)
              get_type (assoc_symbol i sl)
            with 
                Symbols_not_found -> raise 
                  (Type_error (("undeclared"^
                                " located variable symbol "^i),
                               var.lid_loc))
              | (Symbols_error msg) -> raise 
                  (Type_error (("located variable symbol "^i^
                                " is not declared as first order"),
                               var.lid_loc))
          )
    | EVA_QUANTIFIED (_, t) -> t
    | EVA_P t -> 
        (* the casted term must have type principal *)
        if ((type_lterm sl t) = EVA_PRINCIPAL)
        then EVA_NUMBER
        else raise (Internal_error "type_lterm: cast P")
    | EVA_A t -> 
        (* the casted term must have type algo *)
        if ((type_lterm sl t) = EVA_TALGO)
        then EVA_NUMBER
        else raise (Internal_error "type_lterm: cast A")
    | EVA_SA t -> 
        (* the casted term must have type s_algo *)
        if ((type_lterm sl t) = EVA_SALGO)
        then EVA_TALGO
        else raise (Internal_error "type_lterm: cast SA")
    | EVA_AA t -> 
        (* the casted term must have type principal *)
        if ((type_lterm sl t) = EVA_AALGO)
        then EVA_TALGO
        else raise (Internal_error "type_lterm: cast AA")
    | EVA_U (i, t) -> 
        (* the casted term must have type usertype *)
        let sy = (assoc_symbol i sl) in
        if (is_basetype sy)
        then (
          (* more precise than equals *)
          if ((type_lterm sl t) = (EVA_BASETYPE i))
          then EVA_NUMBER
          else raise (Internal_error "type_lterm: cast U basetype"))
        else if (is_type sy)
        then (
          if ((type_lterm sl t) = (EVA_USERTYPE i))
          then EVA_NUMBER
          else raise (Internal_error "type_lterm: cast U usertype"))
        else raise (Internal_error 
                      "type_lterm: cast P symbol not declared as type")
     (* all the members of the list must have type number *)
    | EVA_CONS tl -> 
        (* check that all members have type NUMBER *)
        if (for_all (fun x -> (x = EVA_NUMBER)) (map (type_lterm sl) tl))
        then EVA_NUMBER        
        else raise (Type_error (("type error in list"),term.lte_loc))
      (* compute the type of subterms, check with signature of root 
       * and if correct returns type of root symbol 
       * app_root = i; app_args = tl *)
    | EVA_APP (i, tl) -> (
        match i.lid_id with
            (EVA_ID name) -> 
              try
                let sf = (assoc_symbol name sl) in
                let signature = (get_signature sf) in 
                  (
                    (* check that f was declared as function symbol or key *)
                    if (not ((is_functional sf) || (is_key sf)))
                    then raise (Type_error 
                                  (("symbol "^name^
                                    " used in application is not "^
                                    "declared as function"),
                                   i.lid_loc))
                    (* check arity *)
                    else if ((length tl) != (length (get_signature sf)))
                    then raise (Type_error 
                                  (("function symbol "^name^
                                    " applied to wrong number of arguments"),
                                   term.lte_loc))
                    (* we do not care the signature of i, 
                     * in EVA type system, all the args of tl 
                     * must have type NUMBER *)
                    else 
                      if (for_all (fun x -> (x = EVA_NUMBER)) 
                            (map (type_lterm sl) tl))
                      then (get_type sf)
                      else raise (Type_error 
                                    (("type error in term arguments "^
                                      "(non number argument)"), 
                                     term.lte_loc))
                  )
              with
                  Symbols_not_found -> raise 
                    (Type_error (("undeclared symbol "^name), i.lid_loc))
      )
      (* check the type of the components of the cipher, returns EVA_NUMBER *)
    | EVA_CRYPT (algo, key, text)
    | EVA_SIGN (algo, key, text) -> 
        if (verify_cipher_type sl algo key text)
        then EVA_NUMBER
        (* never reached *)
        else raise (Type_error ("type error in ciphertext", term.lte_loc))
        (* check that both sides must have the same type, returns this type *)
    | EVA_PCENT (t1, t2) ->
        let type1 = (type_lterm sl t1)
        and type2 = (type_lterm sl t2)
        in 
          if (equals type1 type2)
          then type1            
          else raise (Type_error 
                        ("two sides of % have different types", 
                         term.lte_loc))
(* returns true if the given components of a ciphertext are
 * correctly typed
 * otherwise fails with Type_error.
 * 
 * crypt_algo = algo; crypt_key = key; crypt_content = term *)
and verify_cipher_type sl algo key text =
  (* check type of the algo *)
  let cond_algo = 
    match algo with
        EVA_VANILLA -> true
      | EVA_ALGO a -> 
          let t = (type_lterm sl a)
          in (if ((t = EVA_AALGO) || (t = EVA_SALGO) || (t = EVA_TALGO))
              then true
              else raise 
                (Type_error 
                   ("algorithm expression in ciphertext not typed as an algorithm",
                    a.lte_loc)))
  (* check that the key has type NUMBER *)
  and cond_key = (
    if ((type_lterm sl key) <> EVA_NUMBER)
    then raise (Type_error ("type error for key in ciphertext", key.lte_loc))
    else true)
  (* check that the excrypted text has type NUMBER *)
  and cond_text = (
    if ((type_lterm sl text) <> EVA_NUMBER)
    then raise (Type_error ("type error for encrypted text in ciphertext", text.lte_loc))
    else true)
  in (cond_algo && cond_key && cond_text);;




(* type_lterm_laeva sl t
 * checks the typing of the given located term t
 * wrt to the given symbol list sl
 * according to the type discipline of LAEVA (aka concrete syntax) 
 * which is less resctrictive than the discipline for the abstract syntax 
 * (see remark above):
 *  - in APP(c, args), 
 *    the type of the terms of args must follow the declared signature of c,
 *  - in CONS(t, l), the type of t is not checked.
 *  - in CRYPT(a, k, t) a must have type EVA_SALGO, the
 *    types of k and t are not checked.
 * 
 * The purpose of this function is to check 
 * the terms obtained from the EVA parser,
 * after conversion with Translator.convert_lterm, 
 * and not yet casted,
 * for conformity of terms against
 * the type declaration of function symbols. 
 * It returns the type (not located) of the given located term t.
 *
 * {b Note:} the term must NOT contain alias symbols
 *
 * {b Note:} this function should not be applied directly 
 * to terms obtained after parsing EVA protocol spec.
 * Translator.convert_leterm (optionaly Translator.check_lterm)
 * should be applied first.
 * 
 * {b Note:} this function will likely fail when
 * applied to coerced located terms as returned by
 * coerce_lterm (hence well typed terms wrt EVA semantics).
 *
 * raise Type_error if the given term 
 * contains symbols not declared in sl
 * or is not well formed (arity) wrt sl 
 * or is not well typed wrt sl as described above.
 *
 * raise Type_error if the given term contains alias symbols 
 * or some leaf symbol is not conform to the symbol list sl *)
let rec type_lterm_laeva sl term = 
  match term.lte_term with
      (* returns the declared type of the identifier *) 
      EVA_TERM_ID i 
    | EVA_TERM_CST i -> 
        type_lterm sl term
      (* args not checked *)
    (* returns the declared type of the value *)
    | EVA_TERM_VALUE EVA_INTRUDER -> 
        EVA_PRINCIPAL 
    | EVA_TERM_VALUE (EVA_VALUE name) -> (
        try 
          (get_type (assoc_symbol name sl))
        with 
            Symbols_not_found -> raise 
              (Type_error (("value symbol not declared "^name), term.lte_loc))
          | (Symbols_error msg) -> raise 
              (Type_error (("symbol "^name^
                            " used as value symbol but declared differently"),
                           term.lte_loc))
      )
    | EVA_TERM_ALIAS (a, ta) -> 
        (type_lterm_laeva sl ta)
    (* check that the role has type principal, and returns the type of the var *)
    | EVA_LOCATED (var, role) -> 
        type_lterm sl term 
    | EVA_QUANTIFIED (_, t) -> 
        t
    | EVA_P _
    | EVA_A _
    | EVA_SA _
    | EVA_AA _
    | EVA_U (_, _) ->
        raise (Internal_error "type_lterm_laeva: called with casted lterm")
    | EVA_CONS tl -> 
        (* recursively check typing of args but do not care of returned type *)
        let pipo = (map (type_lterm_laeva sl) tl) 
        in EVA_NUMBER        
     (* compute the type of subterms, check with signature of root 
      * and if correct returns type of root symbol 
      * app_root = i; app_args = tl *)
    | EVA_APP (i, tl) -> (
        match i.lid_id with
            (EVA_ID name) -> 
              try
                let sf = (assoc_symbol name sl) in
                let signature = (get_signature sf) in 
                  (
                    (* check that f was declared as function symbol or key *)
                    if (not ((is_functional sf) || (is_key sf)))
                    then raise (Type_error 
                                  (("symbol "^name^
                                    " used in application is not "^
                                    "declared as function"),
                                   i.lid_loc))
                    (* check arity *)
                    else if ((length tl) != (length (get_signature sf)))
                    then raise (Type_error 
                                  (("function symbol "^name^
                                    " applied to wrong number of arguments"),
                                   term.lte_loc))
                   (* compares the typing of arguments 
                    * to the signature declaration *)
                    else 
                      let ttl = (map (type_lterm_laeva sl) tl) in                      
                        if (List.for_all2 equals ttl signature)
                        then (get_type sf)                     
                        else raise (Type_error 
                                      (("type error in term arguments"), 
                                       term.lte_loc))
                  )
              with
                  Symbols_not_found -> raise 
                    (Type_error (("undeclared symbol "^name), i.lid_loc))
      )
      (* check the type of the algo component of the cipher, returns EVA_NUMBER *)
    | EVA_CRYPT (algo, key, text)
    | EVA_SIGN (algo, key, text) -> (
        (* recursively check typing of key, text, but do not care returned type *)
        let pipo1 = (type_lterm_laeva sl key)
        and pipo2 = (type_lterm_laeva sl text)
        in match algo with
            EVA_VANILLA -> EVA_NUMBER
          | EVA_ALGO a -> 
              let t = (type_lterm_laeva sl a)
              in (if ((t = EVA_AALGO) || (t = EVA_SALGO) || (t = EVA_TALGO))
                  then EVA_NUMBER
                  else raise 
                    (Type_error 
                       ("algorithm expression in ciphertext not typed as an algorithm",
                        a.lte_loc))))
      (* check that both sides must have the same type, returns this type *)
    | EVA_PCENT (t1, t2) ->
        let type1 = (type_lterm_laeva sl t1)
        and type2 = (type_lterm_laeva sl t2)
        in 
          if (equals type1 type2)
          then type1            
          else raise (Type_error 
                        ("two sides of % have different types", 
                         term.lte_loc));;


       
(**************************************************************
 * Coercions and typecheck
 **************************************************************)

(* coerce_lterm sl t
 * typecheck the given located term t
 * wrt the given symbol list sl, 
 * casts it to the type EVA_NUMBER,
 * and return the casted copy.
 * 
 * {b Note:} the symbols in the term must be conform to 
 * the entries in symbol list, in particular
 * this function should not be applied directly 
 * to terms obtained after parsing EVA protocol spec.
 * see Translator.convert_leterm and Translator.check_lterm
 *
 * raise Type_error if the given term contains symbols not declared in sl
 * or is not well formed (arity) according sl
 * or is not well typed according to sl.
 *
 * raise Type_error if some leaf symbol 
 * in the given term is not conform to symbol list 
 * 
 * raise Type_error if some leaf (TERM_ID or TERM_CST)
 * is not a first order symbol 
 * (even if it is a function or keypair symbol). *)
let rec coerce_lterm sl term = 
  match term.lte_term with
      EVA_TERM_ID i 
    | EVA_TERM_CST i -> (
        try          
          let sy = (assoc_symbol i sl) in
            if (is_first_order sy)
            then (cast term (get_type sy))
            else raise 
              (Type_error (("first order symbol expected: "^i), term.lte_loc))
        with
            (Symbols_error msg) ->  
              raise (Internal_error ("coerce_lterm: Symbols_error "^msg))
          | Symbols_not_found -> raise
              (Type_error (("undeclared symbol "^i), term.lte_loc))
      )
    | EVA_TERM_VALUE EVA_INTRUDER -> 
        (cast term EVA_PRINCIPAL)
    | EVA_TERM_VALUE (EVA_VALUE name) -> (
        try
          (cast term (get_type (assoc_symbol name sl)))
        with
            (Symbols_error msg) ->  raise 
              (Type_error (("value symbol expected: "^name),
                           term.lte_loc))
          | Symbols_not_found -> raise
              (Type_error (("undeclared symbol "^name),
                          term.lte_loc))
      )
    | EVA_TERM_ALIAS (a, ta) -> 
        { term with lte_term = EVA_TERM_ALIAS (a, coerce_lterm sl ta) }        
    | EVA_LOCATED (var, role) -> (
        let i = (string_of_lid var) in
          try
            let var_type = (get_type (assoc_symbol i sl))
            in (cast term var_type) 
          with
              (Symbols_error msg) ->  raise 
                (Type_error (("located variable symbol "^i^
                              " is not declared"^
                              " as a first order identifier"),
                             var.lid_loc))
            | Symbols_not_found -> raise
                (Type_error (("undeclared symbol "^i), var.lid_loc))
      )
    | EVA_QUANTIFIED (_, t) -> 
        (cast term t)
    | EVA_P _
    | EVA_A _
    | EVA_SA _
    | EVA_AA _
    | EVA_U _ ->
        raise (Internal_error "coerce_lterm: coercion symbols")
    | EVA_CONS tl -> 
        { term with lte_term = EVA_CONS (map (coerce_lterm sl) tl) }
      (* the args of tl are coerced (to type NUMBER) *)
    | EVA_APP (f, tl) -> (
        let i = (string_of_lid f) in 
          try
            let sy = (assoc_symbol i sl) in
            let f_type = (get_type sy) in
            let f_signature = (get_signature sy) in 
            (* we check here the types of the args tl agains signature,
             * after coercion of the args (to NUMBER), 
             * the declared signature becomes obsolete *)
            let ttl = (map (type_lterm_laeva sl) tl) in                      
              if (List.for_all2 equals ttl f_signature) 
              then (cast 
                      { term with 
                          lte_term = EVA_APP (f, (map (coerce_lterm sl) tl)) }
                      f_type)
              else raise (Type_error 
                            (("type error in term arguments"), 
                             term.lte_loc))
          with
              (Symbols_error msg) ->  raise 
                (Type_error (("symbol "^i^
                              " used as functional symbol"^
                              " but declared differently"), f.lid_loc))
            | Symbols_not_found -> raise
                (Type_error (("undeclared symbol "^i), f.lid_loc))
      )
    | EVA_CRYPT (algo, key, text) -> 
        { term with 
            lte_term = EVA_CRYPT ((coerce_algo sl algo), 
                                 (coerce_lterm sl key), 
                                 (coerce_lterm sl text))
        }
    | EVA_SIGN (algo, key, text) -> 
        { term with 
            lte_term = EVA_SIGN ((coerce_algo sl algo), 
                                 (coerce_lterm sl key), 
                                 (coerce_lterm sl text))
        }
    | EVA_PCENT (t1, t2) -> 
        { term with 
            lte_term = EVA_PCENT ((coerce_lterm sl t1), (coerce_lterm sl t2)) }
and coerce_algo sl = function
    EVA_VANILLA  ->  EVA_VANILLA
  | EVA_ALGO ta -> EVA_ALGO (coerce_lterm sl ta)
(** [cast_term t ty]
  - param  [t] a located term
  - param [ty] an EVA type
  @return the term of type NUMBER obtained from [t] by applying 
  the appropriate coercion function.
  NB: the coercion is not applied recursively. *) 
and cast term = function 
    EVA_VOID -> 
      raise (Internal_error "cast: void type") 
  (* coercion P : principal -> number *)
  | EVA_PRINCIPAL -> 
      (* we keep the same location for the casted term *)
      { term with lte_term = EVA_P term }
  | EVA_NUMBER -> term
  (* coercion _aa : asym_algo -> *algo* *)
  | EVA_AALGO -> 
      (cast { term with lte_term = EVA_AA term } EVA_TALGO)
  (* coercion _sa : sym_algo -> *algo* *)
  | EVA_SALGO -> 
      (cast { term with lte_term = EVA_SA term } EVA_TALGO)
  (* coercion _a : *algo* -> number *)
  | EVA_TALGO -> 
      { term with lte_term = EVA_A term }
  | EVA_USERTYPE ut 
  | EVA_BASETYPE ut -> 
      { term with lte_term = EVA_U (ut, term) }
  (* should not happen *)
  | _ ->
      raise (Internal_error "cast-term: invalid type");;

(* descend_coerce_lterm sl t
 * return a term obtained from [t] by 
 * appropriate application of {!Types.coerce_lterm}
 * to the strict subterms, but not to [t] itself.
 * 
 * {b Note:} the symbols in the term must be conform to 
 * the entries in symbol list, in particular
 * this function should not be applied directly 
 * to terms obtained after parsing EVA protocol spec.
 * see Translator.convert_leterm and Translator.check_lterm
 *
 * raise Type_error if the given term contains symbols not declared in sl
 * or is not well formed (arity) according sl
 * or is not well typed according to sl. *)
let rec descend_coerce_lterm sl term =  
  match term.lte_term with 
      (* it also works when [i] is not first order *)
      EVA_TERM_ID i 
    | EVA_TERM_CST i -> 
        term
    | EVA_TERM_VALUE EVA_INTRUDER -> 
        term 
    | EVA_TERM_VALUE (EVA_VALUE name) -> 
        term
    | EVA_TERM_ALIAS (a, ta) -> 
        { term with lte_term = EVA_TERM_ALIAS (a, descend_coerce_lterm sl ta) }        
    | EVA_LOCATED (var, role) ->
        term
    | EVA_QUANTIFIED (_, t) -> 
        term
    | EVA_P _
    | EVA_A _
    | EVA_SA _
    | EVA_AA _
    | EVA_U _ ->
        raise (Internal_error "descend_coerce_lterm: coercion symbols")
    | EVA_PCENT (t1, t2) -> 
        { term with 
            lte_term = EVA_PCENT ((descend_coerce_lterm sl t1), 
                                  (descend_coerce_lterm sl t2)) }
    (* the other terms have type number, 
       they are not coerced at root by [coerce_lterm] *)
    | _ ->
        (coerce_lterm sl term);;


(* coerce_spec sl sp
 * applies coerce_lterm sl 
 * to every term of the given spec sp,
 * except in sessions, 
 * and returns the sepc obtained.  *)
let coerce_spec sl sp = 
  let tr = (coerce_lterm sl) 
  and dtr = (descend_coerce_lterm sl) 
  in { sp with
         (* coerce terms in declarations, ie in aliases, know, axioms *)
         sp_declarations = (map (lterm_replace_ldcl dtr) sp.sp_declarations); 
         (* coerce terms in protocol messages *)         
         sp_protocol = lterm_replace_lprotocol tr sp.sp_protocol;
         (* coerce terms in formulas *)         
         sp_statements = (map (lterm_replace_lstatement tr) sp.sp_statements);
     }
;;
  
(**************************************************************
 * Typecheck only (TODO ?)
 **************************************************************)
