(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the 
 *         consequences of use of this software, no matter how awful, 
 *         even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *	   use is subject to a specific license, obtainable from LSV.
 *	   This was developed as part of the EVA project; as such, access
 *	   to this software by partners of the RNTL EVA Project (Trusted
 *	   Logic S.A., Verimag) is defined by the intellectual property
 *	   agreement signed as part of this project.
 *)

(** Module for type checking and type convertion *)

open Location;;
open Abstract;;
open Symbols;;

(** raised by functions of the [types] module to signal a type error
  in the input file.
  - message
  - location of the error *)
exception Type_error of string * location

(** type equality.

  [equals t1 t2]
  @return whether the 2 given terms [t1] and [t2] are equal.

  {b Note:} We make no differences between user and base types, 
  i.e. [USERTYPE s] and [BASETYPE s] are considered equal. *)
val equals :
  eva_type -> eva_type -> bool


(***************************************************************
 * Types check
 ***************************************************************)

(** {b Type disciplines.} 
  The respective type disciplines of the LAEVA concrete syntax and 
  of the abstract syntax differ.
  Indeed, in the lists of terms that occur under the constructors 
  [EVA_CONS] and [EVA_APP], 
  components of various types are allowed in LAEVA 
  (provided they respect the signature of the function symbol in [EVA_APP])
  and not in the abstract syntax,
  where all the components of a tuple must have type [EVA_NUMBER]
  (see {!Abstract.eva_type}.
  
  We provide here two function for type checking
  which correspond to each of the discipline:
  - {!Types.type_lterm_laeva} which conforms to the LAEVA discipline,
    and checks in [EVA_APP] terms the type of the arguments againts the declared 
    signature. 
  - {!Types.type_lterm} which conforms to the disipline of the abstract syntax.
    It must be call after term conversion and type coercion. *)


(** [type_lterm sl t]
  checks that the given located term t is well typed 
  wrt to the given symbol list sl
  according to the abstract syntax type discipline (see remark above):
   - in [EVA_APP(c, args)], the terms of [args] must all have type [EVA_NUMBER]
   - in [EVA_CONS(t, l)], [t] must have type [EVA_NUMBER]
   - in [EVA_CRYPT(a, k, t)], [a] must have type [EVA_SALGO] and
  [k] and [t] must have type [EVA_NUMBER].
  @return the type (not located) of the given located term [t].
 
  {b Note:} the term [t] must NOT contain alias symbols
 
  {b Note:} this function should not be applied directly 
  to terms obtained after parsing EVA protocol spec.
  {!Translator.convert_leterm} ({!Translator.check_lterm})
  and {!Types.coerce_lterm} should be applied first.
 
  @raise Type_error if the given term 
  - contains symbols not declared in [sl]
  - or is not well formed (arity) according [sl]
  - or is not well typed according to [sl].
 
  @raise Type_error if the given term contains alias symbols 
  or some leaf symbol is not conform to symbol list [sl].  *)
val type_lterm :
  eva_symbol_list -> eva_lterm -> eva_type

(** [type_lterm_laeva sl t]
  checks the typing of the given located term [t]
  wrt to the given symbol list [sl],
  according to the type discipline of LAEVA (aka concrete syntax) 
  which is less resctrictive than the discipline for the abstract syntax 
  (see remark above):
   - in [EVA_APP(c, args)], 
     the type of the terms of [args] must follow the declared signature of [c],
   - in [EVA_CONS(t, l)], the type of [t] is not checked.
   - in [EVA_CRYPT(a, k, t)], [a] must have type [EVA_SALGO], the
     types of [k] and [t] are not checked.
  
  The purpose of this function is to check 
  the terms obtained from the parser,
  after conversion with {!Translator.convert_lterm}, 
  and not yet casted,
  for conformity of terms against
  the type declaration of function symbols. 

  @return the type (not located) of the given located term [t].
 
  {b Note:} the term must NOT contain alias symbols
 
  {b Note:} this function should not be applied directly 
  to terms obtained after parsing EVA protocol spec.
  {!Translator.convert_leterm} (optionaly {!Translator.check_lterm})
  should be applied first.
  
  {b Note:} this function will likely fail when
  applied to coerced located terms as returned by
  {!Types.coerce_lterm} (hence well typed terms wrt EVA semantics).
 
  @raise Type_error if the given term 
  - contains symbols not declared in [sl]
  - or is not well formed (arity) wrt [sl]
  - or is not well typed wrt [sl] as described above.
 
  @raise Type_error if the given term contains alias symbols 
  or some leaf symbol is not conform to the symbol list [sl]. *)
val type_lterm_laeva :
  eva_symbol_list -> eva_lterm -> eva_type


(**************************************************************
 * Coercions
 **************************************************************)

(** [coerce_lterm sl t]
  typechecks the given located term [t]
  wrt the given symbol list [sl]
  and casts it to the type [EVA_NUMBER] (see {!Abstract.eva_type},
  @return the casted copy.
  
  {b Note:} the symbols in the term [t] must be conform to 
  the entries in symbol list [sl], in particular
  this function should not be applied directly 
  to terms obtained after parsing EVA protocol spec,
  see {!Translator.convert_lterm} and {!Translator.check_lterm}.
 
  @raise Type_error if the given term contains symbols not declared in [sl]
  or is not well formed (arity) according [sl]
  or is not well typed according to [sl].
 
  @raise Type_error if some leaf symbol 
  in the given term [t] is not conform to symbol list [sl]. *)
val coerce_lterm :
  eva_symbol_list -> eva_lterm -> eva_lterm



(* descend_coerce_lterm sl t
 * return a term obtained from [t] by 
 * appropriate application of {!Types.coerce_lterm}
 * to the strict subterms, but not to [t] itself.
 * 
 * {b Note:} the symbols in the term must be conform to 
 * the entries in symbol list, in particular
 * this function should not be applied directly 
 * to terms obtained after parsing EVA protocol spec.
 * see Translator.convert_leterm and Translator.check_lterm
 *
 * raise Type_error if the given term contains symbols not declared in sl
 * or is not well formed (arity) according sl
 * or is not well typed according to sl. *)
val descend_coerce_lterm :
  eva_symbol_list -> eva_lterm -> eva_lterm


(** [cast_term t ty]
  - param  [t] a located term
  - param [ty] an EVA type
  @return the term of type NUMBER obtained from [t] by applying 
  the appropriate coercion function.
  NB: the coercion is not applied recursively. *) 
val cast : eva_lterm -> eva_type -> eva_lterm


(** [coerce_spec sl sp]
  applies {!Types.coerce_lterm} [sl]
  to every term of the given spec [sp],
  except in (codomains of) sessions.
  @return the spec obtained.  

  {b Note:} that the session (i.e. substitutions)\
  do not need to be coerced because they are applied
  to coerced (typed) variable. *)
val coerce_spec :
  eva_symbol_list -> eva_spec -> eva_spec
